import numpy as np
import pprint
import json
from .ops import dequantize
from .ops import bounds
from .ops import np_array_from_arcs
from .ops import winding_order


def instance(obj):
    return type(obj).__name__


# ----------------- topology options object ------------------
class TopoOptions(object):
    def __init__(
        self,
        object=None,
        topology=True,
        prequantize=False,
        topoquantize=False,
        presimplify=False,
        toposimplify=False,
        shared_coords=False,
        prevent_oversimplify=True,
        simplify_with="shapely",
        simplify_algorithm="dp",
        winding_order=None,
        object_name="data",
    ):
        # get all arguments
        arguments = locals()
        if arguments["object"]:
            arguments = arguments["object"]

        if "topology" in arguments:
            self.topology = arguments["topology"]
        else:
            self.topology = True

        if "prequantize" in arguments:
            self.prequantize = arguments["prequantize"]
        else:
            self.prequantize = False

        if "topoquantize" in arguments:
            self.topoquantize = arguments["topoquantize"]
        else:
            self.topoquantize = False

        if "presimplify" in arguments:
            self.presimplify = arguments["presimplify"]
        else:
            self.presimplify = False

        if "toposimplify" in arguments:
            self.toposimplify = arguments["toposimplify"]
        else:
            self.toposimplify = False

        if "shared_coords" in arguments:
            self.shared_coords = arguments["shared_coords"]
        else:
            self.shared_coords = False

        if "prevent_oversimplify" in arguments:
            self.prevent_oversimplify = arguments["prevent_oversimplify"]
        else:
            self.prevent_oversimplify = True

        if "simplify_with" in arguments:
            self.simplify_with = arguments["simplify_with"]
        else:
            self.simplify_with = "shapely"

        if "simplify_algorithm" in arguments:
            self.simplify_algorithm = arguments["simplify_algorithm"]
        else:
            self.simplify_algorithm = "dp"

        if "winding_order" in arguments:
            self.winding_order = arguments["winding_order"]
        else:
            self.winding_order = None

        if "object_name" in arguments:
            if type(arguments["object_name"]) is not list:
                self.object_name = [arguments["object_name"]]
            else:
                self.object_name = arguments["object_name"]
        else:
            self.object_name = ["data"]

    def __repr__(self):
        return "TopoOptions(\n  {}\n)".format(pprint.pformat(self.__dict__))


# --------- supportive functions for serialization -----------
def coordinates(arcs, tp_arcs, geom_type):
    """
    Return GeoJSON coordinates for the sequence(s) of arcs.

    The arcs parameter may be a sequence of ints, each the index of a coordinate
    sequence within tp_arcs within the entire topology describing a line string, a
    sequence of such sequences, describing a polygon, or a sequence of polygon arcs.
    """

    if isinstance(arcs[0], int):
        coord_list = []
        for i, arc in enumerate(arcs):
            arc_coords = tp_arcs[arc if arc >= 0 else ~arc][:: arc >= 0 or -1]
            arc_coords = arc_coords[~np.isnan(arc_coords).any(axis=1)]
            coord_list.append(arc_coords[i > 0 :])
        coords = np.concatenate(coord_list).tolist()
        if geom_type in ["Polygon", "MultiPolygon"]:
            if len(coords) < 3:
                # This may happen if an arc has only two points.
                coords.extend([coords[0]])
        elif geom_type in ["LineString", "MultiLineString"]:
            if len(coords) < 2:
                # This should never happen per the specification.
                coords.extend([coords[0]])
        return coords
    elif isinstance(arcs[0], (list, tuple)):
        return list(coordinates(arc, tp_arcs, geom_type) for arc in arcs)
    else:
        raise ValueError("Invalid input %s", arcs)


def geometry(obj, tp_arcs, transform=None):
    """
    Converts a topology object to a geometry object.

    The topology object is a dict with 'type' and 'arcs' items.
    """
    if obj["type"] == "GeometryCollection":
        geometries = [geometry(feat, tp_arcs) for feat in obj["geometries"]]
        return {"type": obj["type"], "geometries": geometries}

    if obj["type"] == "MultiPoint":
        if transform is not None:
            scale = transform["scale"]
            translate = transform["translate"]
            coords = obj["coordinates"]
            point_coords = dequantize(np.array(coords).T, scale, translate).T.tolist()
        else:
            point_coords = obj["coordinates"]
        return {"type": obj["type"], "coordinates": point_coords}

    if obj["type"] == "Point":
        if transform is not None:
            scale = transform["scale"]
            translate = transform["translate"]
            coords = [obj["coordinates"]]
            point_coord = dequantize(np.array(coords), scale, translate).tolist()
        else:
            point_coord = [obj["coordinates"]]
        return {"type": obj["type"], "coordinates": point_coord[0]}

    else:
        return {
            "type": obj["type"],
            "coordinates": coordinates(obj["arcs"], tp_arcs, obj["type"]),
        }


def prettyjson(obj, indent=2, maxlinelength=80):
    """Renders JSON content with indentation and line splits/concatenations to fit maxlinelength.
    Only dicts, lists and basic types are supported"""

    # https://stackoverflow.com/a/56497521/104668

    items, _ = getsubitems(
        obj, itemkey="", islast=True, maxlinelength=maxlinelength, level=0
    )
    return indentitems(items, indent, level=0)


def getsubitems(obj, itemkey, islast, maxlinelength, level):  # noqa: C901
    items = []
    is_inline = (
        True  # at first, assume we can concatenate the inner tokens into one line
    )

    isdict = isinstance(obj, dict)
    islist = isinstance(obj, list)
    istuple = isinstance(obj, tuple)
    isbasictype = not (isdict or islist or istuple)

    # build json content as a list of strings or child lists
    if isbasictype:
        # render basic type
        keyseparator = "" if itemkey == "" else ": "
        itemseparator = "" if islast else ","
        items.append(itemkey + keyseparator + basictype2str(obj) + itemseparator)

    else:
        # render lists/dicts/tuples
        if isdict:
            opening, closing, keys = ("{", "}", iter(obj.keys()))
        elif islist:
            opening, closing, keys = ("[", "]", range(0, len(obj)))
        elif istuple:
            opening, closing, keys = (
                "[",
                "]",
                range(0, len(obj)),
            )  # tuples are converted into json arrays

        if itemkey != "":
            opening = itemkey + ": " + opening
        if not islast:
            closing += ","

        itemkey = ""
        subitems = []

        # get the list of inner tokens
        for (i, k) in enumerate(keys):
            islast_ = i == len(obj) - 1
            itemkey_ = ""
            if isdict:
                itemkey_ = basictype2str(k)
            inner, is_inner_inline = getsubitems(
                obj[k], itemkey_, islast_, maxlinelength, level + 1
            )
            subitems.extend(inner)  # inner can be a string or a list
            is_inline = (
                is_inline and is_inner_inline
            )  # if a child couldn't be rendered inline, then we are not able either

        # fit inner tokens into one or multiple lines, each no longer than maxlinelength
        if is_inline:
            multiline = True

            # in Multi-line mode items of a list/dict/tuple can be rendered in multiple lines if they don't fit on one.
            # suitable for large lists holding data that's not manually editable.

            # in Single-line mode items are rendered inline if all fit in one line, otherwise each is rendered in a separate line.
            # suitable for smaller lists or dicts where manual editing of individual items is preferred.

            # this logic may need to be customized based on visualization requirements:
            if isdict:
                multiline = False
            if islist:
                multiline = True

            if multiline:
                lines = []
                current_line = ""

                for (i, item) in enumerate(subitems):
                    item_text = item
                    if i < len(inner) - 1:
                        item_text = item + ","

                    if len(current_line) > 0:
                        try_inline = current_line + " " + item_text
                    else:
                        try_inline = item_text

                    if len(try_inline) > maxlinelength:
                        # push the current line to the list if maxlinelength is reached
                        if len(current_line) > 0:
                            lines.append(current_line)
                        current_line = item_text
                    else:
                        # keep fitting all to one line if still below maxlinelength
                        current_line = try_inline

                    # Push the remainder of the content if end of list is reached
                    if i == len(subitems) - 1:
                        lines.append(current_line)

                subitems = lines
                if len(subitems) > 1:
                    is_inline = False
            else:  # single-line mode
                totallength = len(subitems) - 1  # spaces between items
                for item in subitems:
                    totallength += len(item)
                if totallength <= maxlinelength:
                    str = ""
                    for item in subitems:
                        str += (
                            item + " "
                        )  # insert space between items, comma is already there
                    subitems = [str.strip()]  # wrap concatenated content in a new list
                else:
                    is_inline = False

        # attempt to render the outer brackets + inner tokens in one line
        if is_inline:
            item_text = ""
            if len(subitems) > 0:
                item_text = subitems[0]
            if len(opening) + len(item_text) + len(closing) <= maxlinelength:
                items.append(opening + item_text + closing)
            else:
                is_inline = False

        # if inner tokens are rendered in multiple lines already, then the outer brackets remain in separate lines
        if not is_inline:
            items.append(opening)  # opening brackets
            items.append(subitems)  # Append children to parent list as a nested list
            items.append(closing)  # closing brackets

    return items, is_inline


def basictype2str(obj):
    if isinstance(obj, str):
        strobj = '"' + str(obj) + '"'
    elif isinstance(obj, bool):
        strobj = {True: "true", False: "false"}[obj]
    # convert None to 'null' since None is not a valid syntax in JSON
    elif obj is None:
        strobj = "null"
    else:
        strobj = str(obj)
    return strobj


def indentitems(items, indent, level):
    """Recursively traverses the list of json lines, adds indentation based on the current depth"""
    res = ""
    indentstr = " " * (indent * level)
    for (i, item) in enumerate(items):
        if isinstance(item, list):
            res += indentitems(item, indent, level + 1)
        else:
            islast = i == len(items) - 1
            # no new line character after the last rendered line
            if level == 0 and islast:
                res += indentstr + item
            else:
                res += indentstr + item + "\n"
    return res


# ----------------- serialization functions ------------------
def serialize_as_topojson(data, options):

    # change options to reflect this
    options.prequantize = False
    options.presimplify = False

    arcs_asarray = [np.asarray(a) for a in data["arcs"]]
    parse_topo = {
        "type": "Topology",
        "linestrings": arcs_asarray,
        "coordinates": [],
        "options": options,
        "objects": data["objects"],
    }
    if "transform" in data.keys():
        parse_topo["transform"] = data["transform"]
        if "bbox" in data.keys():
            parse_topo["bbox"] = data["bbox"]
        else:
            scale = data["transform"]["scale"]
            translate = data["transform"]["translate"]
            dequ_arcs = dequantize(np_array_from_arcs(data["arcs"]), scale, translate)
            parse_topo["bbox"] = bounds(dequ_arcs)
    else:
        parse_topo["bbox"] = bounds(arcs_asarray)

    return parse_topo, options


def serialize_as_geodataframe(fc, crs=None):
    """
    Convert a topology dictionary or string into a GeoDataFrame.

    Parameters
    ----------
    fc : dict, str
        a complete object representing a GeoJSON file
    crs : str, dict
        coordinate reference system to set on the resulting frame.
        Default is `None`.

    Returns
    -------
    geopandas.GeoDataFrame
        topojson object parsed as GeoDataFrame
    """
    from geopandas import GeoDataFrame
    from pandas import json_normalize

    features = fc["features"]
    return GeoDataFrame.from_features(features=features, crs=crs).set_axis(
        json_normalize(features)["id"].values
    )


def serialize_as_svg(topo_object, separate=False, include_junctions=False):
    from IPython.display import SVG, display
    from shapely import geometry

    if "arcs" in topo_object:
        arcs = topo_object["arcs"]
        if arcs:
            # dequantize if quantization is applied
            if "transform" in topo_object:

                np_arcs = np_array_from_arcs(arcs)

                transform = topo_object["transform"]
                scale = transform["scale"]
                translate = transform["translate"]

                np_arcs = dequantize(np_arcs, scale, translate)
                l_arcs = []
                for ls in np_arcs:
                    l_arcs.append(ls[~np.isnan(ls)[:, 0]].tolist())
                arcs = l_arcs

            arcs = [geometry.LineString(arc) for arc in arcs]

    else:
        arcs = topo_object["linestrings"]

    if separate and not include_junctions:
        for ix, line in enumerate(arcs):
            line = geometry.LineString(line)
            svg = line._repr_svg_()
            print(ix, line.wkt)
            display(SVG(svg))
    elif separate and include_junctions:
        pts = topo_object["junctions"]
        for ix, line in enumerate(arcs):
            line = geometry.LineString(line)
            svg = geometry.GeometryCollection(
                [line, geometry.MultiPoint(pts)]
            )._repr_svg_()
            print(ix, line.wkt)
            display(SVG(svg))

    elif not separate and include_junctions:
        pts = topo_object["junctions"]
        display(
            geometry.GeometryCollection(
                [geometry.MultiLineString(arcs), geometry.MultiPoint(pts)]
            )
        )
    else:
        display(geometry.MultiLineString(arcs))


def serialize_as_json(topo_object, fp, pretty=False, indent=4, maxlinelength=88):
    # https://stackoverflow.com/a/57915246
    class NpEncoder(json.JSONEncoder):
        def default(self, obj):
            if isinstance(obj, np.integer):
                return int(obj)
            if isinstance(obj, np.floating):
                return float(obj)
            if isinstance(obj, np.ndarray):
                return obj.tolist()
            return super(NpEncoder, self).default(obj)

    if fp:
        with open(fp, "w") as f:
            if pretty is True:
                print(
                    prettyjson(topo_object, indent=indent, maxlinelength=maxlinelength),
                    file=f,
                )
            elif pretty is False:
                json.dump(topo_object, separators=(",", ":"), fp=f, cls=NpEncoder)
            else:
                json.dump(topo_object, fp=f, cls=NpEncoder)
    else:
        if pretty is True:
            return prettyjson(topo_object, indent=indent, maxlinelength=maxlinelength)
        elif pretty is False:
            return json.dumps(topo_object, separators=(",", ":"), cls=NpEncoder)
        else:
            return json.dumps(topo_object, cls=NpEncoder)


def serialize_as_geojson(
    topo_object,
    fp=None,
    validate=False,
    objectname="data",
    order="CCW_CW",
    decimals=None,
):
    from shapely.geometry import shape

    # prepare arcs from topology object
    arcs = topo_object["arcs"]
    transform = None
    if "transform" in topo_object:
        transform = topo_object["transform"]
        scale = transform["scale"]
        translate = transform["translate"]

    if arcs:
        np_arcs = np_array_from_arcs(arcs)
        # dequantize if quantization is applied
        if transform:
            np_arcs = dequantize(np_arcs, scale, translate)
    else:
        np_arcs = None

    # evenly round the coordinates to the given number of decimals
    if decimals is not None and isinstance(decimals, int):
        np_arcs = np.around(np_arcs, decimals=decimals)

    # select object member from topology object
    if objectname not in topo_object["objects"]:
        raise LookupError(f"'{objectname}' is not an object name in your topojson file")
    features = topo_object["objects"][objectname]["geometries"]

    # prepare geojson featurecollection
    fc = {"type": "FeatureCollection", "features": []}

    # fill the featurecollection with geometry object members
    for index, feature in enumerate(features):
        f = {"id": feature.get("id", index), "type": "Feature"}
        f["properties"] = feature.get("properties", {})

        # the transform is only used in cases of points or multipoints
        geom_map = geometry(feature, np_arcs, transform)

        # enforce right-hand rule on geometry for GeoJSON
        geom_map = winding_order(geom=shape(geom_map), order=order)

        if validate:
            geom = shape(geom_map).buffer(0)

            assert geom.is_valid
            f["geometry"] = geom_map.__geo_interface__
        else:
            f["geometry"] = geom_map.__geo_interface__

        fc["features"].append(f)
    return fc


def serialize_as_altair(
    topo_object,
    color=None,
    tooltip=True,
    projection="identity",
    objectname="data",
    geo_interface=False,
):
    import altair as alt

    if projection == "identity":
        reflectY = True
    else:
        reflectY = False
    # create a mesh visualization
    if geo_interface:
        # chart object
        chart = (
            alt.Chart(topo_object, width=300)
            .mark_geoshape(filled=False)
            .project(type=projection, reflectY=reflectY)
        )

    # create a mesh visualization
    elif color is None:
        data = alt.InlineData(
            values=topo_object, format=alt.DataFormat(mesh=objectname, type="topojson")
        )
        chart = (
            alt.Chart(data, width=300)
            .mark_geoshape(filled=False)
            .project(type=projection, reflectY=reflectY)
        )

    # creating a chloropleth visualization
    else:
        data = alt.InlineData(
            values=topo_object,
            format=alt.DataFormat(feature=objectname, type="topojson"),
        )
        if tooltip is True:
            tooltip = [color]
        chart = (
            alt.Chart(data, width=300)
            .mark_geoshape()
            .encode(
                color=alt.Color(color, legend=alt.Legend(columns=2)), tooltip=tooltip
            )
            .project(type=projection, reflectY=reflectY)
        )

    return chart


def serialize_as_ipywidgets(topo_object, toposimplify, topoquantize):
    from ipywidgets import interact
    from ipywidgets import fixed
    import ipywidgets as widgets

    style = {"description_width": "initial"}
    ts = toposimplify
    tq = topoquantize

    # set to simplification package for speed
    topo_object.options.simplify_with = "simplification"

    alg = widgets.RadioButtons(
        options=[("Douglas-Peucker", "dp"), ("Visvalingam-Whyatt", "vw")],
        value="vw",
        description="Simplify algorithm",
        disabled=False,
        style=style,
    )
    eps = widgets.FloatSlider(
        min=ts["min"],
        max=ts["max"],
        step=ts["step"],
        value=ts["value"],
        description="Toposimplify Factor",
        style=style,
    )
    qnt = widgets.FloatLogSlider(
        min=tq["min"],
        max=tq["max"],
        step=tq["step"],
        value=tq["value"],
        base=tq["base"],
        description="Topoquantize Factor",
        style=style,
    )

    return interact(
        toposimpquant, epsilon=eps, quant=qnt, algo=alg, topo=fixed(topo_object)
    )


def toposimpquant(epsilon, quant, algo, topo):
    topo.options.simplify_algorithm = algo
    return topo.toposimplify(epsilon).topoquantize(quant).to_alt()


def example_data_africa():
    import pickle

    dict_bytes = b"\x80\x04\x95h\xca\x00\x00\x00\x00\x00\x00}\x94(K\x00}\x94(\x8c\x02id\x94\x8c\x011\x94\x8c\x04type\x94\x8c\x12GeometryCollection\x94\x8c\nproperties\x94}\x94(\x8c\tcontinent\x94\x8c\x06Africa\x94\x8c\ngdp_md_est\x94GA\x02b@\x00\x00\x00\x00\x8c\x06iso_a3\x94\x8c\x03TZA\x94\x8c\x04name\x94\x8c\x08Tanzania\x94\x8c\x07pop_est\x94J\xd797\x03u\x8c\x04bbox\x94(G@=W\n\x15\x0e9\xc8G\xc0'q\x1e\xc9+\x7f\xd9G@D(\x86\x05h\x1e\xd0G\xbf\xeefffffgt\x94\x8c\ngeometries\x94]\x94}\x94(h\x04\x8c\x07Polygon\x94\x8c\x0bcoordinates\x94(G@@\xf3\xac\xce\xfaK\xe6G\xbf\xeefffffg\x86\x94G@A\tK\x9c\xb6\x84\x88G\xbf\xf0\xf5\x05\xd0\xfaX\x00\x86\x94G@B\xd9n\xac\x86\x05`G\xc0\x08\xc6\xa2\xb1pO\x80\x86\x94G@B\xe2)\xc7y\xa6\xb5G\xc0\rj\xbd\xe3\xfb\xbd|\x86\x94G@C\x99\xe2XOLnG\xc0\x12\xb5\x031\xe3\xa7\xdb\x86\x94G@C^\xca\x03\xc4\xb0\x98G\xc0\x17\xa2\xc3\xc9\xee\xcb\xc0\x86\x94G@Cf^\xdd\x05)@G\xc0\x19\xe7\x13j@\x0f\xc0\x86\x94G@C\xb8Q\xeb\x85\x1e\xb8G\xc0\x1b\\(\xf5\xc2\x8e\xc0\x86\x94G@C\xbc(\xf5\xc2\x8fpG\xc0\x1cfffff@\x86\x94G@C\x98\xeb\x9a\x17m\xd8G\xc0\x1e\xd0\xcb)^\x9e\x00\x86\x94G@C\xa0B\x84\xdf\xce8G\xc0 \x03\xff\xac\x1d)\xc0\x86\x94G@C\x97\xdf\xe3*\x06pG\xc0 \xf8\x94\xc4G\xc2\xe0\x86\x94G@C\xc4\x93 \xd9\x94hG\xc0\"9\x88\x8f\x86\x1a \x86\x94G@C\xf9\x8c~($\x10G\xc0$2a|\x1b\xda`\x86\x94G@D(\x85\xe5\xc62\x92G\xc0$\xa2Z\xa2\xfb\xf8\xf2\x86\x94G@D(\x86\x05h\x1e\xd0G\xc0$\xa2Z\xeec\x1f@\x86\x94G@C\xc2\xb0 \xc4\x9b\xa0G\xc0%\xcb3\xda\xf8\xdf\x80\x86\x94G@C6\xba,\xa9S\x04G\xc0&\x92\x06\n\x05\xf0\x10\x86\x94G@B\xe9\xf0\x1b\x86n@G\xc0&\x89\x9e\xd7\xc6\xfb\xa0\x86\x94G@B\xbcS;\x10w@G\xc0'#4\x82\xbe\x8b\xc0\x86\x94G@Bc8%\xd5\xae\xf4G\xc0'0g6e\x85\x8c\x86\x94G@BA\xcdm\x83\xc2\xdcG\xc0'q\x1e\xc9+\x7f\xd9\x86\x94G@A\xa7\xfc\xa7\x8a\x93\xd8G\xc0&\xe0\xd7\xcc\x95M\x94\x86\x94G@AG\xad\xb8\x9b\xa3@G\xc0'\n@\x10\xd9\x92\x8c\x86\x94G@A#\xd7\n=p\xa0G\xc0$Q\xeb\x85\x1e\xb8`\x86\x94G@@\xf8m^\xdb\xeb8G\xc0#c)7\xc2\xca\xc0\x86\x94G@@\xde\xaf%\x1c\x19HG\xc0\"\xd5\x94\xafO\r\x80\x86\x94G@@a36\xe6\xb70G\xc0\"v\x11\x14CN\xa0\x86\x94G@@\x18\x8f\x07\x1dR\xf0G\xc0!\xdcX\x032:\xe0\x86\x94G@?\x8el\xd43\xcfxG\xc0!\x86+D\x0c\xad|\x86\x94G@?(bdA\x19\xcaG\xc0!0l\xa0%?\xdd\x86\x94G@>\xbdqG\x1b-\xfdG\xc0 \xae\x15A\xde\xa1m\x86\x94G@>\xbdq\xa7\xd49 G\xc0 \xae\x15s\xd6\x0c\xc4\x86\x94G@>32\xfd)\x8c\x10G\xc0\x1cQ\xe6\x88\x19=`\x86\x94G@=\x9e\xbam\xcd`\x91G\xc0\x1a\x14~\xda\x04\xb8\xa8\x86\x94G@=k\x84\xa4jc\x18G\xc0\x17\xc2\x8f\x10\xa1\xb1\xd8\x86\x94G@=\x85\x1d\xd7\x9d\x96\xb8G\xc0\x15\xae\x0e\xf5S'\x00\x86\x94G@=W\n\x15\x0e9\xc8G\xc0\x11\xff\xfb\xa6\xd1a\xe8\x86\x94G@=\xc0\xe60\\\xbd@G\xc0\x11\xcf?+\xf0\x19\x80\x86\x94G@>\x1d\xc7%\xc3\xde\xf0G\xc0\x10\\Hj\xd2\xdc\xc0\x86\x94G@>\x81k\x11\xc6\xd1\xe0G\xc0\x0c\x8cs\xab\xc9F\x80\x86\x94G@>\xc0\x92\xcc\xf6\xbePG\xc0\n\xdf\xdd\xeb\xd9\x01\x80\x86\x94G@>\xbe5\xe7B\x99\xe0G\xc0\x08FDR?g\x80\x86\x94G@>\x87\x14\xb9\xcbhPG\xc0\x06v\x01y|\xc3\x80\x86\x94G@>x<\x884ViG\xc0\x03O\x93\x158Z\xaf\x86\x94G@>x<K\t\xe9\x90G\xc0\x03O\x86\x1a`\xd4\x00\x86\x94G@>\xc2 \x89\x1b\xe7HG\xc0\x02LI\xdc\xfe\x9e\xf0\x86\x94G@>\xd0\xee70biG\xbf\xfb.\xc0\x86\xc9\xe1?\x86\x94G@>kJt\xa1\x04\x80G\xbf\xf2'\x90P,\xb6\x00\x86\x94G@>\xc5\x15\x8b\x82\x7f\xc0G\xbf\xf0;\x98\xc7\xe2\x82\x00\x86\x94G@?\xdd\xbdQ.\xc6\xd0G\xbf\xf0p\x11\n\x13~\x00\x86\x94G@@\xf3\xac\xce\xfaK\xe6G\xbf\xeefffffg\x86\x94t\x94\x85\x94uauK\x01}\x94(h\x02\x8c\x012\x94h\x04h\x05h\x06}\x94(h\x08\x8c\x06Africa\x94h\nG@\x8cT\x00\x00\x00\x00\x00h\x0b\x8c\x03ESH\x94h\r\x8c\tW. Sahara\x94h\x0fJu4\t\x00uh\x10(G\xc01\x10<\x81\"[\xb4G@4\xff\xef\xc0\xf6\xc2XG\xc0!T\x8b2\r\xd0\x88G@;\xa8\x0b\x86\xf0\x0c\x00t\x94h\x12]\x94}\x94(h\x04h\x15h\x16(G\xc0!T\xc8'\xd0_\xe0G@;\xa8\x0b\x86\xf0\x0c\x00\x86\x94G\xc0!T\x8b2\r\xd0\x88G@;\x96\xe8\x19\xb6\x04\xdf\x86\x94G\xc0!^i\xa6\x1bw\xf0G@;eO|\xb2\x9a8\x86\x94G\xc0!_\xe4\xf4\x9d,\xe0G@9\xe1\x8c\xe6\x82\xca,\x86\x94G\xc0'\xf0W\xac\xef5HG@9\xee\xf05\x04\x7fK\x86\x94G\xc0'\xdf\xdb\xe3\x8c8hG@7_\xe5h7\xa7@\x86\x94G\xc0)\xbf\x99\xf8\x07!\xf1G@7H\xea\xc4`\x9c\x9c\x86\x94G\xc0*<\xcda\xd7Z\xb0G@6\xc5n\xaf\xe5\xb9\x14\x86\x94G\xc0)\xdb\xb3?\xb56\xe0G@5S\xba\xe6\x82\xb6$\x86\x94G\xc00\xd8^\x9cp\xdctG@5UT\xaf\xe5\xb2\xc1\x86\x94G\xc01\x10<\x81\"[\xb4G@4\xff\xef\xc0\xf6\xc2X\x86\x94G\xc01\x05:\xcc:@@G@5l\x1c\x86\xef\xf0\xb0\x86\x94G\xc01\x00\xc2\x1a\xbb\xf5\x0cG@5k\xb5;\xd8\x0b\xe1\x86\x94G\xc0-\x80}\x1d\x93$\xc8G@5\x80'S\xbc\xbd\xb8\x86\x94G\xc0-B\xfc\x80\x8f\xb9\xe8G@5\xdcf\x8d\xc3\x92\xf0\x86\x94G\xc0,q<\xe6\xf6\x1c\xb8G@6Of\xd8\xdby\x9c\x86\x94G\xc0+\xc8?\x9fH\x05PG@7\xb0\xe5\xf7\x93\xd1\x99\x86\x94G\xc0)\x00~.\xa4\"\x10G@8\xc5&W&\x9cP\x86\x94G\xc0(\x0f\xbf\x9fG\xf5\x00G@:\x07\xe6\xd8\xdb\x8a\x00\x86\x94G\xc0'o\xba\x80\x8f\xa0HG@:\x1a\xa5\xc0\xf6\xd8\xcc\x86\x94G\xc0&\xc8\xfc\xf4\x9dD\xc0G@:\xe2(\x12\xe2a`\x86\x94G\xc0%\x1a?\x16\xbf_hG@:\xfd\xa5\x91-y\xb8\x86\x94G\xc0$`\xfc5w\xae(G@:\xdcf\xdf\xaf.\r\x86\x94G\xc0#x~\xed\xc9\x95\xa8G@:\xdcf\xdf\xaf.\r\x86\x94G\xc0\"\xd3y\xa6\x1b~`G@;\x16\xa6]\xfaG\x00\x86\x94G\xc0!\x96\xfb\t\x18\x0f\x18G@;\x1e\xe5\xf4*\x10y\x86\x94G\xc0!\xa2\xbaeA\x05\x10G@;\xa8\x0b\x86\xf0\x0c\x00\x86\x94G\xc0!T\xc8'\xd0_\xe0G@;\xa8\x0b\x86\xf0\x0c\x00\x86\x94t\x94\x85\x94uauK\x02}\x94(h\x02\x8c\x0211\x94h\x04h\x05h\x06}\x94(h\x08\x8c\x06Africa\x94h\nG@\xf0\x1d\xa0\x00\x00\x00\x00h\x0b\x8c\x03COD\x94h\r\x8c\x0fDem. Rep. Congo\x94h\x0fJ\x1f\x13\xf7\x04uh\x10(G@(][B\x00\x8e G\xc0*\x83\xb36e\x94,G@?,\x95\n\xd0\xd18G@\x15\x06;\xdeN\x02\x10t\x94h\x12]\x94}\x94(h\x04h\x15h\x16(G@=W\n\x15\x0e9\xc8G\xc0\x11\xff\xfb\xa6\xd1a\xe8\x86\x94G@=\x85\x1d\xd7\x9d\x96\xb8G\xc0\x15\xae\x0e\xf5S'\x00\x86\x94G@=k\x84\xa4jc\x18G\xc0\x17\xc2\x8f\x10\xa1\xb1\xd8\x86\x94G@=\x9e\xbam\xcd`\x91G\xc0\x1a\x14~\xda\x04\xb8\xa8\x86\x94G@>32\xfd)\x8c\x10G\xc0\x1cQ\xe6\x88\x19=`\x86\x94G@>\xbdq\xa7\xd49 G\xc0 \xae\x15s\xd6\x0c\xc4\x86\x94G@>\xbdqG\x1b-\xfdG\xc0 \xae\x15A\xde\xa1m\x86\x94G@>X\x99\x18x\x0ehG\xc0 y\xfc\xc2W\xc0\xc8\x86\x94G@=\x00\xbe\xdb\x07d\xb0G\xc0 \xd0fw?\xdd\x90\x86\x94G@<\xbc 70Y?G\xc0!\r\x99/\x91\xca$\x86\x94G@<s*\xbf\xb8\xe0\x88G\xc0\"Tp,'\xff\x88\x86\x94G@<\xacvf\xf9\xbb\xf8G\xc0#6;\xcc\x95=p\x86\x94G@<~\xfem\xcd[\xa0G\xc0%\x94k\xa3\x9f\x85L\x86\x94G@<_K\xf9\xbf\xb3\xd0G\xc0'\x96X\xdd\xa6a\xc0\x86\x94G@<\xa4ux\n\xcc\xe8G\xc0'\xf1qs\xd6,\xb5\x86\x94G@=Wo\xae\xa7\xd3`G\xc0(\xb8\xb3m\x02\x8f\xb4\x86\x94G@=\x9d\xb2D\xd7\x9d\xf8G\xc0([\x98\x10\xd9\x98X\x86\x94G@=\xb3\x19\xe5D\xd8`G\xc0*\x83\xb36e\x94,\x86\x94G@<\xef-\\\xbcLwG\xc0*\x7fwz\xa9\xd8`\x86\x94G@<\x86\x08\"\xb5w\x08G\xc0)e\xafz\xa9\xd3\x88\x86\x94G@<'\xb53\xc6\x86\x7fG\xc0(\x8b\x82\x92\x8e\x81H\x86\x94G@;c\x88R~\xd5\x08G\xc0(C\xf7z\xa9\xce\x8c\x86\x94G@;*\x17jc\x85\x88G\xc0'7\xad\xe1\x100Y\x86\x94G@:\x8d\x97&\x1f>\x98G\xc0'\xd9P,(\x17\xd0\x86\x94G@9\xc0\x97\\\xbc>xG\xc0'\x91\xe6\xf2!B\xf4\x86\x94G@9k\t\xc9\xf6C\xd8G\xc0&\xa9pplV\xdc\x86\x94G@8\xc8}\xd0\xc9\xe1yG\xc0&z6\n\x05\xef\xa4\x86\x94G@8P\x84\"\xb5d\x88G\xc0&\x86\x91/\x91\xe24\x86\x94G@8A\xd4\xef\x821\x19G\xc0%\xe7k\x95\xf8E\xe1\x86\x94G@7\xe9\x86\xef\x82/\x90G\xc0%\xda\x88\xf8\xf4\xdb\xd8\x86\x94G@7t\xf0>\x03\xe2xG\xc0%\xbcX\x99b\x15X\x86\x94G@6\xd6\\D\xd7\x80(G\xc0&\t\x05\xb8\x1ah\x94\x86\x94G@6g\x1d\xc9\xf66\x90G\xc0%\xfctb\xc5\x13\x08\x86\x94G@6'\xbf\xa7\xd4\x13YG\xc0&+k\ro\xbe\x84\x86\x94G@65p\xdb\x07F\xc8G\xc0#\xca\"\xbb\x84.\xec\x86\x94G@5\xe0\x0b\xec\x18V`G\xc0#\x0c#m\x02v\xc0\x86\x94G@5\xcdB\xdb\x07E\x00G\xc0!\xd1A\xfc^\x9aL\x86\x94G@5\xf2\xfa>\x03\xdb\xd9G\xc0 \x9c\x9f\ro\xa6\x10\x86\x94G@5\xbf\x17\xbcN\xf2\xd8G\xc0\x1f\xae*\xb1\x0f\x0e\xb8\x86\x94G@5\xbaex\n\xae\x80G\xc0\x1d)\xdazr\x00@\x86\x94G@4\x83\xc6\x89\x1b\xba8G\xc0\x1d2\xcb\xddn\x96\x98\x86\x94G@4\x9a\x11\x11\xa4C G\xc0\x1b\xc1\xdc\x81E\x9ax\x86\x94G@4\x17t\x82H\x17\xf0G\xc0\x1b\xc5\xb9i`\xe9\x10\x86\x94G@4\t\xa8703\x00G\xc0\x1cw'b\x8dK\xb0\x86\x94G@3j\xe1q7\x03\xe8G\xc0\x1c\x9f(\xaa;`\xd7\x86\x94G@3*\xa7,\xf2\xbe\x88G\xc0\x1e\xf3\xe6l\xca\xc7@\x86\x94G@3\x04I\xd7\x9dh\x88G\xc0\x1f\xf3\xf6\xbe\xb6P\xc8\x86\x94G@2v\xd470,\x18G\xc0\x1fcW\xb4x\xdd\xa9\x86\x94G@2\"\\YRL\xc0G\xc0\x1f\xf3a\xbbL\x80\x90\x86\x94G@1y\x14\x8f\xefM@G\xc0 #\x19!\xea\x85@\x86\x94G@1\x17\t\xf9\xbf\x820G\xc0\x1e.\xc9\x17un\x88\x86\x94G@0\xdc5x\n\x99\x10G\xc0\x1c\xe3\xa2\r7\xf88\x86\x94G@0\x92\xbb\xec\x18?\x10G\xc0\x1a}\x96\x88\x195Y\x86\x94G@0S\x97\\\xbc\x15\x08G\xc0\x17\x82\x87\x99*9H\x86\x94G@*\xc0NL>\t@G\xc0\x17t\xfb\xa6\xd1y\xe8\x86\x94G@*\x0c\xbb\xaf:\x9c`G\xc0\x17\xf0\x03\xa6\xd1|\x08\x86\x94G@)xh`\xb8\xe4\xe0G\xc0\x17\xdc\xdb\xc2\x1f\xfdh\x86\x94G@(\xa5\x15\xc3\xb5wqG\xc0\x18f~\xa3g\xad\xd8\x86\x94G@(][B\x00\x8e G\xc0\x17(\xe3\x8b\x82\xf6\xe8\x86\x94G@(\xdf\x95\x9a\xbf\xb5\xf1G\xc0\x16\xbc\xba(\x86^\xd8\x86\x94G@(\xef\x9e>\x96\xc0qG\xc0\x14\xfeRzq\xdcO\x86\x94G@)Cb\x9e)\x87\xd1G\xc0\x13\xf7\x0f\xcf\xc7- \x86\x94G@)\xfd\xb4n`(\x10G\xc0\x13\x1f\xd9\x84\xafD\xb8\x86\x94G@*\x848\x0e\xcdd_G\xc0\x13\x88%\xff\x90\x8e0\x86\x94G@+3Q\xfaR\x86 G\xc0\x12\x00$J\xa8l(\x86\x94G@,J7\xaf:\xa60G\xc0\x12\n?\xb4x\xa2\xf0\x86\x94G@,k\x06\x9e)\x95\xb0G\xc0\x13, XO\xb2(\x86\x94G@-*K\x0bc\x9f\xd0G\xc0\x13\xe1\x86Q|\x14\xd8\x86\x94G@.W\x8c7\xc37\xbfG\xc0\x11_\xc0XO\xaa@\x86\x94G@/\x81\xd0\x01&9\x80G\xc0\x0e\xd7`\xb0\x9fCP\x86\x94G@0\x01\x9c0\\\x80\xd8G\xc0\x0cG\xf3\xad5g\xe0\x86\x94G@/\xf2\x13B\x00\xafpG\xc0\x05\xb2\xfa\xb7r\xbb\x90\x86\x94G@0h7,\xf2\xb2`G\xbf\xfb\xda\xd6IY@`\x86\x94G@0\xdd\x84\xbf\xb8\xad\x98G\xbf\xf3\x9c\xf1\x97\xda\xd1\x00\x86\x94G@1\x86\x12D\xd7h\xc8G\xbf\xe7\xcdu\x1b:}\x00\x86\x94G@1\xa3~70(pG\xbf\xdb0q\n\x15\r\x00\x86\x94G@1\xa9\xde\x96\xc2\xee\x88G\xbf\xad\xbd/\x91\x7f\xa0\x00\x86\x94G@1\xd3\x98\"\xb5H\x00G?\xd2}\xb7\xeb\xaeK\x00\x86\x94G@1\xc61q6\xfc\xa8G?\xeba\x8eN\x96\x9a\xff\x86\x94G@1\xe6\x1a\x15\x0e\x07xG?\xfb\xde\x8b4\xf2\xcc\xc0\x86\x94G@2\x18\"t\xa0\xceHG@\x02\xec\xffj\xb0\x19@\x86\x94G@2d\xcf\x93Y!\x88G@\x074\x1b\xae\xf4p`\x86\x94G@2s\xfc\x15\x0e\t\xe8G@\x0c\x08\xfbxW\x82@\x86\x94G@2\x8b\x00\xe1\xda\xd7\x18G@\x10\xce\xa0\xc6i>\x00\x86\x94G@2\xee\xac\x07f\xcb(G@\x12\xd6\x88\xc6iF\xf0\x86\x94G@3w\xc0\xab=\xd7\xc0G@\x14 H\xd4\x10\x8dp\x86\x94G@4Ji\xf2\xeb\xef\xd8G@\x12\xc4G,\xcf\xad\x10\x86\x94G@4\xedv\x9d\x96\x9dPG@\x11J\x88K\x87\xf8p\x86\x94G@5\xa8\xbcD\xd7z\xf8G@\x10\xe5\xb9\xe5!\x90P\x86\x94G@6g\xb60\\\x9d\x01G@\x10\x1d\xdc\"\x920\xc0\x86\x94G@6\xb4Aq7\x12XG@\x12\x88>{Q`\xf0\x86\x94G@6\xd7k3\xc6o G@\x12\xd7+U\xc5o\xf0\x86\x94G@7L\x1670AXG@\x12pScl\xaf\x00\x86\x94G@8i\x18\x8f\xefk\xc8G@\x14oe\x18T\xd3\x10\x86\x94G@8\xce\x16`&\n\x88G@\x13\x96\xc7\xd0\xa6\xba\xe0\x86\x94G@9 \xfb:\x9a\x19\x98G@\x13\xb5\x7f\xa7\xb0\xf8\xe0\x86\x94G@9G_U\xe8\x9b\xf8G@\x14\xae\x7f~\xbb:\x8f\x86\x94G@9\xa6\x84>\x03\xec G@\x15\x06;\xdeN\x02\x10\x86\x94G@:g\x1bU\xe8\xa0\xe7G@\x14\x9a~\xda\xe40\x01\x86\x94G@;\x0bG\xdeq,GG@\x14\x82\xeb\xd0\xa6\xbe\xf0\x86\x94G@;_\xcdHAdXG@\x14\xef\x8f\x1f(u\xc0\x86\x94G@;\xfa\xdf\xc9\xf6O\x18G@\x11\xa27\x1f(g0\x86\x94G@<m\xd2\x89\x1b\xdd\x08G@\x11&\x0b\xdeM\xf1\x00\x86\x94G@<\xb2Yx\n\xcd G@\x11\xd1\xff\xc2\xffr@\x86\x94G@=(\xb9\\\xbcMxG@\x11\x8e\x9c\x14\xea\xf60\x86\x94G@=\xb7Kx\n\xd1\xa0G@\x12g9\\\x99\x0f\x00\x86\x94G@=\xf4\x18\x96\xc3$\x98G@\x10\xb1\xde)e\xd3\xb0\x86\x94G@>\xd5wZ0\xf5JG@\x0c\x12\xc8\x8f\xe1\x7f\xa8\x86\x94G@>\xd5w\xd7\x9d\x9c\x80G@\x0c\x12\xc5\x9ay\xa4\x80\x86\x94G@>\xc5\xfa\x0e:\x9f\xa9G@\x02\xb8\x14\xc0\x05m\xc0\x86\x94G@?,\x95\n\xd0\xd18G@\x01\xa2\xbe\xab\x8a\x87\xa0\x86\x94G@>\xdaH\x96\xc3(\x89G?\xfd\x97 \xc0\xe5-\x00\x86\x94G@>w\xf0\x1b\xe1\xdf0G?\xf9WDaRT\x00\x86\x94G@>\x16\x0e)\x89\x1eXG?\xf0\xff;\xa2,\xa3\xc0\x86\x94G@=\xe03\n\xd0\xcbpG?\xe3\x1d\xbcc\x11Z\x00\x86\x94G@=\xd1\xca\xf6U\xe9\xf8G\xbf\xcaG\x9af\x15$\x00\x86\x94G@=\x96|\x89\x1b\xe2 G\xbf\xe2\xcc\x07\x06\xbf\x82\x00\x86\x94G@=\x94W\xe5D\xd7\xd8G\xbf\xf5v\x04\xcb\x0e\x0c`\x86\x94G@=J\xb9\x18x\t\xc9G\xbf\xf9\xeb\xbf\xacU\xcd\xff\x86\x94G@=A<\xdb\x07e\xc9G\xc0\x01\xb8\x8b\x91\xe6\xb7\xb0\x86\x94G@=\x1e\x13\x18x\t\x08G\xc0\x02Vr\xd2\xc1.p\x86\x94G@=\x06a\x93YPHG\xc0\x06\xb6\xcc\xd9\x94\xe2 \x86\x94G@=F\xc1\x18x\t\xb8G\xc0\nY\xeb\xff \xe4\x80\x86\x94G@=W\n\x15\x0e9\xc8G\xc0\x11\xff\xfb\xa6\xd1a\xe8\x86\x94t\x94\x85\x94uauK\x03}\x94(h\x02\x8c\x0212\x94h\x04h\x05h\x06}\x94(h\x08\x8c\x06Africa\x94h\nG@\xb2o\x00\x00\x00\x00\x00h\x0b\x8c\x03SOM\x94h\r\x8c\x07Somalia\x94h\x0fJz\xebr\x00uh\x10(G@D}\x93\x0b\xe0\xde\xd3G\xbf\xfa\xee\x97\x8dO\xdf<G@I\x91\"\xa6\xf3\xf50G@(\x0c\x9d\x9d4X\xcet\x94h\x12]\x94}\x94(h\x04h\x15h\x16(G@D\xca\xe5\x8a2\xf4IG\xbf\xfa\xee\x97\x8dO\xdf<\x86\x94G@D\x7f\x1a\x9f\xbev\xc9G\xbf\xebw\x1c\x97\x0f{\x9e\x86\x94G@D}\x93\x0b\xe0\xde\xd3G@\x06F\xb2k\xf8v\x9e\x86\x94G@D\xeds\\\xdf\"[G@\x0fY\xee~3\x0e\xaa\x86\x94G@E\x10vJ\xdf\xf8\"G@\x10\xef\xbf\xc6T\x0c\xc8\x86\x94G@Eb\x84\x8b\xeb[-G@\x11\x02\xa6\xf3\xf5/\xc2\x86\x94G@E\xd4\x97c^t(G@\x13\xd4\x87\xfc\xb9$\x00\x86\x94G@F{W>\xab6{G@\x14\x01\xa8\xac\\\x13\xfd\x86\x94G@G\xe5\x0b\xb6\xedgwG@ \x01\x897K\xc6\xa8\x86\x94G@H>M\\r\xcc G@!\xac\xddX\xf7W(\x86\x94G@Hx\x14\xa0\xb7\x11`G@\"\xe7K\xa4\x0fA@\x86\x94G@Hx\x18\x03\xb3\xa7\x90G@#\xf2n\x99\xd1\xd50\x86\x94G@Hx {+\x1f\x08G@%\xf6\xf3\x9d;\xaeH\x86\x94G@Hx\x93\xa0\xb7\x11dG@&\xc9\xdd=\xa8\xeb\xf0\x86\x94G@Hy^\xc6\x05\xda\xffG@&\xd2<m\xa8!^\x86\x94G@Hy^\xc6\x05\xdb\x0fG@&\xd2<m\xa8!b\x86\x94G@H\xa2E\xf5\xad\x96\xa7G@&\xdcT6\xb8\xf9\xb1\x86\x94G@H\xddCk\x8f\x9b\x13G@'(e\x94\xafO\x0e\x86\x94G@I!\x1f\xb3\xfam\xf1G@'[\xf0\x99Z\xafy\x86\x94G@I]\xb2\xd4\xd4\x02KG@(\x0b6z\x0f\x90\x98\x86\x94G@I\x8e;\xcd5\xa8XG@(\x0c\x9d\x9d4X\xce\x86\x94G@I\x91\"\xa6\xf3\xf50G@'\x7f\r\x84M\x01;\x86\x94G@I\x85P\xda\xe3\xe6\xc5G@&U@\xccx\xe9\xf7\x86\x94G@I\x85\xcc\xb7\xd4\x17DG@%H$\x0bx\x03G\x86\x94G@Ij\xc6i\x05}\x17G@$\x8f7u\xb8\x13\x02\x86\x94G@IF\xb4\xb7,Q\x98G@\"e\xc1?\xd0\xd0h\x86\x94G@I\t\x13\xe8\x14P\xf0G@ )\xd8\x83\xba4D\x86\x94G@H\xb9\xf2\x12\xd7s\x1aG@\x1b7\xf8\xca\x81\x98\xf2\x86\x94G@HL\x1a6\xe2\xeb\x1dG@\x15[?\xa6\xde\xfc{\x86\x94G@G\xde\xd24\xeb\x9a\x18G@\x10\xe0\xaad\xc2\xf88\x86\x94G@GHJ\x0eA\x0bcG@\x06\xd7\xa2H\x94\xc4H\x86\x94G@F\xc80\xd3\x06\xa2\xb2G@\x00]\xb7k;\xb8=\x86\x94G@F\x08\xb9#\xa2\x9cwG?\xf0\xd8dE#\xf6\x80\x86\x94G@E\x91gw\x07\x9eZG?\xd2\xb3g\xa0\xf9\tl\x86\x94G@E\x05R*o?TG\xbf\xedi\xc2;yR\xd3\x86\x94G@D\xe7\xcd5\xa8XyG\xbf\xf7$\xbd\xba\nRj\x86\x94G@D\xca\xe5\x8a2\xf4IG\xbf\xfa\xee\x97\x8dO\xdf<\x86\x94t\x94\x85\x94uauK\x04}\x94(h\x02\x8c\x0213\x94h\x04h\x05h\x06}\x94(h\x08\x8c\x06Africa\x94h\nGA\x02\xa3\xe0\x00\x00\x00\x00h\x0b\x8c\x03KEN\x94h\r\x8c\x05Kenya\x94h\x0fJ\xfb\x8e\xd6\x02uh\x10(G@@\xf2`w\xce\xb8PG\xc0\x12\xb5\x031\xe3\xa7\xdbG@D\xeds\\\xdf\"[G@\x16\x06$\xdd/\x1a\xa0t\x94h\x12]\x94}\x94(h\x04h\x15h\x16(G@C\x99\xe2XOLnG\xc0\x12\xb5\x031\xe3\xa7\xdb\x86\x94G@B\xe2)\xc7y\xa6\xb5G\xc0\rj\xbd\xe3\xfb\xbd|\x86\x94G@B\xd9n\xac\x86\x05`G\xc0\x08\xc6\xa2\xb1pO\x80\x86\x94G@A\tK\x9c\xb6\x84\x88G\xbf\xf0\xf5\x05\xd0\xfaX\x00\x86\x94G@@\xf3\xac\xce\xfaK\xe6G\xbf\xeefffffg\x86\x94G@@\xf2`w\xce\xb8PG?\xbc\x1c\xbdq\xc6`\x00\x86\x94G@A\x17\n=p\xa3\xd7G?\xe0z\xe1G\xae\x14{\x86\x94G@AV\x07_o\xd2 G?\xf2\xd4\xbf\t\x95\xaa\xf7\x86\x94G@A\x84\x9bR\x00}\xd4G?\xfe~R\x15v\x89\xca\x86\x94G@ALL\x05\x92\x108G@\x08n\x0f<\xb3\xe6\x80\x86\x94G@A=T!\xc0D(G@\x0cq\xdei\xadC\x80\x86\x94G@A\x00\xa3\xd7\n=qG@\x10\xff\xe1\xd6\xf2\xbb\x10\x86\x94G@AOb\x97_\xc4OG@\x13ct$\xe3\xc2\x8b\x86\x94G@A\xa6%\x18\xe5d\x97G@\x16\x06$\xdd/\x1a\xa0\x86\x94G@A\xe8\xa2 \x00\x00\x00G@\x15ZY\x82\xd8-\x81\x86\x94G@A\xe8\xa2 \x00\x00\x00G@\x13\x1b\x9c\xe3\n\xbc+\x86\x94G@B\x14\\\xb0J\xcd\xbfG@\x11\xca\x9c\xe4\xd5\x9f\xdc\x86\x94G@Bms\xb1\xfa#HG@\x11\xca\x9c\xe4\xd5\x9f\xdc\x86\x94G@C\x0fz$\x89LDG@\x0c\xc9\xf1k\x11\xc6\xd2\x86\x94G@C7\xee\xa2\t\xaa\xa4G@\x0c\xb5D\xbb\x1a\xf3\xa1\x86\x94G@CU\xe7\xea_\x84\xcbG@\x0c\xed\xb6\x1b\xb0_\xb0\x86\x94G@Cr=\xc4\x86\xad.G@\x0c\x01\x83\xf9\x1edp\x86\x94G@C\xc7\x99\xe7D\xa5\xf5G@\x0b``\xfeG\x99\x1c\x86\x94G@C\xedn\xac\x86\x05hG@\x0e\xb5\xd7\x88\x11\xb1\xd9\x86\x94G@Db]\x8dy\xd0\xa6G@\x11\x070@9\xab\xf4\x86\x94G@D\x95\xfd\x8a\xda\xb9\xf5G@\x0fZK\xdb\xa0\xa5&\x86\x94G@D\xeds\\\xdf\"[G@\x0fY\xee~3\x0e\xaa\x86\x94G@D}\x93\x0b\xe0\xde\xd3G@\x06F\xb2k\xf8v\x9e\x86\x94G@D\x7f\x1a\x9f\xbev\xc9G\xbf\xebw\x1c\x97\x0f{\x9e\x86\x94G@D\xca\xe5\x8a2\xf4IG\xbf\xfa\xee\x97\x8dO\xdf<\x86\x94G@Dq@$\xb3=\xb0G\xc0\x00\xa9\x0f\xf9rGE\x86\x94G@DQ\xa5\x11\x9c\xe0wG\xc0\x03\xff\x91\xe6F\xf1V\x86\x94G@D!\xabKr\xc5\x1bG\xc0\x04\x95\xb05\xbdQ/\x86\x94G@D\x0f\x83'gM\x16G\xc0\n8\xb0J\xb6\x06\xb8\x86\x94G@C\xe6h]\xb7k<G\xc0\rs\x04\x03\x9a\xbf4\x86\x94G@C\xcdm\t\x17\xd6\xb7G\xc0\x11b\xd8\xc2\xa4T\xdf\x86\x94G@C\x99\xe2XOLnG\xc0\x12\xb5\x031\xe3\xa7\xdb\x86\x94t\x94\x85\x94uauK\x05}\x94(h\x02\x8c\x0214\x94h\x04h\x05h\x06}\x94(h\x08\x8c\x06Africa\x94h\nGA\x05\x85`\x00\x00\x00\x00h\x0b\x8c\x03SDN\x94h\r\x8c\x05Sudan\x94h\x0fJ\x8f\xda9\x02uh\x10(G@5\xef\xd2\xc7\xb8\x90\xd0G@ uX\x1e\xf0~(G@C4}\xd3\xea-\x80G@6\x00\x00\x00\x00\x00\x00t\x94h\x12]\x94}\x94(h\x04h\x15h\x16(G@8\x91?\x18w\xf5\x00G@ uX\x1e\xf0~(\x86\x94G@7\xceI\xc9\xf6<\xc0G@!U'\xbf]\xbc@\x86\x94G@7u\x81\xdeq\x1c\x80G@!\xe8\x98%\xc4$\xf8\x86\x94G@7e\x10>\x03\xe20G@\"\x87\xb6\xf9d\x94\xf0\x86\x94G@7\x8e\xa7\xec\x18]\xc7G@#\\\xc8\xa0\xa5s\x10\x86\x94G@7\x8d\xe6\xe1\xda\xed G@$-\xb2\xde\x16\x1ay\x86\x94G@6\xfa@K\xab!8G@%m\xce\nu\xb2\xc0\x86\x94G@6\xdd9\xf2\xeb\xfb(G@&H\xe8\x03\xa2\x16\x18\x86\x94G@6\xe0O\xf44\x19\xd0G@&\xc4\xeb\x9a\x17n \x86\x94G@6\x829\x82\x01\xcd`G@'[\xd5\x12\xeck\xe0\x86\x94G@6\x7fd\x069\xd5\xe0G@(\x85>-b9\x00\x86\x94G@6I\xbb\x05\xfa\xeb\xd0G@)J\xc7\x10\xcb)@\x86\x94G@5\xef\xd2\xc7\xb8\x90\xd0G@)-%\xed\xd0R\xe0\x86\x94G@6\t\x9f\x7f\x8c\xa8\x10G@)\xe92\r\x99E\xa0\x86\x94G@6K\xec\xaa\xb8\xa5\xd0G@*\xbe\xa0\xba\x1fK\x00\x86\x94G@6.\xec\x17\xeb\xaf\x10G@+\x92\xad\x81\xad\xea\xa0\x86\x94G@6\x83\x13\xbe\"\xe5\xe0G@,/\xb5I\xf9H@\x86\x94G@6M\xb2\xd4\xd4\x02@G@,\xa7T\xf3w[\xa0\x86\x94G@6\x91e+\xd3\xc3\x80G@-\xe3y\xfa\x97\xe1\x80\x86\x94G@7\x06K\x87\xbd\xcf G@/\\\x87T\xf3w\x80\x86\x94G@7\xe3\x0b9\x19&`G@/8\xc0\x05>-`\x86\x94G@7\xd6p\xe2\xc1+\x00G@3\x94\x99\xae\x92O@\x86\x94G@7\xd9\x99\x99\x99\x99\xc0G@4\x00\x00\x00\x00\x00\x10\x86\x94G@9\x00\x00\x00\x00\x00 G@4\x00\xc7:\xbc\x94\x80\x86\x94G@9\x00\x00\x00\x00\x00\x00G@6\x00\x00\x00\x00\x00\x00\x86\x94G@=\x05\x1e\xb8Q\xeb\x80G@6\x00\x00\x00\x00\x00\x00\x86\x94G@@s33330G@6\x00\x00\x00\x00\x00\x00\x86\x94G@Bn\xe0\x9f\xe8h0G@6\x00\x00\x00\x00\x00\x00\x86\x94G@B\x98'\xfa\x1a\r\x00G@5\x04\xd3Z\x85\x87\xa0\x86\x94G@B|\x15\xa0{5 G@4\xd6bw\xc4\\\xe0\x86\x94G@B\x8e\xae}Vm\x08G@3\xce\xd6wpz\x00\x86\x94G@B\xbd\xabKr\xc5(G@2\x9d5\x00\x92\xcd\x10\x86\x94G@B\xeen\xebp&\x10G@2^,\x12\xad\x81\xa0\x86\x94G@C4}\xd3\xea-\x80G@1\xff\x91\x12\xe2:0\x86\x94G@B\xf3\xb6E\xa1\xca\xd0G@1msB\xed\xbb`\x86\x94G@B\x95o\xa8.\x87\xe0G@1C]$\x9eE \x86\x94G@Bm\x1f\xb3\xfan\x00G@0\xf4\xe0u\xf6\xfd \x86\x94G@B`\x7fw\xafd\x10G@0J\xb7VC\x02\xc0\x86\x94G@B)_E\xe0\xb4\xe0G@-\xa5\x1dh\xc6\x93\x00\x86\x94G@B6\xfa/\x05\xa7\x10G@,\xd8\x1e\xcdJ\xa1 \x86\x94G@B\"\x96\x91\xa7\\\xd8G@+ l\xca-\xb6`\x86\x94G@A\xee\x8bm\x86\xec\x18G@)(\x14P\xef\xdc\xc0\x86\x94G@A\xa1W\xbc\x7fw\xb8G@(*l\xa0<K \x86\x94G@Ajr\xda\x12/\xc0G@&\xa3N\xb9\xa1w \x86\x94G@A]\x96R\xbd<HG@%\xd2\x01\xcd_\x9a\x00\x86\x94G@A \xf4\x1f!-\x80G@%B\x9b(\x0f\x12\xc0\x86\x94G@@\xfb\x16]9\x97\x08G@#*\xca\xffm3@\x86\x94G@@\xfc\xcc%\x07 \x90G@!^~\xa5\xf8L\x80\x86\x94G@@\xfbPt\x81\xcd8G@\"\xed\xb6\xcb(\xc1@\x86\x94G@@\xe9\x98gAr0G@\"\xf7\xd6\xd2\xbb>@\x86\x94G@@\xeb\xca\xf1\x9cZ\xb0G@#\xf6\xbd\x83\xec\x9c\xe0\x86\x94G@@\xdci)\x1f\xfe\xe0G@$\xa6\x88\xc0UH\x00\x86\x94G@@\x9a|\xf2x\xbe\xb0G@%p\xb2y\x00\xd9\xc0\x86\x94G@@\x8b\x1b)\xfcb\xe0G@&\xe1\xddD\xa9t\x80\x86\x94G@@\x9a|\xf2x\xbe\xb0G@([\xd29\xbd\xb1`\x86\x94G@@_(Z\xe28\x18G@(~\xfa\xdfl: \x86\x94G@@V^1v\x95\xe8G@(\x0c\xb6\xc4\xf4\xfe\x00\x86\x94G@@\tuG\x08\xcb\x08G@'\xf2XH\xb2\x17`\x86\x94G@@(8\xd8\x01\x82\x98G@'\\\xeb\x88\x8cR\xe0\x86\x94G@@35\x8b\xc8\rPG@&)G\xde\xd5'\x00\x86\x94G@?\xd9\xc8\x80\xd5\xf8\x00G@%\x10\x02\xb1`\xe2 \x86\x94G@?ZU(=H\xb0G@#\x9e\xd7\xe5\xb8F\xe0\x86\x94G@>\xd6|\xba\xee\xc8 G@#j\x1a\xed2z \x86\x94G@=\xff#\xc4!\xc3@G@$\x94\xf4m~\x03\x80\x86\x94G@=\x9es\xfc\x81\xcbpG@$+z|ri\xc0\x86\x94G@=\x84\x15\x80>\xe5\x00G@#\x96\r\xbcL\xa4\xe0\x86\x94G@=\x00=\x12\xf0d\x80G@#5]\xf4\xac\xad@\x86\x94G@<\xf7r\xe9\x84\xc2`G@\"\xcb\xe4\x03\xa1\x13 \x86\x94G@;\xf8\x8c8Sc\x90G@\"\xcb\xe4\x03\xa1\x13 \x86\x94G@;\xd5c\x92\xa4\xdb\x10G@#5]\xf4\xac\xad@\x86\x94G@;\x1c\xce,\xd0\x8d\xb0G@#F\xf2G\x83\xf1`\x86\x94G@:\xc0\x83y\xe6f\xe0G@\"\xef\x0c\xa9O\x9b\xc0\x86\x94G@:z2.\x89U\xb0G@#\x1a\xffxi\xc6\xa0\x86\x94G@9\xf6Y\xc1:\xd5 G@$E\xd8\xf8\xb5P`\x86\x94G@9\xcaf\xf2 \xaa`G@$\xd2{\x8for\xc0\x86\x94G@9\x11\xd1\x8cL\\\xf0G@$\x8c*D\x12a`\x86\x94G@8\xcb\x80@\xefK\xc0G@#\x9e\xd7\xe5\xb8F\xe0\x86\x94G@8\x89\x94\nH\x0b\x90G@!\xd5\xc7{\xdbV\xa0\x86\x94G@81\xael\x13\xb5\xd0G@!u\x17\xb4;^\xe0\x86\x94G@7\xe3\x11\x18\x04\x040G@!=M6\x82\xd0`\x86\x94G@8\x91?\x18w\xf5\x00G@ uX\x1e\xf0~(\x86\x94t\x94\x85\x94uauK\x06}\x94(h\x02\x8c\x0215\x94h\x04h\x05h\x06}\x94(h\x08\x8c\x06Africa\x94h\nG@\xdd\xdf\x80\x00\x00\x00\x00h\x0b\x8c\x03TCD\x94h\r\x8c\x04Chad\x94h\x0fJ\xd1C\xb8\x00uh\x10(G@+\x14\xaeu3\xcd\x80G@\x1d\xb0\x0c\xfd\x06z\x10G@7\xe3\x0b9\x19&`G@7h\xe3h\xf0\x84`t\x94h\x12]\x94}\x94(h\x04h\x15h\x16(G@7\xd6p\xe2\xc1+\x00G@3\x94\x99\xae\x92O@\x86\x94G@7\xe3\x0b9\x19&`G@/8\xc0\x05>-`\x86\x94G@7\x06K\x87\xbd\xcf G@/\\\x87T\xf3w\x80\x86\x94G@6\x91e+\xd3\xc3\x80G@-\xe3y\xfa\x97\xe1\x80\x86\x94G@6M\xb2\xd4\xd4\x02@G@,\xa7T\xf3w[\xa0\x86\x94G@6\x83\x13\xbe\"\xe5\xe0G@,/\xb5I\xf9H@\x86\x94G@6.\xec\x17\xeb\xaf\x10G@+\x92\xad\x81\xad\xea\xa0\x86\x94G@6K\xec\xaa\xb8\xa5\xd0G@*\xbe\xa0\xba\x1fK\x00\x86\x94G@6\t\x9f\x7f\x8c\xa8\x10G@)\xe92\r\x99E\xa0\x86\x94G@5\xef\xd2\xc7\xb8\x90\xd0G@)-%\xed\xd0R\xe0\x86\x94G@6I\xbb\x05\xfa\xeb\xd0G@)J\xc7\x10\xcb)@\x86\x94G@6\x7fd\x069\xd5\xe0G@(\x85>-b9\x00\x86\x94G@6\x829\x82\x01\xcd`G@'[\xd5\x12\xeck\xe0\x86\x94G@6\xe0O\xf44\x19\xd0G@&\xc4\xeb\x9a\x17n \x86\x94G@6\xdd9\xf2\xeb\xfb(G@&H\xe8\x03\xa2\x16\x18\x86\x94G@6;+HAM\xb0G@%\xf1\x9bf\x9e\xab\x00\x86\x94G@5\xb9L`%\xfc\xf8G@%\"U\x1b\x86\xc2\x88\x86\x94G@5\x008\xe8\xae\x82XG@\"\xf3\xb4U\x8d\x8c\x90\x86\x94G@4\x0fG\x8c\x85\x88pG@\"\x06\x81f\x9e\x99\x88\x86\x94G@3\x18\x10\xe8\xaey\xf0G@\"&RU\x8d\x89\x08\x86\x94G@2\xcf\xdf\xdeq\x08\x10G@!\xf7@\x92\xfe,\x10\x86\x94G@2\xe98\xb8\xe5\x16(G@!C\x04\xa0\xa5i\xd0\x86\x94G@2c\xb9\xdeq\x068G@ \x90\x07\x07\x0b\xcd \x86\x94G@1\xf7\x05\xa1\x00`\x80G@\x1f\x90K\xc2\xff\xaf@\x86\x94G@0\xb4\xbb\xa7\xd3\xfb`G@\x1e\x08\x87\x03\xda\x1c\x00\x86\x94G@0t\xc8\x82H\x07\xf0G@\x1e\xf0h\x82%7\xe0\x86\x94G@0Jb>\x03\xc2\xf0G@\x1f\x04i%\xfcC\x00\x86\x94G@0\x1b2\x00\x93\x1e@G@\x1d\xfd\x04\x9ds\xb5@\x86\x94G@.\x8f\x15q\xca\x0c`G@\x1d\xb0\x0c\xfd\x06z\x10\x86\x94G@.\xdfGj\xf6m@G@\x1e\xc5p\x9ds\xb8\x80\x86\x94G@.=\xe2\x15\xa1\x15 G@ \xc3\xa90\x01\x90\xa0\x86\x94G@-\xf5\xc1\xfaR\x92?G@!\x97\x9a\xf9d\x90\xd0\x86\x94G@-\x16\xc4S\x11\xb3\xf0G@!\xee\x85_\xca\xf8\xb0\x86\x94G@+\xe8\x8fO\xa7\xde\x90G@#\x19Wf\x9e\x9e@\x86\x94G@,W\xcag\x8c\x91\xf0G@$\n\xf2\x18\x1c\xed\x80\x86\x94G@-A n`6`G@#\xd7\x82\xbb\xf3\xf6\xe0\x86\x94G@-\xd1\x96\xd4\xc6\x9fQG@#\xfb\xf8c4\xd1\xf0\x86\x94G@.\xef\x8d\x04\x90\x07 G@#\xf6\xf4\xd7By`\x86\x94G@-\xd8\xdd\x7fqJ\x10G@%\xc8[\xc61p\x08\x86\x94G@-\xeb\x99\x04\x90\x02\xb0G@'\x1ct3k|\xb9\x86\x94G@-\xc9f{_\x1c\x00G@(p'RT`\xa0\x86\x94G@,\xfd\xd7\xd80k\xe0G@)\xb8\x02\xc9\x9bQ\x91\x86\x94G@-1\n>\x96\xd3 G@*\xa9-\xb8\x8aD\xe0\x86\x94G@+\xe8\xb1-\x85\xbc\xc0G@*\xb4\xf7=\xa8\xfd \x86\x94G@+\xe9\xd4n`0\xc0G@+\xfeNN\xba\x14\x00\x86\x94G@+\x14\xaeu3\xcd\x80G@,\xbb\xf8\xf2\x91!\x80\x86\x94G@+\xf1\xc0D(N\x00G@/^e\xbe\xa0\xba@\x86\x94G@.~\xd6\x9e)\x9e\xc0G@0\xa0\x97\x1d\x1f\xa4\xd0\x86\x94G@.\x99\xd3j\xf6l`G@1\xed\x8e \x89z\xd0\x86\x94G@/_\x19d\"\xcf\x00G@3\xf5\t\xc0\xf6\xbd\xe0\x86\x94G@/\xcevY\xe5``G@4c:\xfegc\x90\x86\x94G@.\xf9kx\x9d\xae\xa0G@4\xba\xfcru\x0cP\x86\x94G@.\xf1.\xc6\xbc\xe8\x80G@5\x0cg8\x1d}\xd0\x86\x94G@.1\x9bB\x00\xa7\xa0G@5N\xfb\x16L\x190\x86\x94G@-\xb3\xdd\x97\xf6+\x80G@6\xdc\xeaJ\x8c\x15p\x86\x94G@/\xb8\xc1T\xc9\x86 G@7h\xe3h\xf0\x84`\x86\x94G@3\xd9i\x1au\xcd G@5~\xbe7\xde\x93\xb0\x86\x94G@7\xd6p\xe2\xc1+\x00G@3\x94\x99\xae\x92O@\x86\x94t\x94\x85\x94uauK\x07}\x94(h\x02\x8c\x0225\x94h\x04h\x05h\x06}\x94(h\x08\x8c\x06Africa\x94h\nGA&\x8e8\x00\x00\x00\x00h\x0b\x8c\x03ZAF\x94h\r\x8c\x0cSouth Africa\x94h\x0fJ\xd0\xd0D\x03uh\x10(G@0XPf\xf9\x85\xc0G\xc0Ah\xdaqp\x9e\xb4G@@jAcFJ\x1cG\xc06\x17`E\xdd\x9b\x9ct\x94h\x12]\x94}\x94(h\x04h\x15h\x16(G@0XPf\xf9\x85\xc0G\xc0<\x93\xa2\xf0\x88b\xce\x86\x94G@0\xd2\xf2\xcd_\xee@G\xc0<\x15\x08\x8a!\xfa:\x86\x94G@18\x0b\xb5{>\x80G\xc0<[\x1f\x19~$c\x86\x94G@1c3\x03\xfc\xf4(G\xc0<\xc8\x94a,:\xc2\x86\x94G@1\xd6\x0e\x0e:f\xc0G\xc0<\xdb;\x94_nF\x86\x94G@2w\x03\xa1\x00b\xb0G\xc0=\x0b\xa3d\x96\x0c\x1d\x86\x94G@3\x00\x8bjca\xa8G\xc0<\xf8\xf2\x08m\x16\n\x86\x94G@3\xe5\rO\x14\xe3\xe0G\xc0<v\n\xf7\\\x02\xb8\x86\x94G@3\xe5Q\n\xd0\x9f\x98G\xc08\xc4\x8d\xe6J\xe1j\x86\x94G@4*l\xfd)_\xf0G\xc08\xea\xff\x8d\x8b\xbc\x86\x86\x94G@4\xc24706(G\xc09\xde>1b\xca\xf2\x86\x94G@4\xaa\x9d\xc9\xf6.\xf0G\xc0:z:a,0\x9b\x86\x94G@4\xe3\xbdjci\xefG\xc0:\xd4\x1bd\x96\x02^\x86\x94G@5\x9b\x1c\x00\x936xG\xc0:\xb9\xfe\x1c\xe7\xedp\x86\x94G@6\x1b \xc6\x8ce\x0fG\xc0:G\xbe\xdc\rwl\x86\x94G@6\x94\\0\\\x9d\xc0G\xc09\xfa\xbd\x12\xaay\x84\x86\x94G@6\xd3\x03q7\x12\xe0G\xc09\x80\x1e\x0f@\xa72\x86\x94G@7O\xe5\x93Y7(G\xc09D\xc8\xdc\rr\xfa\x86\x94G@7\xbb\xcb:\x9a\x13xG\xc09c\xdf\x86\xb8\x1e,\x86\x94G@86\x15\x93Y; G\xc09\xab\x93Bs\xdb$\x86\x94G@9\x06q\x93Y>\xb0G\xc09\xb8<L\xb1L\x00\x86\x94G@9\xaa'\x93YA\x80G\xc09|\x9f\xfa\xc5\xc5\xda\x86\x94G@9\xc4\x0e\xab=\xf3qG\xc09,\xc2\xacD\x0f\x94\x86\x94G@9\xf1\x10\x1b\xe1\xcb@G\xc08\xb2E\x86\xb8\x1b \x86\x94G@:|ZR~\xd1\x10G\xc08\x9d\xc7\x94_[\x9f\x86\x94G@:\xc9Q\xf2\xec\x0chG\xc08=\x9d\xe6J\xdf\x19\x86\x94G@;\x1e\x91\xa1\x00\x88\xc0G\xc07\x93\x06\xd59\xcb\x19\x86\x94G@<\x04i\x93YK\xd8G\xc06\xd3\xe7\xa8\xda5\n\x86\x94G@=n\xa3\xe5D\xd71G\xc06\x17`E\xdd\x9b\x9c\x86\x94G@=\xd6\xcb\x1fK\xac\x9fG\xc06\x1a*\xdc\re\x0c\x86\x94G@>R\xa8{t\xa4\x80G\xc06E\x88ZX}\xaf\x86\x94G@>\xa8\xec\xef\x82M?G\xc06&\xcd Q\xa9\x88\x86\x94G@?1\x000\\\xc3\xa0G\xc06@b\xf0\x88F\xfc\x86\x94G@?\xab\x9f3\xc6\x95\xf0G\xc07\xa8\xb21b\xc1;\x86\x94G@?\xee;\x11\xa4t\xf7G\xc08^\x92\x16\x14B\xa6\x86\x94G@?\xc0\x9d\xd7\x9d\xa0\x90G\xc09{\xfa\x08m\x06\xb2\x86\x94G@?\xd6x\x9d\x96\xcdQG\xc09\xd7\xe4\x97\xc91<\x86\x94G@?UI\xd0\xc9\xfeHG\xc09\xa9\x02?\n\n\xe2\x86\x94G@?\x0bH\xcd`,\xc8G\xc09\xbb@u\xa7\x0e\x9c\x86\x94G@>\xf3\x1d\\\xbcUXG\xc0:\x05\xccS\x84\xed\xc2\x86\x94G@>\xad670a\xc8G\xc0:e\xe8u\xa7\x11\x8a\x86\x94G@>\xaf\x9b3\xc6\x91\x9fG\xc0:\xbel\xa5pv\x0c\x86\x94G@?Hc\xd0\xc9\xfe\x10G\xc0;I/d\x96\x04b\x86\x94G@?\xde93\xc6\x96\xd1G\xc0;-\x8c\xa5pw\xf7\x86\x94G@@\t,U\x9f\x07\x98G\xc0:\xbb\xdb\xa2\x06\xa5\xca\x86\x94G@@jAcFJ\x1cG\xc0:\xbe\x00E\xdd\xb0\x12\x86\x94G@@JF\x1f\x02\x05LG\xc0;x\\?\n\x12\xd8\x86\x94G@@;')?u\xacG\xc0<M\x0f\x12\xaa\x83\xba\x86\x94G@@\x1a\x08\xa4 \xbc\xc8G\xc0<\xc0\x9d\x9b3\x0e>\x86\x94G@?\x85`YR\x87\xa8G\xc0=A\xe4\x1c\xe7\xf8\x92\x86\x94G@?SW\xf9\xbf\xc0\xd0G\xc0=f\xe8\x01\x99w\x7f\x86\x94G@>\xe6\xd9\xec\x18~\x18G\xc0=\xe8\xf2\xf0\x88h\xac\x86\x94G@>\x9fp\xb2\x11\xa9?G\xc0>l|\x90\xf5\xa4\xf4\x86\x94G@>\x0eCjc\x92@G\xc0?#\xe8\xb3\x17\xca>\x86\x94G@<\xec\xf1\x03\xfd&\xe9G\xc0@\x16\x05qp\x98\xe2\x86\x94G@<8A\xec\x18rGG\xc0@b\xcfY\x8b\xe8\xb5\x86\x94G@;v\xf0\x8f\xefy8G\xc0@\x9d\r&X\xb6\x82\x86\x94G@:ka:\x9a\x1fGG\xc0@\xce\xb6\xb2K\x10\x1c\x86\x94G@9\xe8\xdf\xc3\"\xa5\x90G\xc0@\xd5a\x93\x92\xbeN\x86\x94G@9\xc7\xd7Am\xbc\xe8G\xc0@\xf8\xea)\xc2\x88M\x86\x94G@9,@\xab=\xf0\xd8G\xc0@\xe5\xff:\xd3\x99\x0b\x86\x94G@8\xad\x87\xc9\xf6@\x98G\xc0@\xfe[\xc6\xc5\xf25\x86\x94G@7\x98\x13:\x9a\x12\xd8G\xc0@\xe5\xb1V\"\x1a\xbe\x86\x94G@6\xfc\xf9\xf2\xeb\xfb\xb8G\xc0@\xf5M\x9af_H\x86\x94G@6\x92\xfb\xf9\xbf\x9aPG\xc0@\xee\x9aA\xa79\x9e\x86\x94G@5\x8a\xf4\xe1\xda\xe4HG\xc0A!!\xa1:\x00v\x86\x94G@4\xb0e\xc3\"\x8e\xa0G\xc0A5f\x00\xcc\xc6\xc8\x86\x94G@4\x12>)\x88\xf2HG\xc0Ae\xc7\x0b\n8@\x86\x94G@3\x9d\xcc\xb8\xe5\x19@G\xc0Ah\xdaqp\x9e\xb4\x86\x94G@31z\xb2\x11v\xf8G\xc0A;6qp\x9d\xec\x86\x94G@2\xda\xf5\xe5D\xa8\xb0G\xc0A8\xdf\x00\xcc\xc6\xd7\x86\x94G@2l\xb5jc_\x18G\xc0@\xff\xbaK\xe4\xaa\x8d\x86\x94G@2`\x9e\x00\x93(HG\xc0A\x11y\x82\x81\xaeE\x86\x94G@2>\x97~\xde?\x98G\xc0@\xef\x12{\xae\r@\x86\x94G@2@\x05Am\x9b\xc8G\xc0@\xa4\x05\xecQ\xe3\x01\x86\x94G@1\xec\xd9HA:\xc8G\xc0@N>\xc6\xc5\xef/\x86\x94G@2?w\x03\xfc\xf7\xf0G\xc0@6\xed\xc6\xc5\xee\xc8\x86\x94G@28\xc5\\\xbc\x1dXG\xc0?\xa9`\xc7\x92\xad\xd1\x86\x94G@1\x91!\x85\xb1\xdd\x08G\xc0>\xb9\xc8\xdc\r\x8a\xfa\x86\x94G@1\x10}\x93Y\x1b\xb0G\xc0=\xe0\xee\x9e\x9c\xe3j\x86\x94G@1\x10\x1b\\\xbc\x18@G\xc0=\xe0>\x83Na\xb2\x86\x94G@0XPf\xf9\x85\xc0G\xc0<\x93\xa2\xf0\x88b\xce\x86\x94t\x94(G@<\xfaojc\x8d\x89G\xc0<\xf4\xa1\xfa\xc5\xd5\x1b\x86\x94G@<\x8a\xac\xdb\x07b\xa8G\xc0<\xa5\xc2\xacD\x1e\xda\x86\x94G@<\x13\x07\xd7\x9d\x90`G\xc0<\xd9\xf9\xd8\xa3\xb2\x84\x86\x94G@;\x88R\xa4jZ\xc7G\xc0=>\"L\xb1[}\x86\x94G@:\xff\xcf\xa1\x00\x888G\xc0=\xe0>\x83Na\xb2\x86\x94G@;\xbf\xd8{t\x990G\xc0>\xa5%\xa8\xdaWl\x86\x94G@<\x1bq\xc3\"\xaf@G\xc0>\x8b\xb5\x19~.\x06\x86\x94G@<J\x83\x85\xb2\x0c8G\xc0>9\xe9V\xee\xd0v\x86\x94G@<\xd90\xb8\xe5A\xd8G\xc0>\x11\xee\xd59\xe7\xaa\x86\x94G@=\x04\xb6\xdb\x07d\xc0G\xc0=\xbegki\xaf\x9e\x86\x94G@=S>\x1b\xe1\xda(G\xc0=A\xe4\x1c\xe7\xf8\x92\x86\x94G@<\xfaojc\x8d\x89G\xc0<\xf4\xa1\xfa\xc5\xd5\x1b\x86\x94t\x94\x86\x94uauK\x08}\x94(h\x02\x8c\x0226\x94h\x04h\x05h\x06}\x94(h\x08\x8c\x06Africa\x94h\nG@\xb7\x83\x00\x00\x00\x00\x00h\x0b\x8c\x03LSO\x94h\r\x8c\x07Lesotho\x94h\x0fJ\x9a\xe0\x1d\x00uh\x10(G@:\xff\xcf\xa1\x00\x888G\xc0>\xa5%\xa8\xdaWlG@=S>\x1b\xe1\xda(G\xc0<\xa5\xc2\xacD\x1e\xdat\x94h\x12]\x94}\x94(h\x04h\x15h\x16(G@<\xfaojc\x8d\x89G\xc0<\xf4\xa1\xfa\xc5\xd5\x1b\x86\x94G@=S>\x1b\xe1\xda(G\xc0=A\xe4\x1c\xe7\xf8\x92\x86\x94G@=\x04\xb6\xdb\x07d\xc0G\xc0=\xbegki\xaf\x9e\x86\x94G@<\xd90\xb8\xe5A\xd8G\xc0>\x11\xee\xd59\xe7\xaa\x86\x94G@<J\x83\x85\xb2\x0c8G\xc0>9\xe9V\xee\xd0v\x86\x94G@<\x1bq\xc3\"\xaf@G\xc0>\x8b\xb5\x19~.\x06\x86\x94G@;\xbf\xd8{t\x990G\xc0>\xa5%\xa8\xdaWl\x86\x94G@:\xff\xcf\xa1\x00\x888G\xc0=\xe0>\x83Na\xb2\x86\x94G@;\x88R\xa4jZ\xc7G\xc0=>\"L\xb1[}\x86\x94G@<\x13\x07\xd7\x9d\x90`G\xc0<\xd9\xf9\xd8\xa3\xb2\x84\x86\x94G@<\x8a\xac\xdb\x07b\xa8G\xc0<\xa5\xc2\xacD\x1e\xda\x86\x94G@<\xfaojc\x8d\x89G\xc0<\xf4\xa1\xfa\xc5\xd5\x1b\x86\x94t\x94\x85\x94uauK\t}\x94(h\x02\x8c\x0248\x94h\x04h\x05h\x06}\x94(h\x08\x8c\x06Africa\x94h\nG@\xdb\xaa\x80\x00\x00\x00\x00h\x0b\x8c\x03ZWE\x94h\r\x8c\x08Zimbabwe\x94h\x0fJ\x1c\xa6\xd2\x00uh\x10(G@9C\xa4K\xab+HG\xc06E\x88ZX}\xafG@@l\xc8=\xbaW\xd0G\xc0/\x03\xfc\xa7\t\x7f\x05t\x94h\x12]\x94}\x94(h\x04h\x15h\x16(G@?1\x000\\\xc3\xa0G\xc06@b\xf0\x88F\xfc\x86\x94G@>\xa8\xec\xef\x82M?G\xc06&\xcd Q\xa9\x88\x86\x94G@>R\xa8{t\xa4\x80G\xc06E\x88ZX}\xaf\x86\x94G@=\xd6\xcb\x1fK\xac\x9fG\xc06\x1a*\xdc\re\x0c\x86\x94G@=n\xa3\xe5D\xd71G\xc06\x17`E\xdd\x9b\x9c\x86\x94G@<\xcbn\x96\xc3\x1f\x80G\xc05\xa3\xb3Bs\xc9j\x86\x94G@<\x05x\x82H:\xd0G\xc05|h\xdc\rbV\x86\x94G@;\xba+\x9a,\xea\xffG\xc04\xda\x0f\xaf\xad\xcc\xc6\x86\x94G@;\xb9\x89\n\xd0\xc2\x08G\xc04\x7f\xc2L\xb15\n\x86\x94G@;K\xe7\xbcO\x0b@G\xc04d:\xa5pZ\x1c\x86\x94G@:*/\xbcO\x06GG\xc03K\x07\xa8\xda%~\x86\x94G@9\xd9\xb3Am\xbd8G\xc02\xb6\xe3\xc4(\xa4\xa6\x86\x94G@9\xa6/\x93YApG\xc02\x898\xfe/w~\x86\x94G@9C\xa4K\xab+HG\xc01\xbc\x8d\xdfw\"\x0e\x86\x94G@:a\xc6\x82H3\x98G\xc01\xd8\x9686H\x16\x86\x94G@:\xb4\xef\x18w\xfehG\xc01\xf6\x13\x19}\xf6\xac\x86\x94G@;\x0b_\x93YG\x90G\xc01\xf0\"|z\x8c\xc2\x86\x94G@;\x99&{t\x98\x87G\xc01Js\xdfw \x18\x86\x94G@<w\xc8\xb2\x11\x9f\xc0G\xc00w\xe9\x12\xaaO\xae\x86\x94G@<\xd3l\"\xb5x`G\xc00c\xc6\x90\xf5g:\x86\x94G@<\xf2\x8c\xf6U\xe6 G\xc00\x0b\x05kis\\\x86\x94G@=\x84OAm\xcdXG\xc0/J\x136e\xa90\x86\x94G@>F5\xa1\x00\x96\x9fG\xc0/\x03\xfc\xa7\t\x7f\x05\x86\x94G@>V\xc5\xbcO\x18\x9fG\xc0/\xc2\xfdX\x87\xcdd\x86\x94G@?,M\xec\x18\x7fIG\xc0/\xb8\xcd\x9c\xcc\x11{\x86\x94G@?\xa2\xf1\x8c\x85\xbbXG\xc00\x12m\xf3\xf1\xfc\x04\x86\x94G@?\xda\x1fU\xe8\xb8\xe1G\xc00Q\xc5P\x1a\xf2\xde\x86\x94G@@*\x03\xbc\x05n\x90G\xc00d^\xf7[\xcd\xa4\x86\x94G@@l\x7fm\x83\xba\xccG\xc00\xb6\xa1Bs\xb3\xbe\x86\x94G@@l\xc8=\xbaW\xd0G\xc01\xfa\xa3\x7f\xe4]&\x86\x94G@@S\xd3Ka\x98<G\xc02\xac\x0e\x16\x14)\x96\x86\x94G@@NU\xd3\xea \xacG\xc03k\\\xacC\xf6B\x86\x94G@@b\xe8\x18.eLG\xc03\xb71\x0b\xd6\xbd\x8a\x86\x94G@@Trw\xc1+\x04G\xc04M\xe5\xf3\xf2\x0e\xa2\x86\x94G@@A\x1c\xda\xbd\xc0\xe0G\xc04e1\xdfw-\xc1\x86\x94G@@\x1f[\xc6B\xdf\x04G\xc05\x1d\xd21b\xb6\x0c\x86\x94G@?1\x000\\\xc3\xa0G\xc06@b\xf0\x88F\xfc\x86\x94t\x94\x85\x94uauK\n}\x94(h\x02\x8c\x0249\x94h\x04h\x05h\x06}\x94(h\x08\x8c\x06Africa\x94h\nG@\xe1\x87\x80\x00\x00\x00\x00h\x0b\x8c\x03BWA\x94h\r\x8c\x08Botswana\x94h\x0fJ\xca\xcb!\x00uh\x10(G@3\xe5<\xb8\xe5\x1axG\xc0:\xd4\x1bd\x96\x02^G@=n\xa3\xe5D\xd71G\xc01\xa9l\xc0\xbe\xcf\xd0t\x94h\x12]\x94}\x94(h\x04h\x15h\x16(G@=n\xa3\xe5D\xd71G\xc06\x17`E\xdd\x9b\x9c\x86\x94G@<\x04i\x93YK\xd8G\xc06\xd3\xe7\xa8\xda5\n\x86\x94G@;\x1e\x91\xa1\x00\x88\xc0G\xc07\x93\x06\xd59\xcb\x19\x86\x94G@:\xc9Q\xf2\xec\x0chG\xc08=\x9d\xe6J\xdf\x19\x86\x94G@:|ZR~\xd1\x10G\xc08\x9d\xc7\x94_[\x9f\x86\x94G@9\xf1\x10\x1b\xe1\xcb@G\xc08\xb2E\x86\xb8\x1b \x86\x94G@9\xc4\x0e\xab=\xf3qG\xc09,\xc2\xacD\x0f\x94\x86\x94G@9\xaa'\x93YA\x80G\xc09|\x9f\xfa\xc5\xc5\xda\x86\x94G@9\x06q\x93Y>\xb0G\xc09\xb8<L\xb1L\x00\x86\x94G@86\x15\x93Y; G\xc09\xab\x93Bs\xdb$\x86\x94G@7\xbb\xcb:\x9a\x13xG\xc09c\xdf\x86\xb8\x1e,\x86\x94G@7O\xe5\x93Y7(G\xc09D\xc8\xdc\rr\xfa\x86\x94G@6\xd3\x03q7\x12\xe0G\xc09\x80\x1e\x0f@\xa72\x86\x94G@6\x94\\0\\\x9d\xc0G\xc09\xfa\xbd\x12\xaay\x84\x86\x94G@6\x1b \xc6\x8ce\x0fG\xc0:G\xbe\xdc\rwl\x86\x94G@5\x9b\x1c\x00\x936xG\xc0:\xb9\xfe\x1c\xe7\xedp\x86\x94G@4\xe3\xbdjci\xefG\xc0:\xd4\x1bd\x96\x02^\x86\x94G@4\xaa\x9d\xc9\xf6.\xf0G\xc0:z:a,0\x9b\x86\x94G@4\xc24706(G\xc09\xde>1b\xca\xf2\x86\x94G@4*l\xfd)_\xf0G\xc08\xea\xff\x8d\x8b\xbc\x86\x86\x94G@3\xe5Q\n\xd0\x9f\x98G\xc08\xc4\x8d\xe6J\xe1j\x86\x94G@3\xe5<\xb8\xe5\x1axG\xc05\xd9bZX{\xd3\x86\x94G@4\xe1\x92\x00\x933HG\xc05\xd0w\xbdU\x11\xdc\x86\x94G@4\xe9\x1f\xc9\xf60\x01G\xc02@\x91ki}\x12\x86\x94G@5\xa7\xb0\xb8\xe5\"8G\xc028\x19\xf3\xf2\x05v\x86\x94G@72eO\x14\xf2`G\xc01\xdeyIGYB\x86\x94G@7\x949\xb5{ZxG\xc02H\x00\xb9\xeb2\x1c\x86\x94G@87\xa53\xc6u'G\xc01\xe3\xac?\t\xe8\xb4\x86\x94G@8\x85L\xef\x822@G\xc01\xe3\x1a\x9e\x9c\xae\xaa\x86\x94G@9\x15\x9e\x15\x0e'\x10G\xc01\xa9l\xc0\xbe\xcf\xd0\x86\x94G@9C\xa4K\xab+HG\xc01\xbc\x8d\xdfw\"\x0e\x86\x94G@9\xa6/\x93YApG\xc02\x898\xfe/w~\x86\x94G@9\xd9\xb3Am\xbd8G\xc02\xb6\xe3\xc4(\xa4\xa6\x86\x94G@:*/\xbcO\x06GG\xc03K\x07\xa8\xda%~\x86\x94G@;K\xe7\xbcO\x0b@G\xc04d:\xa5pZ\x1c\x86\x94G@;\xb9\x89\n\xd0\xc2\x08G\xc04\x7f\xc2L\xb15\n\x86\x94G@;\xba+\x9a,\xea\xffG\xc04\xda\x0f\xaf\xad\xcc\xc6\x86\x94G@<\x05x\x82H:\xd0G\xc05|h\xdc\rbV\x86\x94G@<\xcbn\x96\xc3\x1f\x80G\xc05\xa3\xb3Bs\xc9j\x86\x94G@=n\xa3\xe5D\xd71G\xc06\x17`E\xdd\x9b\x9c\x86\x94t\x94\x85\x94uauK\x0b}\x94(h\x02\x8c\x0250\x94h\x04h\x05h\x06}\x94(h\x08\x8c\x06Africa\x94h\nG@\xd9a\x80\x00\x00\x00\x00h\x0b\x8c\x03NAM\x94h\r\x8c\x07Namibia\x94h\x0fJ,\xea%\x00uh\x10(G@'w\xe8\xe9Ad\xa0G\xc0=\x0b\xa3d\x96\x0c\x1dG@9\x15\x9e\x15\x0e'\x10G\xc00\xf0\xfb\xd8\xa3~\"t\x94h\x12]\x94}\x94(h\x04h\x15h\x16(G@3\xe5Q\n\xd0\x9f\x98G\xc08\xc4\x8d\xe6J\xe1j\x86\x94G@3\xe5\rO\x14\xe3\xe0G\xc0<v\n\xf7\\\x02\xb8\x86\x94G@3\x00\x8bjca\xa8G\xc0<\xf8\xf2\x08m\x16\n\x86\x94G@2w\x03\xa1\x00b\xb0G\xc0=\x0b\xa3d\x96\x0c\x1d\x86\x94G@1\xd6\x0e\x0e:f\xc0G\xc0<\xdb;\x94_nF\x86\x94G@1c3\x03\xfc\xf4(G\xc0<\xc8\x94a,:\xc2\x86\x94G@18\x0b\xb5{>\x80G\xc0<[\x1f\x19~$c\x86\x94G@0\xd2\xf2\xcd_\xee@G\xc0<\x15\x08\x8a!\xfa:\x86\x94G@0XPf\xf9\x85\xc0G\xc0<\x93\xa2\xf0\x88b\xce\x86\x94G@/4!\x7fqP\x11G\xc0;\xd2=Bs\xe4\x9a\x86\x94G@.k\xc3\x0bc\xa5`G\xc0;\x17H\xe2\xe1\x1bj\x86\x94G@-\xfa\xbb]O(\x80G\xc0:\x1e\x0c\x16\x14JV\x86\x94G@-|\x86\x90\x82Y\x90G\xc09d\x96g\xff\xccD\x86\x94G@,\xd0\xf8Ejq\xf1G\xc07\xda_ Q\xb1\x04\x86\x94G@,\xc5|\xa4\xfd7\xb1G\xc06\xa8\x1ag\xff\xc0=\x86\x94G@,\x83\xf3\x19\n\xdd\xd0G\xc06\x1cx#\xbby\x90\x86\x94G@+\xbc\xbe\xab\xd0\xd3\x90G\xc05\xb2\xf4\x16\x146\xe6\x86\x94G@*\xb4z\x9e)\x8e1G\xc04\xdfr\x0f@\x92\xd6\x86\x94G@)\xa7XEjd\x00G\xc03\xacT\x97\xc9\x16\x18\x86\x94G@)7\x95\xb6\x0e9\x1fG\xc03\x0b\x9b\xfa\xc5\xa9\x86\x86\x94G@'\x96\xff\x93\xec\x0f\xd0G\xc02\x11\xb2u\xa6\xec\xe8\x86\x94G@'w\xe8\xe9Ad\xa0G\xc01MH\x9e\x9c\xac\x18\x86\x94G@(nP\xf6\xe8\xa9\xc0G\xc01\x1c\x96L\xb1&\"\x86\x94G@)\xa0\xcfB\x00\x93\xb0G\xc00\xf0\xfb\xd8\xa3~\"\x86\x94G@*\xec\xba\xb9x\x10\xe0G\xc00\xf8\xa1V\xee\x96(\x86\x94G@,\x1d\xf3\xe5\xd7\xa8\xe0G\xc01lb\xacC\xedz\x86\x94G@,k^\xab\xd0\xd6\x90G\xc01Zd\xcef\x0fN\x86\x94G@2Ch>\x03\xcb\xa0G\xc01OX\xf0\x881@\x86\x94G@2\xf4\xc8\xab=\xd5\x80G\xc01\xca\x02\x1c\xe7\xc6\"\x86\x94G@5`\x8e\x9d\x96\x9fHG\xc01\xee>1b\xa8\x08\x86\x94G@77\rjct0G\xc01\x85\xea\xf0\x8820\x86\x94G@8\x08\xab&\x1f3\x80G\xc01K\xbca,\x08:\x86\x94G@8\xae\xaem\xcdJ\xd8G\xc01Zy Q\x94n\x86\x94G@9\x13\xb3\x03\xfd\x15\xf9G\xc01\x94-\xc4(\x9f\xa8\x86\x94G@9\x15\x9e\x15\x0e'\x10G\xc01\xa9l\xc0\xbe\xcf\xd0\x86\x94G@8\x85L\xef\x822@G\xc01\xe3\x1a\x9e\x9c\xae\xaa\x86\x94G@87\xa53\xc6u'G\xc01\xe3\xac?\t\xe8\xb4\x86\x94G@7\x949\xb5{ZxG\xc02H\x00\xb9\xeb2\x1c\x86\x94G@72eO\x14\xf2`G\xc01\xdeyIGYB\x86\x94G@5\xa7\xb0\xb8\xe5\"8G\xc028\x19\xf3\xf2\x05v\x86\x94G@4\xe9\x1f\xc9\xf60\x01G\xc02@\x91ki}\x12\x86\x94G@4\xe1\x92\x00\x933HG\xc05\xd0w\xbdU\x11\xdc\x86\x94G@3\xe5<\xb8\xe5\x1axG\xc05\xd9bZX{\xd3\x86\x94G@3\xe5Q\n\xd0\x9f\x98G\xc08\xc4\x8d\xe6J\xe1j\x86\x94t\x94\x85\x94uauK\x0c}\x94(h\x02\x8c\x0251\x94h\x04h\x05h\x06}\x94(h\x08\x8c\x06Africa\x94h\nG@\xe3e\x00\x00\x00\x00\x00h\x0b\x8c\x03SEN\x94h\r\x8c\x07Senegal\x94h\x0fJ\xea\xd2\xdf\x00uh\x10(G\xc01\xa0\x02\xcc:B\xe0G@(\xaa\x07\xb1\xb6\x9bqG\xc0&\xef\x90y\xbb\xfd\xa8G@0\x99'\xce\x9d\xef\xd8t\x94h\x12]\x94}\x94(h\x04h\x15h\x16(G\xc00\xb6\xb6\xee\\a\x00G@+0\x9ej\x08\x92\x10\x86\x94G\xc01 H\x87\xf5\xfchG@,\xbf=tF\t\x90\x86\x94G\xc01\xa0\x02\xcc:B\xe0G@-u\x86U\x8d\xba\xbf\x86\x94G\xc01/g}\xb8\x8c\x08G@-\xd6\xc5\xb8\x8aR\x9f\x86\x94G\xc00\xb3a}\xb8\x89\xe8G@/>8\xd7B\xaa\xb8\x86\x94G\xc00v\x8d\x99\x07\n\x90G@0\"\x91\xba#\x0c\x88\x86\x94G\xc00\x1e\xe5\x8b_\xc84G@0t\xa6L\xe9\x07\x1c\x86\x94G\xc0/?Q+:mPG@0^\x8c\xdf\xae\xff\xe8\x86\x94G\xc0.E\x7f[\x03\xcb\xffG@0\x96X#\xf3E \x86\x94G\xc0-'\x9a\x1a)S\x08G@0\x99'\xce\x9d\xef\xd8\x86\x94G\xc0,2\xf4y\xbc\x14\xd0G@0M\xe6\xc0\xf6\xad\xb4\x86\x94G\xc0*\xdf\x19\x02D\x97\x80G@0\n\x15\x01\xd1 \x94\x86\x94G\xc0)\xa9L\x0c\x82\x02\xd1G@.\x9b}tF\x11\xc0\x86\x94G\xc0(Wl\x95\n\x85\x90G@-;\xd1\xb1\xb6\xaf\x88\x86\x94G\xc0(?\xf1?\xb5/\xd1G@+\xfdL\xeb\xbd}\xb0\x86\x94G\xc0'\xda\xfd\x98tS\xa0G@*\xd8\x1a:?-\x90\x86\x94G\xc0'\x1bV\xf4\x9dF\x18G@*HM).\x1a\x08\x86\x94G\xc0&\xef\x90y\xbb\xfd\xa8G@)\x82PN\xba\x08\xf8\x86\x94G\xc0'\x07#\x83\xf9n\xb8G@(\xe2\xcfD|\x95\x98\x86\x94G\xc0'Q\x0c\xd2{$\xe1G@(\xc5\xee,\x97\xe3\x99\x86\x94G\xc0(h9\xa6\x1b\x96\xe7G@(\xeei_\xcb\x17\x80\x86\x94G\xc0(\x8e\xa4\x87cE\xa8G@(\xb5y).\x13\x18\x86\x94G\xc0(\xff\x83\x91\xa0\xb88G@(\xaa\x07\xb1\xb6\x9bq\x86\x94G\xc0*o\x85\xdc\xb8\xa3@G@)&\xd8\xe4\xe9\xd0\xc8\x86\x94G\xc0+f\xa4\xcb\xa7\x96hG@), ,\x97\xe5Y\x86\x94G\xc0/\x18\xd1\xf8\x079pG@)A\x9f\x81\xed;\x10\x86\x94G\xc0/\xa2\x16\x05\xae|\xa9G@)\x07\xf8j\x08\x88\x90\x86\x94G\xc00%\xd0\xc5f\x9b\xf4G@)\x18t3k\x85p\x86\x94G\xc00\xadm}\xb8\x89\xccG@(\xc5\x0bD|\x95\x10\x86\x94G\xc00\xd7n(c50G@*M\x83\x81\xed?\xa8\x86\x94G\xc0/\xdc\xd2\xd2{JwG@*B\xb4\x99\xd1\xf0\xf8\x86\x94G\xc0/a\xca\xd2{HWG@*\x8ak\xb8\x8aD \x86\x94G\xc0/\x06\x0cI\xf2\xbe@G@*\x8e\xa0\xaeL\xd3\x90\x86\x94G\xc0.HF\x8e6\xff@G@+\x04\xde\xb5 u\xff\x86\x94G\xc0-l\xa5\x1d\x93$pG@*\x98\xae\x8c*\xb1\x99\x86\x94G\xc0,\x8e.\xed\xc9\xbd\xa0G@*\x8f\xa8\xd7B\x96 \x86\x94G\xc0+\xb0\x9f\t\x18;\x87G@+\x02\x94\xd0n\xf7\xb0\x86\x94G\xc0,\x18\x0fa\xd7b\xd7G@+\x96\x90\x11InH\x86\x94G\xc0,\xc0\xe0\xa2\xb1\xd9\xd0G@+@Y).\x1eH\x86\x94G\xc0-_\xc2\x80\x8f\xbahG@+B\xbe%\xc4N\x19\x86\x94G\xc0.)\xd98\xe1\xa9hG@+\xc0\xc3\x88\xc0\xe6x\x86\x94G\xc0.\xcc+\x9fH\x12\x97G@+\xb8\x82A\x12\xd1\xd8\x86\x94G\xc0/?\xcb\x16\xbf\x8c\x08G@+?F\xd7B\x99(\x86\x94G\xc00\xb6\xb6\xee\\a\x00G@+0\x9ej\x08\x92\x10\x86\x94t\x94\x85\x94uauK\r}\x94(h\x02\x8c\x0252\x94h\x04h\x05h\x06}\x94(h\x08\x8c\x06Africa\x94h\nG@\xe2\x99@\x00\x00\x00\x00h\x0b\x8c\x03MLI\x94h\r\x8c\x04Mali\x94h\x0fJ=\xe8\x10\x01uh\x10(G\xc0(Wl\x95\n\x85\x90G@$1V3ko\xe0G@\x11\x14\xb1\xed\xd0\xc7\xa0G@8\xf9}\xaf\xe5\xc2\xc3t\x94h\x12]\x94}\x94(h\x04h\x15h\x16(G\xc0'\x07#\x83\xf9n\xb8G@(\xe2\xcfD|\x95\x98\x86\x94G\xc0&\xef\x90y\xbb\xfd\xa8G@)\x82PN\xba\x08\xf8\x86\x94G\xc0'\x1bV\xf4\x9dF\x18G@*HM).\x1a\x08\x86\x94G\xc0'\xda\xfd\x98tS\xa0G@*\xd8\x1a:?-\x90\x86\x94G\xc0(?\xf1?\xb5/\xd1G@+\xfdL\xeb\xbd}\xb0\x86\x94G\xc0(Wl\x95\n\x85\x90G@-;\xd1\xb1\xb6\xaf\x88\x86\x94G\xc0'\xab\x1d?\xb5-@G@-\x99#=\xa9\t\xe0\x86\x94G\xc0'U\x085w\xbb G@.\xc6\xc3=\xa9\x0f\x10\x86\x94G\xc0&\xb2\xbc\x95\n~PG@.\xd2\x90%\xc4]\xc8\x86\x94G\xc0%M4\x87c7PG@.C\xf7D|\xad@\x86\x94G\xc0$,w$f\x9c0G@.\xa95mrq\x88\x86\x94G\xc0#f\x87\xd5\xe4\xe3\xe0G@.\x879\x14\xb3Kh\x86\x94G\xc0#\x19\xb8\xd9N\xb2\xc0G@.\xf9\x16\x1e\xf0\xbe\x00\x86\x94G\xc0\x16&\xa6q\xc2\xaa\xe0G@/\x00\xdd{\x19\xb3\xe9\x86\x94G\xc0\x15B\xd8\x0b\\@\x90G@03\xac\xaf\xe5\x9c,\x86\x94G\xc0\x15\xf4?>\x8fv\xd0G@0S9\xe3\x18\xcf\xec\x86\x94G\xc0\x17\xe2o\x90{\x04pG@4\xa4\r\xa9\x12\x0fH\x86\x94G\xc0\x19\xd0\xadnX\xea\xc0G@8\xf4\xe3 \x89\x99\xb8\x86\x94G\xc0\x13\xb1\x7fY\xdd\xee\xa0G@8\xf9}\xaf\xe5\xc2\xc3\x86\x94G\xbf\xf8\xcd\x06]9\\\x80G@6\xca\xec']0\xa4\x86\x94G?\xfd+\xf0\xad\x05\x02\x00G@4\x9c^\x01\xd14\xb0\x86\x94G@\x00|\xe8\xc3\xbc\x90\x00G@4$ih7\x99\x08\x86\x94G@\x05w\xfd>\x9d\xed\xc0G@3\xdb1\xe6\x82\xaf\xac\x86\x94G@\t,\\\x9a\xc6\xf3\xc0G@3\xb1\x8e]\xfa&l\x86\x94G@\tC\xdbS\x18\xdf\xc0G@3\x0e\xafk\xa1dx\x86\x94G@\x11\x11\xd6h\xb2\x0f@G@3'\xbfu\xde\xd5\x8c\x86\x94G@\x11\x14\xb1\xed\xd0\xc7\xa0G@0\xda+\x94\x97\x1dT\x86\x94G@\r\xc9\x91LES\x00G@0/-8n$\xa4\x86\x94G@\r\x1b'|\x0e\xb3\x00G@/\"\xe0\x99\xd2\x06h\x86\x94G@\x05\xff\xfc-\x8c\xdf\x00G@.\xd1\xad=\xa9\x0f@\x86\x94G?\xf6+\x1f\x9b\xf3\xd2\x01G@.\xa5\xa9\xcd\x057x\x86\x94G?\xf0@\xa6\x02Z\x1e\x00G@-\xef\xb5\x96h1\x00\x86\x94G?\xd7\xfe;\xe0r\x02\x00G@-\xdb\x99\xda\xact\xd7\x86\x94G\xbf\xd1\n\\\x04@0\x00G@-\xd9?\x07\x0c\x07\x90\x86\x94G\xbf\xe0\x81\xe15So\x00G@.;y\x07\x0c\t@\x86\x94G\xbf\xf1\x0f\xd3*\x06\x07\x00G@-\xf2\x97\xe1\x80\x15\xa8\x86\x94G\xc0\x00\x02\x1e\xb3\xbbv@G@-\x1e6U\x8d\xb9@\x86\x94G\xc0\x01\x88\xdbI\xebF\x80G@,~*p\xdc88\x86\x94G\xc0\x07\xbd\xd6\x98m\x16\xc0G@+\x98\xa7).\x1f\xd1\x86\x94G\xc0\x08\xd4d?\xad\xf5\xbfG@+\x15 \xeb\xbdy\xb0\x86\x94G\xc0\x0c.\xb3.\x9c\xf3\x80G@*\xac\xe1\xfc\xce\x88\xf8\x86\x94G\xc0\x10\x06\x8bL6\x9d\xa0G@*\xf1\xe9\x9d;\xc40\x86\x94G\xc0\x11\x1f\"\x7fi\xd5\xa0G@*t\xf6\x8c*\xb0\xf8\x86\x94G\xc0\x11\xb5k\x07\xf2`\xc0G@)\x15\xd5\xa4\x0f\\q\x86\x94G\xc0\x14\xe2>\x9a\xb8g\xe0G@'m~\xeb\xbdi\xa0\x86\x94G\xc0\x14\xca\x97>\x8fq\xb0G@&\xc0\x13\x1b\x86\xc9\xa0\x86\x94G\xc0\x15\xe1\xdb\xc7\x17\xff\x00G@%\xe7\x0c\xd7B\x81\xa1\x86\x94G\xc0\x15\x9e\x0b\xb9p\xbd\x00G@$\xbd\xd16\xd5B\x81\x86\x94G\xc0\x17D\x88O\xa0\x8d\xa0G@$q\xf2\xaeL\xb8\xa8\x86\x94G\xc0\x183\xa9\xb2\x9d'\xf0G@$1V3ko\xe0\x86\x94G\xc0\x18\xd2%\xf6\xe1n\xf1G@%\x0cQ\xb1\xb6\x8b\x88\x86\x94G\xc0\x19\xf9\xd1\xf6\xe1t\x00G@$\xd2\x96G\xe6S\xf0\x86\x94G\xc0\x1a\xaat\xbc\xda\xa3pG@$\xdc\x936\xd5C\x08\x86\x94G\xc0\x1bf\xeb0\xe8M\xf0G@$G*3kp@\x86\x94G\xc0\x1e}\xb4\x93\xe4\xf1\xb1G@$Kb\x8c*\x95\xe0\x86\x94G\xc0\x1f\x99.\x120\x0eqG@$\x98Bw\xaf\xb5\xe7\x86\x94G\xc0 \x0fT\xc4\xd3\xc4\x18G@$i\xbe\xf2\x90\xfc\xc8\x86\x94G\xc0 \xab\xb6\x8e6\xc3`G@$\xfdW\xfc\xceo\xf8\x86\x94G\xc0 \x90\x91\x1d\x92\xeb\xe0G@%\x95\xcfR#\xc7\xe7\x86\x94G\xc0 \xd0\x8b\t\x18\x0b\xb0G@%\xd1\x8a\x1e\xf0\x95\xc1\x86\x94G\xc0!=\x9a\xb7,\x88pG@%\x9f-\x14\xb3$A\x86\x94G\xc0!)\xa0\xd9N\xaa@G@&E\xc1\xfc\xceu\x98\x86\x94G\xc0 \xc0\xab\t\x18\x0bpG@&\xc9\x8b\xf5\xfa\xd7p\x86\x94G\xc0!\x92{\x91\xa0\x97\x90G@'\xa0\x07\xfc\xce{\x90\x86\x94G\xc0!\xcf~\xe0\"M\x88G@(-=D|\x92y\x86\x94G\xc0\"AD5w\xa4\xd0G@(\x9d\xba\x18\x1d\x01\xa0\x86\x94G\xc0\"\xa7\xbdT/\xf8\x80G@(\xab'\x8f\x94yP\x86\x94G\xc0#\"\xc5T/\xfa\x98G@(cs\xd3\xd8\xbcX\x86\x94G\xc0#\xc805w\xab\x88G@(\x1e\xf7\r\xdf\x8e\xd1\x86\x94G\xc0$T\x96\xe6\xf5\xf9\x08G@'\xb0+\xb8\x8a7\x90\x86\x94G\xc0%/\xbb\t\x18\x1e\xf0G@'\xd9\x13KP1p\x86\x94G\xc0%\xbd\xdda\xd7F\xe7G@([\x14\x11I`\x11\x86\x94G\xc0&\x12\xb7vR)\xa8G@(l(A\x12\xc3Y\x86\x94G\xc0&\x98[\x91\xa0\xad\xa8G@('\xeb\xd3\xd8\xbbX\x86\x94G\xc0&\xe9\x8e\xed\xc9\xa4\xd0G@('V\xd0n\xeb\x18\x86\x94G\xc0'\x07#\x83\xf9n\xb8G@(\xe2\xcfD|\x95\x98\x86\x94t\x94\x85\x94uauK\x0e}\x94(h\x02\x8c\x0253\x94h\x04h\x05h\x06}\x94(h\x08\x8c\x06Africa\x94h\nG@\xd0Q\x80\x00\x00\x00\x00h\x0b\x8c\x03MRT\x94h\r\x8c\nMauritania\x94h\x0fJ\xebY9\x00uh\x10(G\xc01\x10<\x81\"[\xb4G@-;\xd1\xb1\xb6\xaf\x88G\xc0\x13\xb1\x7fY\xdd\xee\xa0G@;eO|\xb2\x9a8t\x94h\x12]\x94}\x94(h\x04h\x15h\x16(G\xc01\x10<\x81\"[\xb4G@4\xff\xef\xc0\xf6\xc2X\x86\x94G\xc00\xd8^\x9cp\xdctG@5UT\xaf\xe5\xb2\xc1\x86\x94G\xc0)\xdb\xb3?\xb56\xe0G@5S\xba\xe6\x82\xb6$\x86\x94G\xc0*<\xcda\xd7Z\xb0G@6\xc5n\xaf\xe5\xb9\x14\x86\x94G\xc0)\xbf\x99\xf8\x07!\xf1G@7H\xea\xc4`\x9c\x9c\x86\x94G\xc0'\xdf\xdb\xe3\x8c8hG@7_\xe5h7\xa7@\x86\x94G\xc0'\xf0W\xac\xef5HG@9\xee\xf05\x04\x7fK\x86\x94G\xc0!_\xe4\xf4\x9d,\xe0G@9\xe1\x8c\xe6\x82\xca,\x86\x94G\xc0!^i\xa6\x1bw\xf0G@;eO|\xb2\x9a8\x86\x94G\xc0\x13\xb1\x7fY\xdd\xee\xa0G@8\xf9}\xaf\xe5\xc2\xc3\x86\x94G\xc0\x19\xd0\xadnX\xea\xc0G@8\xf4\xe3 \x89\x99\xb8\x86\x94G\xc0\x17\xe2o\x90{\x04pG@4\xa4\r\xa9\x12\x0fH\x86\x94G\xc0\x15\xf4?>\x8fv\xd0G@0S9\xe3\x18\xcf\xec\x86\x94G\xc0\x15B\xd8\x0b\\@\x90G@03\xac\xaf\xe5\x9c,\x86\x94G\xc0\x16&\xa6q\xc2\xaa\xe0G@/\x00\xdd{\x19\xb3\xe9\x86\x94G\xc0#\x19\xb8\xd9N\xb2\xc0G@.\xf9\x16\x1e\xf0\xbe\x00\x86\x94G\xc0#f\x87\xd5\xe4\xe3\xe0G@.\x879\x14\xb3Kh\x86\x94G\xc0$,w$f\x9c0G@.\xa95mrq\x88\x86\x94G\xc0%M4\x87c7PG@.C\xf7D|\xad@\x86\x94G\xc0&\xb2\xbc\x95\n~PG@.\xd2\x90%\xc4]\xc8\x86\x94G\xc0'U\x085w\xbb G@.\xc6\xc3=\xa9\x0f\x10\x86\x94G\xc0'\xab\x1d?\xb5-@G@-\x99#=\xa9\t\xe0\x86\x94G\xc0(Wl\x95\n\x85\x90G@-;\xd1\xb1\xb6\xaf\x88\x86\x94G\xc0)\xa9L\x0c\x82\x02\xd1G@.\x9b}tF\x11\xc0\x86\x94G\xc0*\xdf\x19\x02D\x97\x80G@0\n\x15\x01\xd1 \x94\x86\x94G\xc0,2\xf4y\xbc\x14\xd0G@0M\xe6\xc0\xf6\xad\xb4\x86\x94G\xc0-'\x9a\x1a)S\x08G@0\x99'\xce\x9d\xef\xd8\x86\x94G\xc0.E\x7f[\x03\xcb\xffG@0\x96X#\xf3E \x86\x94G\xc0/?Q+:mPG@0^\x8c\xdf\xae\xff\xe8\x86\x94G\xc00\x1e\xe5\x8b_\xc84G@0t\xa6L\xe9\x07\x1c\x86\x94G\xc00v\x8d\x99\x07\n\x90G@0\"\x91\xba#\x0c\x88\x86\x94G\xc00\x8c\xb9\xa6\xaeK\xccG@0\xac\x841\x9a\x86\\\x86\x94G\xc00EB\xe0\xb5\x1e0G@1*\xbe\x12\xe26\x9c\x86\x94G\xc00%w\x06A\x10\x00G@2\x1b\xc5ru\x00\xb8\x86\x94G\xc00A\xc3\x1a\xbb\xf1\xc4G@3\x18\xc2]\xfa#\xcc\x86\x94G\xc00`\xad\xbe\x92\xfc\x88G@3\x98\x04h7\x96\xa0\x86\x94G\xc00G e\xd3\xd6\x8cG@4\x17\xafo\x0b9<\x86\x94G\xc00\x89L\x81\"YdG@4\x91_\xaf\xe5\xaf`\x86\x94G\xc01\x10<\x81\"[\xb4G@4\xff\xef\xc0\xf6\xc2X\x86\x94t\x94\x85\x94uauK\x0f}\x94(h\x02\x8c\x0254\x94h\x04h\x05h\x06}\x94(h\x08\x8c\x06Africa\x94h\nG@\xd7\xbd\x80\x00\x00\x00\x00h\x0b\x8c\x03BEN\x94h\r\x8c\x05Benin\x94h\x0fJUp\xa8\x00uh\x10(G?\xe8\xb6\xf9>\xba\xee\x00G@\x18\x91\x91\xc9\xd30`G@\x0e`|d*\x07@G@(x\xa5D|\x93\xa0t\x94h\x12]\x94}\x94(h\x04h\x15h\x16(G@\x05\x88\x9a\xe5\xde\xc8\x80G@\x19\t\x07cl\xcc\x00\x86\x94G?\xfd\xd8\x06o\x94a\x00G@\x18\x91\x91\xc9\xd30`\x86\x94G?\xf9\xe78\xc8Su\x00G@\x1bT\x01\xc9\xd3<\x80\x86\x94G?\xfa\xa1\xb3<a\x1f\x80G@\"A\xd6\x99\xd1\xcd\xc0\x86\x94G?\xf7h\x9f\x9b\xf3\xd7\x80G@\"\xabS\xe1\x7f\xe4\x10\x86\x94G?\xf6\xcd\x0ch\xc0\xa1\x81G@#\xa6\x9a:?\r\xe8\x86\x94G?\xf1>\xa6\x02Z\"\x80G@$Y\xe9\x1b\x86\xbf\x10\x86\x94G?\xe8\xb6\xf9>\xba\xee\x00G@$\xf1\r\xc61lX\x86\x94G?\xec\xc98dF\xf2\x00G@%\xfe\xa3D|\x88\xe0\x86\x94G?\xf3\xe5@vg\xd5\x80G@&8\x94\xde\x16#x\x86\x94G?\xf7'\xa4Mr!\x80G@'\x18n\xa7y#\xe8\x86\x94G?\xfe\xf9\xcb\xfb\x86\xbe\x80G@'HD\xd7B\x87\xb0\x86\x94G@\x01<\\\x9a\xc6\xd0\xc0G@'\xe1[X\xf7rp\x86\x94G@\x03\xeb\xda\xe5\xde\xc1@G@(wR\x99\xd1\xe9\x10\x86\x94G@\x06\xca\x05Y\xect\x80G@(x\xa5D|\x93\xa0\x86\x94G@\x0c\xe3\xb2\x93\xf3c\xbfG@'R\x01mrQA\x86\x94G@\x0c\x93\xe64`\x9c@G@&\xa7\xe7\xaa\xe2\xf2/\x86\x94G@\x0e`|d*\x07@G@%x0\x92\xfe;y\x86\x94G@\x0c\xcc\xf1\x82\xe2R@G@$\xaa\x14N\xb9\xf3\xa8\x86\x94G@\r\xa4\xbc\xd1d\n\xc0G@$ ]\x1b\x86\xbe\x18\x86\x94G@\t\xc3G\xb2\xab\xa7\xbfG@\"\xe3g\xf5\xfa\xc6H\x86\x94G@\x07LhV\x82\xa7@G@\"Ft\x8c*\x8d\x00\x86\x94G@\x05\xcaS\xdb\xa1Y\x00G@!\x03\x81=\xa8\xd2\x88\x86\x94G@\x05\xfe\x14\x7fxd\x00G@\x1f{\xa1\xc9\xd3N\xc0\x86\x94G@\x05\x88\x9a\xe5\xde\xc8\x80G@\x19\t\x07cl\xcc\x00\x86\x94t\x94\x85\x94uauK\x10}\x94(h\x02\x8c\x0255\x94h\x04h\x05h\x06}\x94(h\x08\x8c\x06Africa\x94h\nG@\xd3\xad\x80\x00\x00\x00\x00h\x0b\x8c\x03NER\x94h\r\x8c\x05Niger\x94h\x0fJ \xa9%\x01uh\x10(G?\xd2\xeb\xde\xdd\x08\x1c\x00G@'R\x01mrQAG@/\xcevY\xe5``G@7x\xbfB\xab\xb5Tt\x94h\x12]\x94}\x94(h\x04h\x15h\x16(G@-\xb3\xdd\x97\xf6+\x80G@6\xdc\xeaJ\x8c\x15p\x86\x94G@.1\x9bB\x00\xa7\xa0G@5N\xfb\x16L\x190\x86\x94G@.\xf1.\xc6\xbc\xe8\x80G@5\x0cg8\x1d}\xd0\x86\x94G@.\xf9kx\x9d\xae\xa0G@4\xba\xfcru\x0cP\x86\x94G@/\xcevY\xe5``G@4c:\xfegc\x90\x86\x94G@/_\x19d\"\xcf\x00G@3\xf5\t\xc0\xf6\xbd\xe0\x86\x94G@.\x99\xd3j\xf6l`G@1\xed\x8e \x89z\xd0\x86\x94G@.~\xd6\x9e)\x9e\xc0G@0\xa0\x97\x1d\x1f\xa4\xd0\x86\x94G@+\xf1\xc0D(N\x00G@/^e\xbe\xa0\xba@\x86\x94G@+\x14\xaeu3\xcd\x80G@,\xbb\xf8\xf2\x91!\x80\x86\x94G@+\xe9\xd4n`0\xc0G@+\xfeNN\xba\x14\x00\x86\x94G@+\xe8\xb1-\x85\xbc\xc0G@*\xb4\xf7=\xa8\xfd \x86\x94G@-1\n>\x96\xd3 G@*\xa9-\xb8\x8aD\xe0\x86\x94G@,\xfd\xd7\xd80k\xe0G@)\xb8\x02\xc9\x9bQ\x91\x86\x94G@,mS\xe5\xd7\xaa@G@)\x9a\xa4c4\xea\xa8\x86\x94G@,\\\xd8\x1ct\xad_G@(\xf7\xa1\xe1\x7f\xff\xc0\x86\x94G@+\xfd\x9e\xe2m\xd8\x00G@(\xecRG\xe6e\xe0\x86\x94G@*\xa3,\xdb\x9a1\xc0G@+\x1c\xda\xbb\xf4\x16\xd8\x86\x94G@*+\x00`\xb8\xe7\xe0G@+1:3k\x8e\x80\x86\x94G@(\x9a\xa9\x127,1G@*\x13\np\xdc-\x98\x86\x94G@'\x0e<7\xc3\x17\xc0G@*\xa8p\x11Ij0\x86\x94G@%\xfa\xab\xf3~\xce\xc0G@*\xc6O).\x1c`\x86\x94G@%f\xed\xa8f\xe7\x81G@*~l\x03\xa2(\x98\x86\x94G@$:\xc8\xf6\xe8\x97@G@*\x8d\xf3\xf5\xfa\xe8\x00\x86\x94G@#\x0c\xc3]N\xf8pG@)\xb3\xc3\xaa\xe2\xff\x90\x86\x94G@\"\x07\xa5V{S\x80G@)\xa7?\xe1\x80\x02\xc0\x86\x94G@\x1f7\xfb\xbe\x07\xa2\xe0G@*\xaf\xe2\xc2\xc7\xb5g\x86\x94G@\x1dR\xafC&R\xe0G@*22\nu\xc7\xb8\x86\x94G@\x1bH!\xed\xd0\xf4\x7fG@*:\xed=\xa8\xfb\x11\x86\x94G@\x19\xc8\x1d\xc4\xdb+`G@*\xfcL%\xc4L\xe7\x86\x94G@\x15\xc5\xb1\x13\\\xce\x80G@+\xbbZc4\xf3\xe0\x86\x94G@\x11y/\x0c\x89\x1b\x00G@+~\xb5\xe8S\xab\x80\x86\x94G@\x10n\x89eH<@G@+\x0f\xfb\x81\xedC\x00\x86\x94G@\x0f\xbc\xfe\xbc\xe92\xc0G@)\xe9\x87\x14\xb37\x1f\x86\x94G@\rq\xf0\x04\x97=\x00G@)\x1b\x16%\xc4D\xa0\x86\x94G@\x0c\xe3\xb2\x93\xf3c\xbfG@'R\x01mrQA\x86\x94G@\x06\xca\x05Y\xect\x80G@(x\xa5D|\x93\xa0\x86\x94G@\x03\xeb\xda\xe5\xde\xc1@G@(wR\x99\xd1\xe9\x10\x86\x94G@\x01<\\\x9a\xc6\xd0\xc0G@'\xe1[X\xf7rp\x86\x94G@\x01j\xb7|\x0e\x7f\xc0G@)@\x02U\x8d\xa8?\x86\x94G?\xf0b\xba\x10\x01_\x80G@)\xb4\"~\x83l\xd1\x86\x94G?\xef\xc7\x07\xc0o\xf5\x01G@*\xab\xe7_\xcb\x1f(\x86\x94G?\xdb\x83\xef\x13\xa5C\xffG@+\xfa;6\xd5bP\x86\x94G?\xd2\xeb\xde\xdd\x08\x1c\x00G@,\xe3r\xc2\xc7\xbf\x18\x86\x94G?\xd7\xfe;\xe0r\x02\x00G@-\xdb\x99\xda\xact\xd7\x86\x94G?\xf0@\xa6\x02Z\x1e\x00G@-\xef\xb5\x96h1\x00\x86\x94G?\xf6+\x1f\x9b\xf3\xd2\x01G@.\xa5\xa9\xcd\x057x\x86\x94G@\x05\xff\xfc-\x8c\xdf\x00G@.\xd1\xad=\xa9\x0f@\x86\x94G@\r\x1b'|\x0e\xb3\x00G@/\"\xe0\x99\xd2\x06h\x86\x94G@\r\xc9\x91LES\x00G@0/-8n$\xa4\x86\x94G@\x11\x14\xb1\xed\xd0\xc7\xa0G@0\xda+\x94\x97\x1dT\x86\x94G@\x11\x11\xd6h\xb2\x0f@G@3'\xbfu\xde\xd5\x8c\x86\x94G@\x16\xb5\xd3\xd9U\xff G@3\x99\xe8\xb3O{X\x86\x94G@!%R>\x96\x9e0G@5\x90\xcf#\xf3[\x08\x86\x94G@'\xff\xbf4YK\xafG@7x\xbfB\xab\xb5T\x86\x94G@+)\xb0|\x07n G@7\n^\x9bj\xd9\x10\x86\x94G@,I\xa9q\xca\x02\x80G@6}\xc5\x1d\x1f\xbe\xa0\x86\x94G@-\xb3\xdd\x97\xf6+\x80G@6\xdc\xeaJ\x8c\x15p\x86\x94t\x94\x85\x94uauK\x11}\x94(h\x02\x8c\x0256\x94h\x04h\x05h\x06}\x94(h\x08\x8c\x06Africa\x94h\nGA0\x9d\xe8\x00\x00\x00\x00h\x0b\x8c\x03NGA\x94h\r\x8c\x07Nigeria\x94h\x0fJE\xd1\\\x0buh\x10(G@\x05\x88\x9a\xe5\xde\xc8\x80G@\x10\xf6^R[\x97pG@-'\x83\xd80l\x90G@+\xbbZc4\xf3\xe0t\x94h\x12]\x94}\x94(h\x04h\x15h\x16(G@\x05\x88\x9a\xe5\xde\xc8\x80G@\x19\t\x07cl\xcc\x00\x86\x94G@\x05\xfe\x14\x7fxd\x00G@\x1f{\xa1\xc9\xd3N\xc0\x86\x94G@\x05\xcaS\xdb\xa1Y\x00G@!\x03\x81=\xa8\xd2\x88\x86\x94G@\x07LhV\x82\xa7@G@\"Ft\x8c*\x8d\x00\x86\x94G@\t\xc3G\xb2\xab\xa7\xbfG@\"\xe3g\xf5\xfa\xc6H\x86\x94G@\r\xa4\xbc\xd1d\n\xc0G@$ ]\x1b\x86\xbe\x18\x86\x94G@\x0c\xcc\xf1\x82\xe2R@G@$\xaa\x14N\xb9\xf3\xa8\x86\x94G@\x0e`|d*\x07@G@%x0\x92\xfe;y\x86\x94G@\x0c\x93\xe64`\x9c@G@&\xa7\xe7\xaa\xe2\xf2/\x86\x94G@\x0c\xe3\xb2\x93\xf3c\xbfG@'R\x01mrQA\x86\x94G@\rq\xf0\x04\x97=\x00G@)\x1b\x16%\xc4D\xa0\x86\x94G@\x0f\xbc\xfe\xbc\xe92\xc0G@)\xe9\x87\x14\xb37\x1f\x86\x94G@\x10n\x89eH<@G@+\x0f\xfb\x81\xedC\x00\x86\x94G@\x11y/\x0c\x89\x1b\x00G@+~\xb5\xe8S\xab\x80\x86\x94G@\x15\xc5\xb1\x13\\\xce\x80G@+\xbbZc4\xf3\xe0\x86\x94G@\x19\xc8\x1d\xc4\xdb+`G@*\xfcL%\xc4L\xe7\x86\x94G@\x1bH!\xed\xd0\xf4\x7fG@*:\xed=\xa8\xfb\x11\x86\x94G@\x1dR\xafC&R\xe0G@*22\nu\xc7\xb8\x86\x94G@\x1f7\xfb\xbe\x07\xa2\xe0G@*\xaf\xe2\xc2\xc7\xb5g\x86\x94G@\"\x07\xa5V{S\x80G@)\xa7?\xe1\x80\x02\xc0\x86\x94G@#\x0c\xc3]N\xf8pG@)\xb3\xc3\xaa\xe2\xff\x90\x86\x94G@$:\xc8\xf6\xe8\x97@G@*\x8d\xf3\xf5\xfa\xe8\x00\x86\x94G@%f\xed\xa8f\xe7\x81G@*~l\x03\xa2(\x98\x86\x94G@%\xfa\xab\xf3~\xce\xc0G@*\xc6O).\x1c`\x86\x94G@'\x0e<7\xc3\x17\xc0G@*\xa8p\x11Ij0\x86\x94G@(\x9a\xa9\x127,1G@*\x13\np\xdc-\x98\x86\x94G@*+\x00`\xb8\xe7\xe0G@+1:3k\x8e\x80\x86\x94G@*\xa3,\xdb\x9a1\xc0G@+\x1c\xda\xbb\xf4\x16\xd8\x86\x94G@+\xfd\x9e\xe2m\xd8\x00G@(\xecRG\xe6e\xe0\x86\x94G@,\\\xd8\x1ct\xad_G@(\xf7\xa1\xe1\x7f\xff\xc0\x86\x94G@-'\x83\xd80l\x90G@(+\xb4j\x08\x84\xc8\x86\x94G@,\xef\xb6\xe2m\xdc@G@'\xcf;\x9d;\xb6h\x86\x94G@,\xd4\xac\x89\xae\xb6?G@'%\r\x88\xc0\xd20\x86\x94G@+%Y\xa8g\x00\xc0G@%\x98\xdd\xa4\x0fM\x18\x86\x94G@*\x9e\n\xd4\xc6\x911G@$R\x1a\xf9d\x9c\xcf\x86\x94G@*U\xcf\xa1\x93\\\xc0G@#H\x00,\x97\xcbp\x86\x94G@)\xe93\x19\n\xd2`G@\"\xd5\xe6,\x97\xc9x\x86\x94G@)\x81\xe1;,\xf2\xc0G@!o~\x99\xd1\xca(\x86\x94G@(p\x10\x01&\x1a`G@ \x9c\x94\xf9d\x8c\x80\x86\x94G@( \xbd\x8d\x18q\xc0G@\x1f3\x00\xfd\x06\x80\xc0\x86\x94G@'\xad\xb9\xdf\x03\xf4\xefG@\x1d\x96\x92D\xb4\x8e \x86\x94G@'}\xd6#H8`G@\x1b\xec\xef\xa7\xb1\x1d\x00\x86\x94G@&\x1e\x19q\xc9\xe7@G@\x1a\x93\xe4\x9ds\xa6p\x86\x94G@$\xfe\xa7\xf3~\xca`G@\x1c8\xaf\xb5X_ \x86\x94G@$<\x8e\xc7\x1f4PG@\x1c'\xb3:w\x170\x86\x94G@#\x0b\xa0\x1ct\x84`G@\x19\xd0]\xae\x84\xb4\x1f\x86\x94G@\"wa\x1f\xdeR\x10G@\x19\xc7(\x8f\xccb\x10\x86\x94G@!\x83\xdb]N\xf1\xb0G@\x15\xeb-\x85\x8e\xe0p\x86\x94G@!\x00%\xb6\x0e\x15\x00G@\x13\x16\x82\xb1\xeef\xc0\x86\x94G@\x1d\xd92\xe3\x93\x8f G@\x11\xa5\xff\xb5X0\xa0\x86\x94G@\x1cT\x94+A\x9d\x00G@\x11\xdb\xd7q\x13\xedP\x86\x94G@\x1a\xca\xd3l\x1c\n?G@\x10\xf6^R[\x97p\x86\x94G@\x17\x97\xba\x91\xa7\xee\x80G@\x11\x0c\xc0\x8f\xcc;\xb0\x86\x94G@\x15s\x83\x1a0m\xa0G@\x13\x8dH09\x80\xb0\x86\x94G@\x14\"aI\xf9\xca\xe0G@\x16r|Y/P\x00\x86\x94G@\x11Mk\xf4\xa4i\x00G@\x19\x15%\x936/0\x86\x94G@\x0c\x97\xeb\xc0R\xf5\x00G@\x19\x08\x7f\xeb\xf5T\x90\x86\x94G@\x05\x88\x9a\xe5\xde\xc8\x80G@\x19\t\x07cl\xcc\x00\x86\x94t\x94\x85\x94uauK\x12}\x94(h\x02\x8c\x0257\x94h\x04h\x05h\x06}\x94(h\x08\x8c\x06Africa\x94h\nG@\xf2\xdb\x80\x00\x00\x00\x00h\x0b\x8c\x03CMR\x94h\r\x8c\x08Cameroon\x94h\x0fJEd}\x01uh\x10(G@ \xfaF\x07\xf9\x9a\x00G?\xfb\xa4\x8c\x0ff\xd9?G@0\x03JK\xab\x02\x90G@)\xb8\x02\xc9\x9bQ\x91t\x94h\x12]\x94}\x94(h\x04h\x15h\x16(G@,\xfd\xd7\xd80k\xe0G@)\xb8\x02\xc9\x9bQ\x91\x86\x94G@-\xc9f{_\x1c\x00G@(p'RT`\xa0\x86\x94G@-\xeb\x99\x04\x90\x02\xb0G@'\x1ct3k|\xb9\x86\x94G@-\xd8\xdd\x7fqJ\x10G@%\xc8[\xc61p\x08\x86\x94G@.\xef\x8d\x04\x90\x07 G@#\xf6\xf4\xd7By`\x86\x94G@-\xd1\x96\xd4\xc6\x9fQG@#\xfb\xf8c4\xd1\xf0\x86\x94G@-A n`6`G@#\xd7\x82\xbb\xf3\xf6\xe0\x86\x94G@,W\xcag\x8c\x91\xf0G@$\n\xf2\x18\x1c\xed\x80\x86\x94G@+\xe8\x8fO\xa7\xde\x90G@#\x19Wf\x9e\x9e@\x86\x94G@-\x16\xc4S\x11\xb3\xf0G@!\xee\x85_\xca\xf8\xb0\x86\x94G@-\xf5\xc1\xfaR\x92?G@!\x97\x9a\xf9d\x90\xd0\x86\x94G@.=\xe2\x15\xa1\x15 G@ \xc3\xa90\x01\x90\xa0\x86\x94G@.\xdfGj\xf6m@G@\x1e\xc5p\x9ds\xb8\x80\x86\x94G@.\x8f\x15q\xca\x0c`G@\x1d\xb0\x0c\xfd\x06z\x10\x86\x94G@-\x8d\x97]O&\xa0G@\x19\xa2MN\xf1\xed`\x86\x94G@-\x12\xb8\x01&.\xc0G@\x18\xe8g\xdeN\x13 \x86\x94G@,\xeb7j\xf6d\xb0G@\x15\xce\x9aR[\xac\xc0\x86\x94G@-\x1e,\xdb\x9a<\xa0G@\x14\x1fT\xfd\x06P\x00\x86\x94G@,\xf4\xed;-\x01\xe0G@\x12\xee0\"\x92= \x86\x94G@-\xe6\xe3]O(0G@\x10\xd7pK\x87\xf6\x80\x86\x94G@.\x12\x8b]O(\xf0G@\x0e\xcf\x99\xa8 \xf1`\x86\x94G@.\xcf\x90\x0e\xcdw@G@\n\xae\xb2\x15Z\xe6 \x86\x94G@/\xb9\xb8\x0e\xcd{QG@\x08\x1b\xb9q\x83\xd0\x80\x86\x94G@/\xd0\x947\xc3>@G@\x04u\x88\x97\x0f\xb2\xc1\x86\x94G@0\x03JK\xab\x02\x90G@\x02$ E$#\x80\x86\x94G@/\xe1\xc0\x1ct\xbc\xd0G?\xfb\xa4\x8c\x0ff\xd9?\x86\x94G@.J\xedV{\x89\x80G?\xffl\x9a\xc7\xb8\xd5\x80\x86\x94G@,\xac\xf5\xc3\xb5\x890G@\x01\xd2\xaf\xf38\x9d\x00\x86\x94G@*&\xd20\xef\x84\xf0G@\x02#\x03\xcaB\xdb\xc0\x86\x94G@)\xe7\x15;,\xf4pG@\x02\x92\xabA\xbaU \x86\x94G@(\xb8\x00\xb2\xa4f\xbfG@\x01\x8a\xe1\x1f\x98.\x80\x86\x94G@'\x80\xdaL=\xfb\x00G@\x02\x9d3\x0b\x1dR\x00\x86\x94G@&\x8d\x8a\xb9w\xfd\xa0G@\x02\x16\xa1\xde\xbd\xbc\xe0\x86\x94G@#L^u3\xab\x10G@\x02E[\x93\xa5\xd9\x00\x86\x94G@#\x97#\xe5\xd7\x83`G@\x08\x96U\x11\xf1\x0c\x9f\x86\x94G@\"\xcf\t-\x85\x94`G@\r\xe0O\xa1ML\xe0\x86\x94G@!\xe5oj\xf64@G@\x0f;\xa7\xf38\xd8\x00\x86\x94G@!}f\xab\xd0\xa6\x80G@\x11h\xab\x1f(f0\x86\x94G@ \xfaF\x07\xf9\x9a\x00G@\x11\xfb\x83\x1f(h\xc0\x86\x94G@!\x00%\xb6\x0e\x15\x00G@\x13\x16\x82\xb1\xeef\xc0\x86\x94G@!\x83\xdb]N\xf1\xb0G@\x15\xeb-\x85\x8e\xe0p\x86\x94G@\"wa\x1f\xdeR\x10G@\x19\xc7(\x8f\xccb\x10\x86\x94G@#\x0b\xa0\x1ct\x84`G@\x19\xd0]\xae\x84\xb4\x1f\x86\x94G@$<\x8e\xc7\x1f4PG@\x1c'\xb3:w\x170\x86\x94G@$\xfe\xa7\xf3~\xca`G@\x1c8\xaf\xb5X_ \x86\x94G@&\x1e\x19q\xc9\xe7@G@\x1a\x93\xe4\x9ds\xa6p\x86\x94G@'}\xd6#H8`G@\x1b\xec\xef\xa7\xb1\x1d\x00\x86\x94G@'\xad\xb9\xdf\x03\xf4\xefG@\x1d\x96\x92D\xb4\x8e \x86\x94G@( \xbd\x8d\x18q\xc0G@\x1f3\x00\xfd\x06\x80\xc0\x86\x94G@(p\x10\x01&\x1a`G@ \x9c\x94\xf9d\x8c\x80\x86\x94G@)\x81\xe1;,\xf2\xc0G@!o~\x99\xd1\xca(\x86\x94G@)\xe93\x19\n\xd2`G@\"\xd5\xe6,\x97\xc9x\x86\x94G@*U\xcf\xa1\x93\\\xc0G@#H\x00,\x97\xcbp\x86\x94G@*\x9e\n\xd4\xc6\x911G@$R\x1a\xf9d\x9c\xcf\x86\x94G@+%Y\xa8g\x00\xc0G@%\x98\xdd\xa4\x0fM\x18\x86\x94G@,\xd4\xac\x89\xae\xb6?G@'%\r\x88\xc0\xd20\x86\x94G@,\xef\xb6\xe2m\xdc@G@'\xcf;\x9d;\xb6h\x86\x94G@-'\x83\xd80l\x90G@(+\xb4j\x08\x84\xc8\x86\x94G@,\\\xd8\x1ct\xad_G@(\xf7\xa1\xe1\x7f\xff\xc0\x86\x94G@,mS\xe5\xd7\xaa@G@)\x9a\xa4c4\xea\xa8\x86\x94G@,\xfd\xd7\xd80k\xe0G@)\xb8\x02\xc9\x9bQ\x91\x86\x94t\x94\x85\x94uauK\x13}\x94(h\x02\x8c\x0258\x94h\x04h\x05h\x06}\x94(h\x08\x8c\x06Africa\x94h\nG@\xc6\xad\x00\x00\x00\x00\x00h\x0b\x8c\x03TGO\x94h\r\x8c\x04Togo\x94h\x0fJ\x7f\x89y\x00uh\x10(G\xbf\xa9}a\xd6\xe7\xbf\xffG@\x17\xb7!%\xfc\"`G?\xfd\xd8\x06o\x94a\x00G@&\t\x90\xa7y\x1f@t\x94h\x12]\x94}\x94(h\x04h\x15h\x16(G?\xec\xc98dF\xf2\x00G@%\xfe\xa3D|\x88\xe0\x86\x94G?\xe8\xb6\xf9>\xba\xee\x00G@$\xf1\r\xc61lX\x86\x94G?\xf1>\xa6\x02Z\"\x80G@$Y\xe9\x1b\x86\xbf\x10\x86\x94G?\xf6\xcd\x0ch\xc0\xa1\x81G@#\xa6\x9a:?\r\xe8\x86\x94G?\xf7h\x9f\x9b\xf3\xd7\x80G@\"\xabS\xe1\x7f\xe4\x10\x86\x94G?\xfa\xa1\xb3<a\x1f\x80G@\"A\xd6\x99\xd1\xcd\xc0\x86\x94G?\xf9\xe78\xc8Su\x00G@\x1bT\x01\xc9\xd3<\x80\x86\x94G?\xfd\xd8\x06o\x94a\x00G@\x18\x91\x91\xc9\xd30`\x86\x94G?\xf0\xf6B+O\xe4\x00G@\x17\xb7!%\xfc\"`\x86\x94G?\xea\xc8#\xe9e\xa2\x00G@\x19\x1e\xb2\xbf\x95\xc2\x1f\x86\x94G?\xe2@\x96B$\xa2\x00G@\x1b\xa8M\xa0\xdd{`\x86\x94G?\xdfk\xd8\xe3\xdb\xf3\xffG@\x1d\xa5\xa0K\x88.\xd0\x86\x94G?\xe6\xc8\xf1\x90\xa6k\x00G@ \x9f\xfbX\xf7R\x88\x86\x94G?\xdd\x84*\xcfa\n\x00G@!Z\xbc\xeb\xbdN\xe8\x86\x94G?\xd7j\xe9\xf4\xec\xe0\x00G@\"\xee\x15\x0087 \x86\x94G?\xd7\x86n91&\x00G@$a\xe6\xa7y\x17\xf8\x86\x94G\xbf\xa9}a\xd6\xe7\xbf\xffG@%i\xf1\"Zd0\x86\x94G?\x98_\xb0_\xd9\x1f\xffG@&\t\x90\xa7y\x1f@\x86\x94G?\xec\xc98dF\xf2\x00G@%\xfe\xa3D|\x88\xe0\x86\x94t\x94\x85\x94uauK\x14}\x94(h\x02\x8c\x0259\x94h\x04h\x05h\x06}\x94(h\x08\x8c\x06Africa\x94h\nG@\xfd~\x00\x00\x00\x00\x00h\x0b\x8c\x03GHA\x94h\r\x8c\x05Ghana\x94h\x0fJ\x94\x9d\xa3\x01uh\x10(G\xc0\t\xf4xMU;\xc0G@\x12\xd7\x83cl\xb0\xc0G?\xf0\xf6B+O\xe4\x00G@&2Y\xbf]\xd1ot\x94h\x12]\x94}\x94(h\x04h\x15h\x16(G?\x98_\xb0_\xd9\x1f\xffG@&\t\x90\xa7y\x1f@\x86\x94G\xbf\xa9}a\xd6\xe7\xbf\xffG@%i\xf1\"Zd0\x86\x94G?\xd7\x86n91&\x00G@$a\xe6\xa7y\x17\xf8\x86\x94G?\xd7j\xe9\xf4\xec\xe0\x00G@\"\xee\x15\x0087 \x86\x94G?\xdd\x84*\xcfa\n\x00G@!Z\xbc\xeb\xbdN\xe8\x86\x94G?\xe6\xc8\xf1\x90\xa6k\x00G@ \x9f\xfbX\xf7R\x88\x86\x94G?\xdfk\xd8\xe3\xdb\xf3\xffG@\x1d\xa5\xa0K\x88.\xd0\x86\x94G?\xe2@\x96B$\xa2\x00G@\x1b\xa8M\xa0\xdd{`\x86\x94G?\xea\xc8#\xe9e\xa2\x00G@\x19\x1e\xb2\xbf\x95\xc2\x1f\x86\x94G?\xf0\xf6B+O\xe4\x00G@\x17\xb7!%\xfc\"`\x86\x94G\xbf\xe0>\x91\xd9*x\x00G@\x15_\xb7H\x1e:0\x86\x94G\xbf\xf1\x04\x9bET\x88\x80G@\x14\x00\x8f\x9a\t\xb9P\x86\x94G\xbf\xffop-p\x16\x80G@\x12\xd7\x83cl\xb0\xc0\x86\x94G\xc0\x06\xd9X\x16\xb8*\x80G@\x13\xfaW\xdeM\xfdp\x86\x94G\xc0\x06|Q\x0cz\xb8AG@\x15\x8ecq\x13\xfd\x8f\x86\x94G\xc0\t\xf4xMU;\xc0G@\x19\x00{\x9a\t\xcfP\x86\x94G\xc0\x07\xdea\xcb\xa0J\x80G@\x1d\x84\xd1\\\x99?P\x86\x94G\xc0\x04\x7f]5pr@G@ ps\r\xdfm\x08\x86\x94G\xc0\x06\x9e\xb6a\xd0\x0e@G@#H\xf0\xa0\xa5r\xb8\x86\x94G\xc0\x07\xb6\x0f<D \xc0G@$\xcaiR#\xc4p\x86\x94G\xc0\x07\x85\xf5P\xbf\x01@G@%\xec\xe5\xbf]\xd0@\x86\x94G\xbf\xf3@\xf4\x04z\x1e\x80G@&\x05\x07\x07\x0b\xe5 \x86\x94G\xbf\xe8^\xd4h\x86\xc3\xffG@%\xdf\xb5=\xa8\xe7\xe8\x86\x94G\xbf\xdc\x13\xaf\xa4\xad\x9c\x00G@&2Y\xbf]\xd1o\x86\x94G?\x98_\xb0_\xd9\x1f\xffG@&\t\x90\xa7y\x1f@\x86\x94t\x94\x85\x94uauK\x15}\x94(h\x02\x8c\x0260\x94h\x04h\x05h\x06}\x94(h\x08\x8c\x06Africa\x94h\nG@\xf5E\x00\x00\x00\x00\x00h\x0b\x8c\x03CIV\x94h\r\x8c\x0eC\xc3\xb4te d'Ivoire\x94h\x0fJ\xea\x07q\x01uh\x10(G\xc0!4\xac\xb7,\x88HG@\x11ZhK\x87\xf8\xc0G\xc0\x04\x7f]5pr@G@%\x0cQ\xb1\xb6\x8b\x88t\x94h\x12]\x94}\x94(h\x04h\x15h\x16(G\xc0 \x0fT\xc4\xd3\xc4\x18G@$i\xbe\xf2\x90\xfc\xc8\x86\x94G\xc0\x1f\x99.\x120\x0eqG@$\x98Bw\xaf\xb5\xe7\x86\x94G\xc0\x1e}\xb4\x93\xe4\xf1\xb1G@$Kb\x8c*\x95\xe0\x86\x94G\xc0\x1bf\xeb0\xe8M\xf0G@$G*3kp@\x86\x94G\xc0\x1a\xaat\xbc\xda\xa3pG@$\xdc\x936\xd5C\x08\x86\x94G\xc0\x19\xf9\xd1\xf6\xe1t\x00G@$\xd2\x96G\xe6S\xf0\x86\x94G\xc0\x18\xd2%\xf6\xe1n\xf1G@%\x0cQ\xb1\xb6\x8b\x88\x86\x94G\xc0\x183\xa9\xb2\x9d'\xf0G@$1V3ko\xe0\x86\x94G\xc0\x17D\x88O\xa0\x8d\xa0G@$q\xf2\xaeL\xb8\xa8\x86\x94G\xc0\x15\x9e\x0b\xb9p\xbd\x00G@$\xbd\xd16\xd5B\x81\x86\x94G\xc0\x13\xd1\x90\xa1\x8c\x03\xa0G@$N0\x85V\xf5\x80\x86\x94G\xc0\x13\x1e\x99\xcd\xeb\x93@G@#\xa4\xdb0\x01\x9d@\x86\x94G\xc0\x11R,A\xf92\xa0G@#8\xbfX\xf7]\xf0\x86\x94G\xc0\x0f\xd7\xf5\xbd\xf9,\x80G@#\xb9\x85)-\xfd0\x86\x94G\xc0\x0c\x18^}\x1e\xa7\xffG@#\xcc\xf7\x8f\x94c\xe8\x86\x94G\xc0\x06\x9e\xb6a\xd0\x0e@G@#H\xf0\xa0\xa5r\xb8\x86\x94G\xc0\x04\x7f]5pr@G@ ps\r\xdfm\x08\x86\x94G\xc0\x07\xdea\xcb\xa0J\x80G@\x1d\x84\xd1\\\x99?P\x86\x94G\xc0\t\xf4xMU;\xc0G@\x19\x00{\x9a\t\xcfP\x86\x94G\xc0\x06|Q\x0cz\xb8AG@\x15\x8ecq\x13\xfd\x8f\x86\x94G\xc0\x06\xd9X\x16\xb8*\x80G@\x13\xfaW\xdeM\xfdp\x86\x94G\xc0\n}\x19\xcb\xa0V\x00G@\x13\xef\xeb,\xcf\xb20\x86\x94G\xc0\x10\t\x07\xfd\xb4\xe8\xc0G@\x14\xb8 \xfd\x06R\xa0\x86\x94G\xc0\x12\x99\x83\xf0\r\xb3 G@\x14\xacMN\xf1\xd7\x90\x86\x94G\xc0\x17V\x86-~k\xd0G@\x13\xf9\x8c\xab\x1a\xca@\x86\x94G\xc0\x1a\x1du\xa4\xf5\xef\x80G@\x12\xd2\x02\x88\xf8\xa3\x00\x86\x94G\xc0\x1e\x13eS\n{\xd0G@\x11ZhK\x87\xf8\xc0\x86\x94G\xc0\x1e\xd9@O\xa0\xaf\x00G@\x11uP\xc6i@\xe0\x86\x94G\xc0\x1e\x8a\x9d\xf6\xe1\x88\x1fG@\x14\xc0\xac\xc6iO`\x86\x94G\xc0\x1e(\xab\x15\x99\xd8PG@\x15@\xdd\x936\x1e`\x86\x94G\xc0\x1eG\xd6\x120\x08\xb0G@\x16\xd4T\"\x92N@\x86\x94G\xc0\x1f\xf9\x8a\x8d\x11W\xd0G@\x18\x817\xdeN\x11`\x86\x94G\xc0 \x9fh\xf4\x9d)\x90G@\x18\xc5\xaa{Q|`\x86\x94G\xc0!4\xac\xb7,\x88HG@\x19\xde\xc9%\xfc+\xdf\x86\x94G\xc0 \xc5Y\xea_\xb9\x98G@\x1b\xa5\xaf\x11\x81R`\x86\x94G\xc0 \xf8\x8cP\xc6 \xe0G@\x1d\x94\xb1\\\x99?\x90\x86\x94G\xc0 \xe0\xeb\xba\x96W\x10G@\x1e\xbe\x82\x00pN\xf0\x86\x94G\xc0 \x8f\xb8^m_\xe8G@\x1e\xbf\xac\x07C\xef`\x86\x94G\xc0 q\x8e\xc4\xd3\xc5\xc8G@ ?$\xf2\x90\xeax\x86\x94G\xc0 \x99\x1c\xe6\xf5\xe8\xa0G@ \xa2\x04\xe4\xe9\xabP\x86\x94G\xc0 h1\x02DixG@ \xe91)-\xf0\xd0\x86\x94G\xc0\x1fT\x12\x1f\xd7N\x1fG@!&\xc2\xb5 J\x98\x86\x94G\xc0 (\x81\x98t1\xc8G@\"\xc0\xa0j\x08l\xf0\x86\x94G\xc0 \x9e\x86\x0c\x81\xdb\x10G@#\x94=\x88\xc0\xc2\x80\x86\x94G\xc0 uk\xac\xef\x14`G@$B\x0e\xf2\x90\xfc\x18\x86\x94G\xc0 \x0fT\xc4\xd3\xc4\x18G@$i\xbe\xf2\x90\xfc\xc8\x86\x94t\x94\x85\x94uauK\x16}\x94(h\x02\x8c\x0261\x94h\x04h\x05h\x06}\x94(h\x08\x8c\x06Africa\x94h\nG@\xcfh\x00\x00\x00\x00\x00h\x0b\x8c\x03GIN\x94h\r\x8c\x06Guinea\x94h\x0fJ\xabk\xbd\x00uh\x10(G\xc0.B\xb8'\xd0\x98\xc0G@\x1d<tK\x88-\x00G\xc0\x1fT\x12\x1f\xd7N\x1fG@), ,\x97\xe5Yt\x94h\x12]\x94}\x94(h\x04h\x15h\x16(G\xc0+f\xa4\xcb\xa7\x96hG@), ,\x97\xe5Y\x86\x94G\xc0*o\x85\xdc\xb8\xa3@G@)&\xd8\xe4\xe9\xd0\xc8\x86\x94G\xc0(\xff\x83\x91\xa0\xb88G@(\xaa\x07\xb1\xb6\x9bq\x86\x94G\xc0(\x8e\xa4\x87cE\xa8G@(\xb5y).\x13\x18\x86\x94G\xc0(h9\xa6\x1b\x96\xe7G@(\xeei_\xcb\x17\x80\x86\x94G\xc0'Q\x0c\xd2{$\xe1G@(\xc5\xee,\x97\xe3\x99\x86\x94G\xc0'\x07#\x83\xf9n\xb8G@(\xe2\xcfD|\x95\x98\x86\x94G\xc0&\xe9\x8e\xed\xc9\xa4\xd0G@('V\xd0n\xeb\x18\x86\x94G\xc0&\x98[\x91\xa0\xad\xa8G@('\xeb\xd3\xd8\xbbX\x86\x94G\xc0&\x12\xb7vR)\xa8G@(l(A\x12\xc3Y\x86\x94G\xc0%\xbd\xdda\xd7F\xe7G@([\x14\x11I`\x11\x86\x94G\xc0%/\xbb\t\x18\x1e\xf0G@'\xd9\x13KP1p\x86\x94G\xc0$T\x96\xe6\xf5\xf9\x08G@'\xb0+\xb8\x8a7\x90\x86\x94G\xc0#\xc805w\xab\x88G@(\x1e\xf7\r\xdf\x8e\xd1\x86\x94G\xc0#\"\xc5T/\xfa\x98G@(cs\xd3\xd8\xbcX\x86\x94G\xc0\"\xa7\xbdT/\xf8\x80G@(\xab'\x8f\x94yP\x86\x94G\xc0\"AD5w\xa4\xd0G@(\x9d\xba\x18\x1d\x01\xa0\x86\x94G\xc0!\xcf~\xe0\"M\x88G@(-=D|\x92y\x86\x94G\xc0!\x92{\x91\xa0\x97\x90G@'\xa0\x07\xfc\xce{\x90\x86\x94G\xc0 \xc0\xab\t\x18\x0bpG@&\xc9\x8b\xf5\xfa\xd7p\x86\x94G\xc0!)\xa0\xd9N\xaa@G@&E\xc1\xfc\xceu\x98\x86\x94G\xc0!=\x9a\xb7,\x88pG@%\x9f-\x14\xb3$A\x86\x94G\xc0 \xd0\x8b\t\x18\x0b\xb0G@%\xd1\x8a\x1e\xf0\x95\xc1\x86\x94G\xc0 \x90\x91\x1d\x92\xeb\xe0G@%\x95\xcfR#\xc7\xe7\x86\x94G\xc0 \xab\xb6\x8e6\xc3`G@$\xfdW\xfc\xceo\xf8\x86\x94G\xc0 \x0fT\xc4\xd3\xc4\x18G@$i\xbe\xf2\x90\xfc\xc8\x86\x94G\xc0 uk\xac\xef\x14`G@$B\x0e\xf2\x90\xfc\x18\x86\x94G\xc0 \x9e\x86\x0c\x81\xdb\x10G@#\x94=\x88\xc0\xc2\x80\x86\x94G\xc0 (\x81\x98t1\xc8G@\"\xc0\xa0j\x08l\xf0\x86\x94G\xc0\x1fT\x12\x1f\xd7N\x1fG@!&\xc2\xb5 J\x98\x86\x94G\xc0 h1\x02DixG@ \xe91)-\xf0\xd0\x86\x94G\xc0 \x99\x1c\xe6\xf5\xe8\xa0G@ \xa2\x04\xe4\xe9\xabP\x86\x94G\xc0 q\x8e\xc4\xd3\xc5\xc8G@ ?$\xf2\x90\xeax\x86\x94G\xc0 \x8f\xb8^m_\xe8G@\x1e\xbf\xac\x07C\xef`\x86\x94G\xc0 \xe0\xeb\xba\x96W\x10G@\x1e\xbe\x82\x00pN\xf0\x86\x94G\xc0!q\xba.\xa4\x00\xd0G@\x1e\xd8\xc1%\xfcA\xc0\x86\x94G\xc0!\xda%$f\x91\xf8G@\x1d<tK\x88-\x00\x86\x94G\xc0\"j\xe6\x0c\x81\xe2\xf8G@\x1dAtt}\xef\xa0\x86\x94G\xc0\"\xce\x83\xa6\x1b~HG@\x1e\x1b\x8d\n\xad\xbc\xc0\x86\x94G\xc0\"\xac\xaf\xf13b`G@\x1f\xb6\xd1\xbc,\x0e\xf0\x86\x94G\xc0#\x82\xbc5w\xaaPG@!\x15\x050\x01\x91\xf8\x86\x94G\xc0$\x08{h\xaa\xdf\xd0G@ \xdbd\xde\x16\x0b\xe0\x86\x94G\xc0$u\xce\xd2{\x18PG@ \xcf\xfa,\x97\xc0\x98\x86\x94G\xc0%\x02\xcd\xea_\xcc8G@ \xb2\xa2\x8c*\x86\x10\x86\x94G\xc0$\xfd\x16\xe0\"[\x80G@!n[X\xf7V\x10\x86\x94G\xc0%O>\x13U\x90\x18G@!\xf4P\xbb\xf3\xee\x90\x86\x94G\xc0%>\xaa\x95\nw\xefG@\"\x89+\x81\xed\x1d\x80\x86\x94G\xc0%\xad\xa5T0\x05\xc8G@#`a\xcd\x05\x05\xe0\x86\x94G\xc0&<&\x80\x8f\x9b\x00G@$\x17|\xa7y\x16\xb0\x86\x94G\xc0'\xd5\xa5a\xd7P G@$\x18\x0eG\xe6P\xb8\x86\x94G\xc0(L\xf9\x1d\x93\r\xe7G@#\xb7\x96\xb5 U\xe0\x86\x94G\xc0(\xda\x13M\\sPG@#\xab\xf2p\xdc\x11h\x86\x94G\xc0)1\x85+:R\xb0G@#=\x89R#\xbd\x98\x86\x94G\xc0)l\x85\xb3\xc2\xdc8G@\"\xafw\xe8S\x84\x90\x86\x94G\xc0*~;\xd5\xe5\x03\x11G@!\xce\\c4\xc8`\x86\x94G\xc0+^\xcc\x80\x8f\xb1\x98G@\"\xfdO\r\xdfx@\x86\x94G\xc0,%\xe98\xe1\xa0\x87G@#\xc5\xb7\xaa\xe2\xe5\x80\x86\x94G\xc0,\xa8\xff\xb3\xc2\xeawG@$\x08\x0cj\x08r\x90\x86\x94G\xc0-(\xceI\xf2\xb6\x08G@$m\xce\xa7y\x18(\x86\x94G\xc0-b\xefM\\\x87?G@%P\x06\xa7y\x1c\x10\x86\x94G\xc0-\xad\xd9\xfe\xda\xd3\xa0G@%\xc0\xcd\xfc\xcesO\x86\x94G\xc0.B\xb8'\xd0\x98\xc0G@&\x14\xb0\xd7B\x82h\x86\x94G\xc0-_\x12eA8\xb0G@'\x0e>\xeb\xbdh\x00\x86\x94G\xc0,\xc3\xae\x9b\xde9pG@'\x04\xbfKP-\xc8\x86\x94G\xc0,>(\xfbp\xfd\x18G@'Z\xaf\x14\xb3+\xd8\x86\x94G\xc0+\xcd5\x9fH\x05gG@'[\x81\r\xdf\x8bp\x86\x94G\xc0+|\x7f\x91\xa0\xc3(G@'\x9f^\xa7y&9\x86\x94G\xc0+\xa8\x13?\xb5>\xc7G@(I\x08\xa7y) \x86\x94G\xc0+o\xff+:\\\x90G@(~\x8f\x1b\x86\xd1P\x86\x94G\xc0+f\xa4\xcb\xa7\x96hG@), ,\x97\xe5Y\x86\x94t\x94\x85\x94uauK\x17}\x94(h\x02\x8c\x0262\x94h\x04h\x05h\x06}\x94(h\x08\x8c\x06Africa\x94h\nG@\xa6F\x00\x00\x00\x00\x00h\x0b\x8c\x03GNB\x94h\r\x8c\rGuinea-Bissau\x94h\x0fJRY\x1b\x00uh\x10(G\xc00\xadm}\xb8\x89\xccG@&\x14\xb0\xd7B\x82hG\xc0+f\xa4\xcb\xa7\x96hG@)A\x9f\x81\xed;\x10t\x94h\x12]\x94}\x94(h\x04h\x15h\x16(G\xc00\xadm}\xb8\x89\xccG@(\xc5\x0bD|\x95\x10\x86\x94G\xc00%\xd0\xc5f\x9b\xf4G@)\x18t3k\x85p\x86\x94G\xc0/\xa2\x16\x05\xae|\xa9G@)\x07\xf8j\x08\x88\x90\x86\x94G\xc0/\x18\xd1\xf8\x079pG@)A\x9f\x81\xed;\x10\x86\x94G\xc0+f\xa4\xcb\xa7\x96hG@), ,\x97\xe5Y\x86\x94G\xc0+o\xff+:\\\x90G@(~\x8f\x1b\x86\xd1P\x86\x94G\xc0+\xa8\x13?\xb5>\xc7G@(I\x08\xa7y) \x86\x94G\xc0+|\x7f\x91\xa0\xc3(G@'\x9f^\xa7y&9\x86\x94G\xc0+\xcd5\x9fH\x05gG@'[\x81\r\xdf\x8bp\x86\x94G\xc0,>(\xfbp\xfd\x18G@'Z\xaf\x14\xb3+\xd8\x86\x94G\xc0,\xc3\xae\x9b\xde9pG@'\x04\xbfKP-\xc8\x86\x94G\xc0-_\x12eA8\xb0G@'\x0e>\xeb\xbdh\x00\x86\x94G\xc0.B\xb8'\xd0\x98\xc0G@&\x14\xb0\xd7B\x82h\x86\x94G\xc0/T\x0fvRR`G@&\xea\xbd\x1b\x86\xcaa\x86\x94G\xc00\x15\xd0\x99\x07\x08\xe8G@'\x0c\x97\x96h\x12\xa0\x86\x94G\xc00P\x95\xddKN,G@'\x9c\xef\x81\xed3\xd0\x86\x94G\xc00O\x17+\xcd\x03\x10G@'\xea\xda\xf9d\xac\xa0\x86\x94G\xc00\x9d$\x81\"Y\xbcG@(W\x81\xaa\xe2\xf9\x98\x86\x94G\xc00\xadm}\xb8\x89\xccG@(\xc5\x0bD|\x95\x10\x86\x94t\x94\x85\x94uauK\x18}\x94(h\x02\x8c\x0263\x94h\x04h\x05h\x06}\x94(h\x08\x8c\x06Africa\x94h\nG@\xaeR\x00\x00\x00\x00\x00h\x0b\x8c\x03LBR\x94h\r\x8c\x07Liberia\x94h\x0fJ}\x8cG\x00uh\x10(G\xc0&\xe0\xa7\xb3\xc2\xd1\x08G@\x11lK\x11\x81%pG\xc0\x1e(\xab\x15\x99\xd8PG@!\x15\x050\x01\x91\xf8t\x94h\x12]\x94}\x94(h\x04h\x15h\x16(G\xc0 \xe0\xeb\xba\x96W\x10G@\x1e\xbe\x82\x00pN\xf0\x86\x94G\xc0 \xf8\x8cP\xc6 \xe0G@\x1d\x94\xb1\\\x99?\x90\x86\x94G\xc0 \xc5Y\xea_\xb9\x98G@\x1b\xa5\xaf\x11\x81R`\x86\x94G\xc0!4\xac\xb7,\x88HG@\x19\xde\xc9%\xfc+\xdf\x86\x94G\xc0 \x9fh\xf4\x9d)\x90G@\x18\xc5\xaa{Q|`\x86\x94G\xc0\x1f\xf9\x8a\x8d\x11W\xd0G@\x18\x817\xdeN\x11`\x86\x94G\xc0\x1eG\xd6\x120\x08\xb0G@\x16\xd4T\"\x92N@\x86\x94G\xc0\x1e(\xab\x15\x99\xd8PG@\x15@\xdd\x936\x1e`\x86\x94G\xc0\x1e\x8a\x9d\xf6\xe1\x88\x1fG@\x14\xc0\xac\xc6iO`\x86\x94G\xc0\x1e\xd9@O\xa0\xaf\x00G@\x11uP\xc6i@\xe0\x86\x94G\xc0\x1f\xe5|]G\xf4pG@\x11lK\x11\x81%p\x86\x94G\xc0\"\x02tP\xc6%pG@\x13Te\x85\x8e\xd5\x10\x86\x94G\xc0#\xd3\xab\xd5\xe4\xe5\xc0G@\x16_\xce`\x02\xf0 \x86\x94G\xc0%\x87\xe0eA\x16/G@\x18\x90\x16{Q{p\x86\x94G\xc0&\xe0\xa7\xb3\xc2\xd1\x08G@\x1b$\xc7cl\xd5Q\x86\x94G\xc0&fLl\x14\xbaoG@\x1clb\xcd=\x11\x90\x86\x94G\xc0&K\x1c\xd2{ `G@\x1d\x96:7\rM@\x86\x94G\xc0%d%\x02D\x7fhG@\x1f\xc2\x02\xda\xe4a\x10\x86\x94G\xc0$u\xce\xd2{\x18PG@ \xcf\xfa,\x97\xc0\x98\x86\x94G\xc0$\x08{h\xaa\xdf\xd0G@ \xdbd\xde\x16\x0b\xe0\x86\x94G\xc0#\x82\xbc5w\xaaPG@!\x15\x050\x01\x91\xf8\x86\x94G\xc0\"\xac\xaf\xf13b`G@\x1f\xb6\xd1\xbc,\x0e\xf0\x86\x94G\xc0\"\xce\x83\xa6\x1b~HG@\x1e\x1b\x8d\n\xad\xbc\xc0\x86\x94G\xc0\"j\xe6\x0c\x81\xe2\xf8G@\x1dAtt}\xef\xa0\x86\x94G\xc0!\xda%$f\x91\xf8G@\x1d<tK\x88-\x00\x86\x94G\xc0!q\xba.\xa4\x00\xd0G@\x1e\xd8\xc1%\xfcA\xc0\x86\x94G\xc0 \xe0\xeb\xba\x96W\x10G@\x1e\xbe\x82\x00pN\xf0\x86\x94t\x94\x85\x94uauK\x19}\x94(h\x02\x8c\x0264\x94h\x04h\x05h\x06}\x94(h\x08\x8c\x06Africa\x94h\nG@\xc4\xc8\x00\x00\x00\x00\x00h\x0b\x8c\x03SLE\x94h\r\x8c\x0cSierra Leone\x94h\x0fJ\xfb\n^\x00uh\x10(G\xc0*~;\xd5\xe5\x03\x11G@\x1b$\xc7cl\xd5QG\xc0$u\xce\xd2{\x18PG@$\x18\x0eG\xe6P\xb8t\x94h\x12]\x94}\x94(h\x04h\x15h\x16(G\xc0*~;\xd5\xe5\x03\x11G@!\xce\\c4\xc8`\x86\x94G\xc0)l\x85\xb3\xc2\xdc8G@\"\xafw\xe8S\x84\x90\x86\x94G\xc0)1\x85+:R\xb0G@#=\x89R#\xbd\x98\x86\x94G\xc0(\xda\x13M\\sPG@#\xab\xf2p\xdc\x11h\x86\x94G\xc0(L\xf9\x1d\x93\r\xe7G@#\xb7\x96\xb5 U\xe0\x86\x94G\xc0'\xd5\xa5a\xd7P G@$\x18\x0eG\xe6P\xb8\x86\x94G\xc0&<&\x80\x8f\x9b\x00G@$\x17|\xa7y\x16\xb0\x86\x94G\xc0%\xad\xa5T0\x05\xc8G@#`a\xcd\x05\x05\xe0\x86\x94G\xc0%>\xaa\x95\nw\xefG@\"\x89+\x81\xed\x1d\x80\x86\x94G\xc0%O>\x13U\x90\x18G@!\xf4P\xbb\xf3\xee\x90\x86\x94G\xc0$\xfd\x16\xe0\"[\x80G@!n[X\xf7V\x10\x86\x94G\xc0%\x02\xcd\xea_\xcc8G@ \xb2\xa2\x8c*\x86\x10\x86\x94G\xc0$u\xce\xd2{\x18PG@ \xcf\xfa,\x97\xc0\x98\x86\x94G\xc0%d%\x02D\x7fhG@\x1f\xc2\x02\xda\xe4a\x10\x86\x94G\xc0&K\x1c\xd2{ `G@\x1d\x96:7\rM@\x86\x94G\xc0&fLl\x14\xbaoG@\x1clb\xcd=\x11\x90\x86\x94G\xc0&\xe0\xa7\xb3\xc2\xd1\x08G@\x1b$\xc7cl\xd5Q\x86\x94G\xc0'j\x98y\xbb\xff\xc8G@\x1bp\xbd\xa0\xddzp\x86\x94G\xc0(\xdb/\xc8=\xbb\x00G@\x1d\r@\xab\x1a\xf20\x86\x94G\xc0)\xe5\xe9\xc1j\x1f(G@\x1f1\xd009\xb3\xf1\x86\x94G\xc0*?\x80C\x1f\x08\xd0G@ S\xf0\xc9\x9b(@\x86\x94G\xc0*~;\xd5\xe5\x03\x11G@!\xce\\c4\xc8`\x86\x94t\x94\x85\x94uauK\x1a}\x94(h\x02\x8c\x0265\x94h\x04h\x05h\x06}\x94(h\x08\x8c\x06Africa\x94h\nG@\xe0\x1b\xc0\x00\x00\x00\x00h\x0b\x8c\x03BFA\x94h\r\x8c\x0cBurkina Faso\x94h\x0fJ\xf5\xd02\x01uh\x10(G\xc0\x15\xe1\xdb\xc7\x17\xff\x00G@#8\xbfX\xf7]\xf0G@\x01j\xb7|\x0e\x7f\xc0G@.;y\x07\x0c\t@t\x94h\x12]\x94}\x94(h\x04h\x15h\x16(G\xc0\x15\x9e\x0b\xb9p\xbd\x00G@$\xbd\xd16\xd5B\x81\x86\x94G\xc0\x15\xe1\xdb\xc7\x17\xff\x00G@%\xe7\x0c\xd7B\x81\xa1\x86\x94G\xc0\x14\xca\x97>\x8fq\xb0G@&\xc0\x13\x1b\x86\xc9\xa0\x86\x94G\xc0\x14\xe2>\x9a\xb8g\xe0G@'m~\xeb\xbdi\xa0\x86\x94G\xc0\x11\xb5k\x07\xf2`\xc0G@)\x15\xd5\xa4\x0f\\q\x86\x94G\xc0\x11\x1f\"\x7fi\xd5\xa0G@*t\xf6\x8c*\xb0\xf8\x86\x94G\xc0\x10\x06\x8bL6\x9d\xa0G@*\xf1\xe9\x9d;\xc40\x86\x94G\xc0\x0c.\xb3.\x9c\xf3\x80G@*\xac\xe1\xfc\xce\x88\xf8\x86\x94G\xc0\x08\xd4d?\xad\xf5\xbfG@+\x15 \xeb\xbdy\xb0\x86\x94G\xc0\x07\xbd\xd6\x98m\x16\xc0G@+\x98\xa7).\x1f\xd1\x86\x94G\xc0\x01\x88\xdbI\xebF\x80G@,~*p\xdc88\x86\x94G\xc0\x00\x02\x1e\xb3\xbbv@G@-\x1e6U\x8d\xb9@\x86\x94G\xbf\xf1\x0f\xd3*\x06\x07\x00G@-\xf2\x97\xe1\x80\x15\xa8\x86\x94G\xbf\xe0\x81\xe15So\x00G@.;y\x07\x0c\t@\x86\x94G\xbf\xd1\n\\\x04@0\x00G@-\xd9?\x07\x0c\x07\x90\x86\x94G?\xd7\xfe;\xe0r\x02\x00G@-\xdb\x99\xda\xact\xd7\x86\x94G?\xd2\xeb\xde\xdd\x08\x1c\x00G@,\xe3r\xc2\xc7\xbf\x18\x86\x94G?\xdb\x83\xef\x13\xa5C\xffG@+\xfa;6\xd5bP\x86\x94G?\xef\xc7\x07\xc0o\xf5\x01G@*\xab\xe7_\xcb\x1f(\x86\x94G?\xf0b\xba\x10\x01_\x80G@)\xb4\"~\x83l\xd1\x86\x94G@\x01j\xb7|\x0e\x7f\xc0G@)@\x02U\x8d\xa8?\x86\x94G@\x01<\\\x9a\xc6\xd0\xc0G@'\xe1[X\xf7rp\x86\x94G?\xfe\xf9\xcb\xfb\x86\xbe\x80G@'HD\xd7B\x87\xb0\x86\x94G?\xf7'\xa4Mr!\x80G@'\x18n\xa7y#\xe8\x86\x94G?\xf3\xe5@vg\xd5\x80G@&8\x94\xde\x16#x\x86\x94G?\xec\xc98dF\xf2\x00G@%\xfe\xa3D|\x88\xe0\x86\x94G?\x98_\xb0_\xd9\x1f\xffG@&\t\x90\xa7y\x1f@\x86\x94G\xbf\xdc\x13\xaf\xa4\xad\x9c\x00G@&2Y\xbf]\xd1o\x86\x94G\xbf\xe8^\xd4h\x86\xc3\xffG@%\xdf\xb5=\xa8\xe7\xe8\x86\x94G\xbf\xf3@\xf4\x04z\x1e\x80G@&\x05\x07\x07\x0b\xe5 \x86\x94G\xc0\x07\x85\xf5P\xbf\x01@G@%\xec\xe5\xbf]\xd0@\x86\x94G\xc0\x07\xb6\x0f<D \xc0G@$\xcaiR#\xc4p\x86\x94G\xc0\x06\x9e\xb6a\xd0\x0e@G@#H\xf0\xa0\xa5r\xb8\x86\x94G\xc0\x0c\x18^}\x1e\xa7\xffG@#\xcc\xf7\x8f\x94c\xe8\x86\x94G\xc0\x0f\xd7\xf5\xbd\xf9,\x80G@#\xb9\x85)-\xfd0\x86\x94G\xc0\x11R,A\xf92\xa0G@#8\xbfX\xf7]\xf0\x86\x94G\xc0\x13\x1e\x99\xcd\xeb\x93@G@#\xa4\xdb0\x01\x9d@\x86\x94G\xc0\x13\xd1\x90\xa1\x8c\x03\xa0G@$N0\x85V\xf5\x80\x86\x94G\xc0\x15\x9e\x0b\xb9p\xbd\x00G@$\xbd\xd16\xd5B\x81\x86\x94t\x94\x85\x94uauK\x1b}\x94(h\x02\x8c\x0266\x94h\x04h\x05h\x06}\x94(h\x08\x8c\x06Africa\x94h\nG@\xa9\x0c\x00\x00\x00\x00\x00h\x0b\x8c\x03CAF\x94h\r\x8c\x14Central African Rep.\x94h\x0fJ\x1e\xd5U\x00uh\x10(G@,\xeb7j\xf6d\xb0G@\x02$ E$#\x80G@;_\xcdHAdXG@&H\xe8\x03\xa2\x16\x18t\x94h\x12]\x94}\x94(h\x04h\x15h\x16(G@;_\xcdHAdXG@\x14\xef\x8f\x1f(u\xc0\x86\x94G@;\x0bG\xdeq,GG@\x14\x82\xeb\xd0\xa6\xbe\xf0\x86\x94G@:g\x1bU\xe8\xa0\xe7G@\x14\x9a~\xda\xe40\x01\x86\x94G@9\xa6\x84>\x03\xec G@\x15\x06;\xdeN\x02\x10\x86\x94G@9G_U\xe8\x9b\xf8G@\x14\xae\x7f~\xbb:\x8f\x86\x94G@9 \xfb:\x9a\x19\x98G@\x13\xb5\x7f\xa7\xb0\xf8\xe0\x86\x94G@8\xce\x16`&\n\x88G@\x13\x96\xc7\xd0\xa6\xba\xe0\x86\x94G@8i\x18\x8f\xefk\xc8G@\x14oe\x18T\xd3\x10\x86\x94G@7L\x1670AXG@\x12pScl\xaf\x00\x86\x94G@6\xd7k3\xc6o G@\x12\xd7+U\xc5o\xf0\x86\x94G@6\xb4Aq7\x12XG@\x12\x88>{Q`\xf0\x86\x94G@6g\xb60\\\x9d\x01G@\x10\x1d\xdc\"\x920\xc0\x86\x94G@5\xa8\xbcD\xd7z\xf8G@\x10\xe5\xb9\xe5!\x90P\x86\x94G@4\xedv\x9d\x96\x9dPG@\x11J\x88K\x87\xf8p\x86\x94G@4Ji\xf2\xeb\xef\xd8G@\x12\xc4G,\xcf\xad\x10\x86\x94G@3w\xc0\xab=\xd7\xc0G@\x14 H\xd4\x10\x8dp\x86\x94G@2\xee\xac\x07f\xcb(G@\x12\xd6\x88\xc6iF\xf0\x86\x94G@2\x8b\x00\xe1\xda\xd7\x18G@\x10\xce\xa0\xc6i>\x00\x86\x94G@2s\xfc\x15\x0e\t\xe8G@\x0c\x08\xfbxW\x82@\x86\x94G@1\xcfU\xa1\x00_\xd0G@\x0c{HE$P\xff\x86\x94G@1\"\x0f\x11\xa43\xe0G@\r\xd3X\xb2^]\xc0\x86\x94G@0\x89|\xa4j*pG@\t\x96\x06\x90<(\xdf\x86\x94G@0\x03JK\xab\x02\x90G@\x02$ E$#\x80\x86\x94G@/\xd0\x947\xc3>@G@\x04u\x88\x97\x0f\xb2\xc1\x86\x94G@/\xb9\xb8\x0e\xcd{QG@\x08\x1b\xb9q\x83\xd0\x80\x86\x94G@.\xcf\x90\x0e\xcdw@G@\n\xae\xb2\x15Z\xe6 \x86\x94G@.\x12\x8b]O(\xf0G@\x0e\xcf\x99\xa8 \xf1`\x86\x94G@-\xe6\xe3]O(0G@\x10\xd7pK\x87\xf6\x80\x86\x94G@,\xf4\xed;-\x01\xe0G@\x12\xee0\"\x92= \x86\x94G@-\x1e,\xdb\x9a<\xa0G@\x14\x1fT\xfd\x06P\x00\x86\x94G@,\xeb7j\xf6d\xb0G@\x15\xce\x9aR[\xac\xc0\x86\x94G@-\x12\xb8\x01&.\xc0G@\x18\xe8g\xdeN\x13 \x86\x94G@-\x8d\x97]O&\xa0G@\x19\xa2MN\xf1\xed`\x86\x94G@.\x8f\x15q\xca\x0c`G@\x1d\xb0\x0c\xfd\x06z\x10\x86\x94G@0\x1b2\x00\x93\x1e@G@\x1d\xfd\x04\x9ds\xb5@\x86\x94G@0Jb>\x03\xc2\xf0G@\x1f\x04i%\xfcC\x00\x86\x94G@0t\xc8\x82H\x07\xf0G@\x1e\xf0h\x82%7\xe0\x86\x94G@0\xb4\xbb\xa7\xd3\xfb`G@\x1e\x08\x87\x03\xda\x1c\x00\x86\x94G@1\xf7\x05\xa1\x00`\x80G@\x1f\x90K\xc2\xff\xaf@\x86\x94G@2c\xb9\xdeq\x068G@ \x90\x07\x07\x0b\xcd \x86\x94G@2\xe98\xb8\xe5\x16(G@!C\x04\xa0\xa5i\xd0\x86\x94G@2\xcf\xdf\xdeq\x08\x10G@!\xf7@\x92\xfe,\x10\x86\x94G@3\x18\x10\xe8\xaey\xf0G@\"&RU\x8d\x89\x08\x86\x94G@4\x0fG\x8c\x85\x88pG@\"\x06\x81f\x9e\x99\x88\x86\x94G@5\x008\xe8\xae\x82XG@\"\xf3\xb4U\x8d\x8c\x90\x86\x94G@5\xb9L`%\xfc\xf8G@%\"U\x1b\x86\xc2\x88\x86\x94G@6;+HAM\xb0G@%\xf1\x9bf\x9e\xab\x00\x86\x94G@6\xdd9\xf2\xeb\xfb(G@&H\xe8\x03\xa2\x16\x18\x86\x94G@6\xfa@K\xab!8G@%m\xce\nu\xb2\xc0\x86\x94G@7\x8d\xe6\xe1\xda\xed G@$-\xb2\xde\x16\x1ay\x86\x94G@7\x8e\xa7\xec\x18]\xc7G@#\\\xc8\xa0\xa5s\x10\x86\x94G@7e\x10>\x03\xe20G@\"\x87\xb6\xf9d\x94\xf0\x86\x94G@7u\x81\xdeq\x1c\x80G@!\xe8\x98%\xc4$\xf8\x86\x94G@7\xceI\xc9\xf6<\xc0G@!U'\xbf]\xbc@\x86\x94G@8\x91?\x18w\xf5\x00G@ uX\x1e\xf0~(\x86\x94G@9\x1dl70IXG@\x1fL\xe8\x14\xeb2\xb0\x86\x94G@9\x1f\xc7\n\xd0\xb6\x98G@\x1e\x00\x16R[\xd0\xd0\x86\x94G@9\xcb\xf1\x1fK\x9a\xd8G@\x1b\xea\xd1\xc9\xd3?\x10\x86\x94G@:6\xa2\x96\xc3\x14\x1fG@\x1a/\xb8\xc6ig@\x86\x94G@:wE\xd7\x9d\x89PG@\x17\xc9pK\x88\x15\x10\x86\x94G@;6\xa1\xf9\xbf\xae\xb8G@\x164-%\xfc\x1b\xc0\x86\x94G@;_\xcdHAdXG@\x14\xef\x8f\x1f(u\xc0\x86\x94t\x94\x85\x94uauK\x1c}\x94(h\x02\x8c\x0267\x94h\x04h\x05h\x06}\x94(h\x08\x8c\x06Africa\x94h\nG@\xdd\x8f\x80\x00\x00\x00\x00h\x0b\x8c\x03COG\x94h\r\x8c\x05Congo\x94h\x0fJ2\x9aK\x00uh\x10(G@&0\x02\xfd\xbc@@G\xc0\x14&\xe5\xff\x90\x90\xf0G@2s\xfc\x15\x0e\t\xe8G@\r\xd3X\xb2^]\xc0t\x94h\x12]\x94}\x94(h\x04h\x15h\x16(G@2s\xfc\x15\x0e\t\xe8G@\x0c\x08\xfbxW\x82@\x86\x94G@2d\xcf\x93Y!\x88G@\x074\x1b\xae\xf4p`\x86\x94G@2\x18\"t\xa0\xceHG@\x02\xec\xffj\xb0\x19@\x86\x94G@1\xe6\x1a\x15\x0e\x07xG?\xfb\xde\x8b4\xf2\xcc\xc0\x86\x94G@1\xc61q6\xfc\xa8G?\xeba\x8eN\x96\x9a\xff\x86\x94G@1\xd3\x98\"\xb5H\x00G?\xd2}\xb7\xeb\xaeK\x00\x86\x94G@1\xa9\xde\x96\xc2\xee\x88G\xbf\xad\xbd/\x91\x7f\xa0\x00\x86\x94G@1\xa3~70(pG\xbf\xdb0q\n\x15\r\x00\x86\x94G@1\x86\x12D\xd7h\xc8G\xbf\xe7\xcdu\x1b:}\x00\x86\x94G@0\xdd\x84\xbf\xb8\xad\x98G\xbf\xf3\x9c\xf1\x97\xda\xd1\x00\x86\x94G@0h7,\xf2\xb2`G\xbf\xfb\xda\xd6IY@`\x86\x94G@/\xf2\x13B\x00\xafpG\xc0\x05\xb2\xfa\xb7r\xbb\x90\x86\x94G@0\x01\x9c0\\\x80\xd8G\xc0\x0cG\xf3\xad5g\xe0\x86\x94G@/\x81\xd0\x01&9\x80G\xc0\x0e\xd7`\xb0\x9fCP\x86\x94G@.W\x8c7\xc37\xbfG\xc0\x11_\xc0XO\xaa@\x86\x94G@-*K\x0bc\x9f\xd0G\xc0\x13\xe1\x86Q|\x14\xd8\x86\x94G@,k\x06\x9e)\x95\xb0G\xc0\x13, XO\xb2(\x86\x94G@,J7\xaf:\xa60G\xc0\x12\n?\xb4x\xa2\xf0\x86\x94G@+3Q\xfaR\x86 G\xc0\x12\x00$J\xa8l(\x86\x94G@*\x848\x0e\xcdd_G\xc0\x13\x88%\xff\x90\x8e0\x86\x94G@)\xfd\xb4n`(\x10G\xc0\x13\x1f\xd9\x84\xafD\xb8\x86\x94G@)=\xd47\xc3!PG\xc0\x11\xc0\x892\xc3\xb9\x90\x86\x94G@(\xa3 \x89\xae\xa3\xd0G\xc0\x12l\xc7\x99*\"\xf0\x86\x94G@'\xd4v\x07\xf9\xb8 G\xc0\x14&\xe5\xff\x90\x90\xf0\x86\x94G@&0\x02\xfd\xbc@@G\xc0\x0f\xd4\xa3\t^m0\x86\x94G@'\xb5\xd2\x82\xda\xffPG\xc0\x0bj;$\xac\xdb\x80\x86\x94G@&\xf4\xc1\x7fq+\xc1G\xc0\x06\x1f\xfc\xd9\x94\xdf\x90\x86\x94G@'\xa4UV{l@G\xc0\x04\x1d\x00\xb0\x9f\x14 \x86\x94G@(\xfd\xcc\xc0K\xa8\xc0G\xc0\x03\"-}k\xdc\xa0\x86\x94G@)&\x8b\xaf:\x98`G\xbf\xff-\x1a c\x8c_\x86\x94G@*8\x1f\xf3~\xe1`G\xc0\x03n\x0fh\xf0\xfc\xa0\x86\x94G@+\xfc\x1c\xcd\xf2\xf6\xd0G\xc0\x03\xc45b\x1d]\xb0\x86\x94G@,\x992\x15\xa1\x0e\x00G\xbf\xff\xf8\xef\xe2\xf2\xec\x1f\x86\x94G@,\xd9\xd5V{\x83 G\xbf\xf5U\xa2;\xb1\xe2\xe1\x86\x94G@,\xa2\x01\x9a\xbf\xc6pG\xbf\xe1\xaf\x1f\xc5\xe5\x0c\x80\x86\x94G@+\xaf\xc7\xbc\xe1\xe4pG?\xa3\xd8\nuT\xd8\x00\x86\x94G@,\x8dr\xb9x\x18\x00G?\xf3&\x9f\xe6p\xf1\x80\x86\x94G@,\r\xa7\x86D\xe2\xa0G?\xf6T\xb1\xd1\xf6\x1e\x00\x86\x94G@*\x90\xb5\x1274\xd0G?\xf5\x06\xe5rcR\x81\x86\x94G@*\x01\x98\x1ct\xa3\x00G?\xfdKY\xedD\xbe\x80\x86\x94G@*&\xd20\xef\x84\xf0G@\x02#\x03\xcaB\xdb\xc0\x86\x94G@,\xac\xf5\xc3\xb5\x890G@\x01\xd2\xaf\xf38\x9d\x00\x86\x94G@.J\xedV{\x89\x80G?\xffl\x9a\xc7\xb8\xd5\x80\x86\x94G@/\xe1\xc0\x1ct\xbc\xd0G?\xfb\xa4\x8c\x0ff\xd9?\x86\x94G@0\x03JK\xab\x02\x90G@\x02$ E$#\x80\x86\x94G@0\x89|\xa4j*pG@\t\x96\x06\x90<(\xdf\x86\x94G@1\"\x0f\x11\xa43\xe0G@\r\xd3X\xb2^]\xc0\x86\x94G@1\xcfU\xa1\x00_\xd0G@\x0c{HE$P\xff\x86\x94G@2s\xfc\x15\x0e\t\xe8G@\x0c\x08\xfbxW\x82@\x86\x94t\x94\x85\x94uauK\x1d}\x94(h\x02\x8c\x0268\x94h\x04h\x05h\x06}\x94(h\x08\x8c\x06Africa\x94h\nG@\xe1\x91\x80\x00\x00\x00\x00h\x0b\x8c\x03GAB\x94h\r\x8c\x05Gabon\x94h\x0fJ\xdf\n\x1b\x00uh\x10(G@!\x98\x92\xe2m\xaa`G\xc0\x0f\xd4\xa3\t^m0G@,\xd9\xd5V{\x83 G@\x02\x9d3\x0b\x1dR\x00t\x94h\x12]\x94}\x94(h\x04h\x15h\x16(G@&\x8d\x8a\xb9w\xfd\xa0G@\x02\x16\xa1\xde\xbd\xbc\xe0\x86\x94G@'\x80\xdaL=\xfb\x00G@\x02\x9d3\x0b\x1dR\x00\x86\x94G@(\xb8\x00\xb2\xa4f\xbfG@\x01\x8a\xe1\x1f\x98.\x80\x86\x94G@)\xe7\x15;,\xf4pG@\x02\x92\xabA\xbaU \x86\x94G@*&\xd20\xef\x84\xf0G@\x02#\x03\xcaB\xdb\xc0\x86\x94G@*\x01\x98\x1ct\xa3\x00G?\xfdKY\xedD\xbe\x80\x86\x94G@*\x90\xb5\x1274\xd0G?\xf5\x06\xe5rcR\x81\x86\x94G@,\r\xa7\x86D\xe2\xa0G?\xf6T\xb1\xd1\xf6\x1e\x00\x86\x94G@,\x8dr\xb9x\x18\x00G?\xf3&\x9f\xe6p\xf1\x80\x86\x94G@+\xaf\xc7\xbc\xe1\xe4pG?\xa3\xd8\nuT\xd8\x00\x86\x94G@,\xa2\x01\x9a\xbf\xc6pG\xbf\xe1\xaf\x1f\xc5\xe5\x0c\x80\x86\x94G@,\xd9\xd5V{\x83 G\xbf\xf5U\xa2;\xb1\xe2\xe1\x86\x94G@,\x992\x15\xa1\x0e\x00G\xbf\xff\xf8\xef\xe2\xf2\xec\x1f\x86\x94G@+\xfc\x1c\xcd\xf2\xf6\xd0G\xc0\x03\xc45b\x1d]\xb0\x86\x94G@*8\x1f\xf3~\xe1`G\xc0\x03n\x0fh\xf0\xfc\xa0\x86\x94G@)&\x8b\xaf:\x98`G\xbf\xff-\x1a c\x8c_\x86\x94G@(\xfd\xcc\xc0K\xa8\xc0G\xc0\x03\"-}k\xdc\xa0\x86\x94G@'\xa4UV{l@G\xc0\x04\x1d\x00\xb0\x9f\x14 \x86\x94G@&\xf4\xc1\x7fq+\xc1G\xc0\x06\x1f\xfc\xd9\x94\xdf\x90\x86\x94G@'\xb5\xd2\x82\xda\xffPG\xc0\x0bj;$\xac\xdb\x80\x86\x94G@&0\x02\xfd\xbc@@G\xc0\x0f\xd4\xa3\t^m0\x86\x94G@$!\xdc|\x07O G\xc0\x07\xc1\x80\x0c\xc8\x19\xef\x86\x94G@\"\xcf|S\x11\x86\xc0G\xc0\x01'\x8d\xb4\x08\xd7P\x86\x94G@!\x98\x92\xe2m\xaa`G\xbf\xf1\xc7\xe3\xf0\x99\xee\x80\x86\x94G@!\xa9\x01\x1f\xdeN\x80G\xbf\xe8\xee+\xb8=\xeb\x80\x86\x94G@\"\x18\xcau3\xa5\xc0G\xbf\xddf\x03\xd0\x0eC\x00\x86\x94G@\"\x95+\xe5\xd7~\xf0G?\xd11\xd3:/\xfa\x00\x86\x94G@\"\xfc[\xe5\xd7\x80\xb0G?\xf0)s\x19\xa4\x17\x80\x86\x94G@#\xa9\x1a\xfd\xbc50G?\xf1\x16\x17\xcb\"f\x80\x86\x94G@&\x91\xf5\xdf\x03\xf0\x10G?\xf0\xec.\xd5_\xd6\x80\x86\x94G@&\x8d\x8a\xb9w\xfd\xa0G@\x02\x16\xa1\xde\xbd\xbc\xe0\x86\x94t\x94\x85\x94uauK\x1e}\x94(h\x02\x8c\x0269\x94h\x04h\x05h\x06}\x94(h\x08\x8c\x06Africa\x94h\nG@\xdf\x06\x80\x00\x00\x00\x00h\x0b\x8c\x03GNQ\x94h\r\x8c\nEq. Guinea\x94h\x0fJv\xe0\x0b\x00uh\x10(G@\"\x9cyV{V G?\xf0)s\x19\xa4\x17\x80G@&\x91\xf5\xdf\x03\xf0\x10G@\x02E[\x93\xa5\xd9\x00t\x94h\x12]\x94}\x94(h\x04h\x15h\x16(G@#L^u3\xab\x10G@\x02E[\x93\xa5\xd9\x00\x86\x94G@&\x8d\x8a\xb9w\xfd\xa0G@\x02\x16\xa1\xde\xbd\xbc\xe0\x86\x94G@&\x91\xf5\xdf\x03\xf0\x10G?\xf0\xec.\xd5_\xd6\x80\x86\x94G@#\xa9\x1a\xfd\xbc50G?\xf1\x16\x17\xcb\"f\x80\x86\x94G@\"\xfc[\xe5\xd7\x80\xb0G?\xf0)s\x19\xa4\x17\x80\x86\x94G@\"\x9cyV{V G?\xf2\x93\x17\xcb\"m@\x86\x94G@#L^u3\xab\x10G@\x02E[\x93\xa5\xd9\x00\x86\x94t\x94\x85\x94uauK\x1f}\x94(h\x02\x8c\x0270\x94h\x04h\x05h\x06}\x94(h\x08\x8c\x06Africa\x94h\nG@\xef\xd2@\x00\x00\x00\x00h\x0b\x8c\x03ZMB\x94h\r\x8c\x06Zambia\x94h\x0fJ\xa0\xb6\xf3\x00uh\x10(G@5\xe3I\xa7\xd4\x12/G\xc01\xf6\x13\x19}\xf6\xacG@@\xbe+\x03\xb3\x85\x94G\xc0 y\xfc\xc2W\xc0\xc8t\x94h\x12]\x94}\x94(h\x04h\x15h\x16(G@>\xbdqG\x1b-\xfdG\xc0 \xae\x15A\xde\xa1m\x86\x94G@?(bdA\x19\xcaG\xc0!0l\xa0%?\xdd\x86\x94G@?\x8el\xd43\xcfxG\xc0!\x86+D\x0c\xad|\x86\x94G@@\x18\x8f\x07\x1dR\xf0G\xc0!\xdcX\x032:\xe0\x86\x94G@@a36\xe6\xb70G\xc0\"v\x11\x14CN\xa0\x86\x94G@@\x9d\x9e\x1f\x02\x06\xbcG\xc0#Z{D\x0c\xb5\x88\x86\x94G@@\xbe+\x03\xb3\x85\x94G\xc0%\r\x16\n\x05\xe9`\x86\x94G@@\xa8\\\x18.f|G\xc0%\x97\xd5f.\xe1\x84\x86\x94G@@\x8e\xa1\x07\x1dT\xfcG\xc0'6\xe2\xad\xdc\xfd \x86\x94G@@\xa78\xd7S\xf2hG\xc0(\xdf\x1eNJ>t\x86\x94G@@~\xf2\"k\xd6lG\xc0)\x91Wz\xa9\xd4H\x86\x94G@@X\x15\xcd\x16\x80lG\xc0+l\xfb\xb1F\xdf\xdb\x86\x94G@@\x9be)?wTG\xc0+\xf1\x97\xa3\x9f\xa1L\x86\x94G@>-\xf2{t\xa3\xe0G\xc0-\x97\x9aNJS4\x86\x94G@>F5\xa1\x00\x96\x9fG\xc0/\x03\xfc\xa7\t\x7f\x05\x86\x94G@=\x84OAm\xcdXG\xc0/J\x136e\xa90\x86\x94G@<\xf2\x8c\xf6U\xe6 G\xc00\x0b\x05kis\\\x86\x94G@<\xd3l\"\xb5x`G\xc00c\xc6\x90\xf5g:\x86\x94G@<w\xc8\xb2\x11\x9f\xc0G\xc00w\xe9\x12\xaaO\xae\x86\x94G@;\x99&{t\x98\x87G\xc01Js\xdfw \x18\x86\x94G@;\x0b_\x93YG\x90G\xc01\xf0\"|z\x8c\xc2\x86\x94G@:\xb4\xef\x18w\xfehG\xc01\xf6\x13\x19}\xf6\xac\x86\x94G@:a\xc6\x82H3\x98G\xc01\xd8\x9686H\x16\x86\x94G@9C\xa4K\xab+HG\xc01\xbc\x8d\xdfw\"\x0e\x86\x94G@9\x15\x9e\x15\x0e'\x10G\xc01\xa9l\xc0\xbe\xcf\xd0\x86\x94G@9\x13\xb3\x03\xfd\x15\xf9G\xc01\x94-\xc4(\x9f\xa8\x86\x94G@8\xae\xaem\xcdJ\xd8G\xc01Zy Q\x94n\x86\x94G@8\x08\xab&\x1f3\x80G\xc01K\xbca,\x08:\x86\x94G@77\rjct0G\xc01\x85\xea\xf0\x8820\x86\x94G@6\x8f\xfe\x96\xc3\x04\x10G\xc00\xe6\x00\xe9\xb4\x8f\x02\x86\x94G@5\xe3I\xa7\xd4\x12/G\xc00\x14\x8f4\xccp\x1c\x86\x94G@5\xef\x13,\xf2\xca\xb0G\xc0)\xcb\xff\xf5\x8b\x1c\xf9\x86\x94G@8\x04!\x85\xb1\xf9hG\xc0)\xd2t\xa7\th,\x86\x94G@7\xeeP\xe8\xae\x8f8G\xc0)!\xb6\xc9+\x87C\x86\x94G@8\x14t\xab=\xec\x08G\xc0(a\xf1\xaas2\x10\x86\x94G@7\xe7v\x9d\x96\xaahG\xc0'q\xce\xe4z\x01\x90\x86\x94G@8\x04\x94\xab=\xeb\xc0G\xc0&y\x7f(\xbeA\x90\x86\x94G@7\xe9\x86\xef\x82/\x90G\xc0%\xda\x88\xf8\xf4\xdb\xd8\x86\x94G@8A\xd4\xef\x821\x19G\xc0%\xe7k\x95\xf8E\xe1\x86\x94G@8P\x84\"\xb5d\x88G\xc0&\x86\x91/\x91\xe24\x86\x94G@8\xc8}\xd0\xc9\xe1yG\xc0&z6\n\x05\xef\xa4\x86\x94G@9k\t\xc9\xf6C\xd8G\xc0&\xa9pplV\xdc\x86\x94G@9\xc0\x97\\\xbc>xG\xc0'\x91\xe6\xf2!B\xf4\x86\x94G@:\x8d\x97&\x1f>\x98G\xc0'\xd9P,(\x17\xd0\x86\x94G@;*\x17jc\x85\x88G\xc0'7\xad\xe1\x100Y\x86\x94G@;c\x88R~\xd5\x08G\xc0(C\xf7z\xa9\xce\x8c\x86\x94G@<'\xb53\xc6\x86\x7fG\xc0(\x8b\x82\x92\x8e\x81H\x86\x94G@<\x86\x08\"\xb5w\x08G\xc0)e\xafz\xa9\xd3\x88\x86\x94G@<\xef-\\\xbcLwG\xc0*\x7fwz\xa9\xd8`\x86\x94G@=\xb3\x19\xe5D\xd8`G\xc0*\x83\xb36e\x94,\x86\x94G@=\x9d\xb2D\xd7\x9d\xf8G\xc0([\x98\x10\xd9\x98X\x86\x94G@=Wo\xae\xa7\xd3`G\xc0(\xb8\xb3m\x02\x8f\xb4\x86\x94G@<\xa4ux\n\xcc\xe8G\xc0'\xf1qs\xd6,\xb5\x86\x94G@<_K\xf9\xbf\xb3\xd0G\xc0'\x96X\xdd\xa6a\xc0\x86\x94G@<~\xfem\xcd[\xa0G\xc0%\x94k\xa3\x9f\x85L\x86\x94G@<\xacvf\xf9\xbb\xf8G\xc0#6;\xcc\x95=p\x86\x94G@<s*\xbf\xb8\xe0\x88G\xc0\"Tp,'\xff\x88\x86\x94G@<\xbc 70Y?G\xc0!\r\x99/\x91\xca$\x86\x94G@=\x00\xbe\xdb\x07d\xb0G\xc0 \xd0fw?\xdd\x90\x86\x94G@>X\x99\x18x\x0ehG\xc0 y\xfc\xc2W\xc0\xc8\x86\x94G@>\xbdqG\x1b-\xfdG\xc0 \xae\x15A\xde\xa1m\x86\x94t\x94\x85\x94uauK }\x94(h\x02\x8c\x0271\x94h\x04h\x05h\x06}\x94(h\x08\x8c\x06Africa\x94h\nG@\xd4\xb4\x00\x00\x00\x00\x00h\x0b\x8c\x03MWI\x94h\r\x8c\x06Malawi\x94h\x0fJV\xe9$\x01uh\x10(G@@X\x15\xcd\x16\x80lG\xc00\xcd!\xfa\xc5\x9f\xa6G@A\xe2\xcd\xc6B\xe6\xc4G\xc0\"v\x11\x14CN\xa0t\x94h\x12]\x94}\x94(h\x04h\x15h\x16(G@@a36\xe6\xb70G\xc0\"v\x11\x14CN\xa0\x86\x94G@@\xde\xaf%\x1c\x19HG\xc0\"\xd5\x94\xafO\r\x80\x86\x94G@@\xf8m^\xdb\xeb8G\xc0#c)7\xc2\xca\xc0\x86\x94G@A#\xd7\n=p\xa0G\xc0$Q\xeb\x85\x1e\xb8`\x86\x94G@AG\xad\xb8\x9b\xa3@G\xc0'\n@\x10\xd9\x92\x8c\x86\x94G@A#\xd7=\xbaZ\xf4G\xc0(\x8f_z\xa9\xcf\xd8\x86\x94G@AG\xad\xb8\x9b\xa3@G\xc0+(\xf5s\xd6:\xdc\x86\x94G@At\x1d\x88\xd2A\x08G\xc0+!\x7f_[Yt\x86\x94G@A\xa2LcFOxG\xc0+\xc6\x922\xfb\xc9\x84\x86\x94G@A\xd7\xea\x8c<\x12\xf4G\xc0-8\xda\xff\xc8\x9c\xac\x86\x94G@A\xe2\xcd\xc6B\xe6\xc4G\xc0/\xcb1\x14C\x89D\x86\x94G@A\xabfj\x19\xf0\x10G\xc00\x1b\x814\xccp:\x86\x94G@A\x84S\xe4\xfb7\x10G\xc00\xcd!\xfa\xc5\x9f\xa6\x86\x94G@A0\xce,\xa9J\x1cG\xc00.\xfd\xc4(\x99\x86\x86\x94G@A'UR5<LG\xc0.\xf5\x10~\x13\xbc4\x86\x94G@ABB\xe1\x91e\xb8G\xc0.\x07\x04\xcf\xff=<\x86\x94G@A:\xd5D\x8d\xfb\xc8G\xc0-9\xdcb\xc52\xe0\x86\x94G@A\x08L3|\xe9\xd8G\xc0,\xb8K_[`p\x86\x94G@@\xe5\x14\xe4\xfb4TG\xc0,\xe7V[\xf1\x91\x08\x86\x94G@@\x9be)?wTG\xc0+\xf1\x97\xa3\x9f\xa1L\x86\x94G@@X\x15\xcd\x16\x80lG\xc0+l\xfb\xb1F\xdf\xdb\x86\x94G@@~\xf2\"k\xd6lG\xc0)\x91Wz\xa9\xd4H\x86\x94G@@\xa78\xd7S\xf2hG\xc0(\xdf\x1eNJ>t\x86\x94G@@\x8e\xa1\x07\x1dT\xfcG\xc0'6\xe2\xad\xdc\xfd \x86\x94G@@\xa8\\\x18.f|G\xc0%\x97\xd5f.\xe1\x84\x86\x94G@@\xbe+\x03\xb3\x85\x94G\xc0%\r\x16\n\x05\xe9`\x86\x94G@@\x9d\x9e\x1f\x02\x06\xbcG\xc0#Z{D\x0c\xb5\x88\x86\x94G@@a36\xe6\xb70G\xc0\"v\x11\x14CN\xa0\x86\x94t\x94\x85\x94uauK!}\x94(h\x02\x8c\x0272\x94h\x04h\x05h\x06}\x94(h\x08\x8c\x06Africa\x94h\nG@\xe1\x18@\x00\x00\x00\x00h\x0b\x8c\x03MOZ\x94h\r\x8c\nMozambique\x94h\x0fJ\x8a{\x95\x01uh\x10(G@>-\xf2{t\xa3\xe0G\xc0:\xbe\x00E\xdd\xb0\x12G@DcB\xc6B\xf1\xc4G\xc0$\xa2Zi\x98\xad\x84t\x94h\x12]\x94}\x94(h\x04h\x15h\x16(G@AG\xad\xb8\x9b\xa3@G\xc0'\n@\x10\xd9\x92\x8c\x86\x94G@A\xa7\xfc\xa7\x8a\x93\xd8G\xc0&\xe0\xd7\xcc\x95M\x94\x86\x94G@BA\xcdm\x83\xc2\xdcG\xc0'q\x1e\xc9+\x7f\xd9\x86\x94G@Bc8%\xd5\xae\xf4G\xc0'0g6e\x85\x8c\x86\x94G@B\xbcS;\x10w@G\xc0'#4\x82\xbe\x8b\xc0\x86\x94G@B\xe9\xf0\x1b\x86n@G\xc0&\x89\x9e\xd7\xc6\xfb\xa0\x86\x94G@C6\xba,\xa9S\x04G\xc0&\x92\x06\n\x05\xf0\x10\x86\x94G@C\xc2\xb0 \xc4\x9b\xa0G\xc0%\xcb3\xda\xf8\xdf\x80\x86\x94G@D(\x86\x05h\x1e\xd0G\xc0$\xa2Z\xeec\x1f@\x86\x94G@D(\x85\xe5\xc62\x92G\xc0$\xa2Z\xa2\xfb\xf8\xf2\x86\x94G@D(\x85\xf9v#\xf9G\xc0$\xa2Zi\x98\xad\x84\x86\x94G@D=;\xcd\x16\x91\x8cG\xc0%\x87\xe7\xda<\x88\x81\x86\x94G@D7\xf7\xe8e\x13(G\xc0'\x85\xfe\xf2!B\xc0\x86\x94G@DG\xc8\xaa\xf4o\x99G\xc0)GB%T}\xac\x86\x94G@DL\xc0\\r\xba\xc4G\xc0,giJ\xe0}\xc4\x86\x94G@DcB\xc6B\xf1\xc4G\xc0-b.\xf2!\\\x89\x86\x94G@D=\x16\x8c<\x1d}G\xc0.\xd0\x05\xd3i\x10\xe5\x86\x94G@D\x0bm\x00I\xc3\xe4G\xc00\x19\xccS\x84\xc2\x1e\x86\x94G@C\xb9\xedp\xed\x99\x88G\xc00\xb8\x8cS\x84\xc4\xd8\x86\x94G@CD\xe8\xae^;\\G\xc01\x19\xdc\xa5pK\xa2\x86\x94G@B\xb4\xa0\x00I\xbe\x00G\xc01\x96\x1c86F\xf2\x86\x94G@B$\x00\xf6\x0cJ\xe0G\xc02\xa8\xe1IG\\\xbc\x86\x94G@A\xf2\xc0f\xb0!\x10G\xc02\xd7\x9ea,\x0f\x08\x86\x94G@A\x99e)?{\xb0G\xc03\x8d\x85\x0b\xd6\xbc\xd2\x86\x94G@Ad\xa86\xe6\xbb\xa4G\xc03\xc8\xb4\xfe/|\xfc\x86\x94G@AY\xd7\x9dM!\xdcG\xc04\x7f>86S\xc1\x86\x94G@A\x96\x8bU\x9f\x0elG\xc05A\x1d\xd1\xcf\xf0\xae\x86\x94G@A\xaf\xcc{+\x014G\xc05\xd7A\x19~\x07\xbc\x86\x94G@A\xb1cy\xbfq\x14G\xc06#\xd7\n=p\xa4\x86\x94G@A\xc8\x01}\xfc\x9b\xa6G\xc06\x17\n=p\xa3\xd8\x86\x94G@A\xc4W\xf9v\x19tG\xc07\x12\x1f'%N\x00\x86\x94G@A\xaf\x96Ka\x9e4G\xc07\x89\rIGr-\x86\x94G@A\xcd\xc1\x96y\x83lG\xc07\xb4\xe1P\x1b\x13]\x86\x94G@A\xba\xb8,\xa9L|G\xc08\x1fc]\xc2V\n\x86\x94G@A\x856\xcd\x16\x85\x98G\xc08zu-\xf8\xf4\x9e\x86\x94G@A\x1b\xa0\x1f\x02\x08\xe4G\xc08\xd0\xf9\xfa\xc5\xc2\xe8\x86\x94G@@\x81\xb0\xde'\x924G\xc09[\x89\xed\x1e\x84n\x86\x94G@@I\x8d\x8c<\x0c\x1cG\xc09\xba1\x86\xb8\x1f\xa8\x86\x94G@@T\x86\xc9\xac\xb0$G\xc0:&\t\xa2\x06\xa38\x86\x94G@@u>\x03\xb3\x84TG\xc0:7C\x12\xaaz\x8e\x86\x94G@@jAcFJ\x1cG\xc0:\xbe\x00E\xdd\xb0\x12\x86\x94G@@\t,U\x9f\x07\x98G\xc0:\xbb\xdb\xa2\x06\xa5\xca\x86\x94G@?\xfc\\\x07g\x04\x90G\xc0:J\xb2\x16\x14K\x1a\x86\x94G@?\xd6x\x9d\x96\xcdQG\xc09\xd7\xe4\x97\xc91<\x86\x94G@?\xc0\x9d\xd7\x9d\xa0\x90G\xc09{\xfa\x08m\x06\xb2\x86\x94G@?\xee;\x11\xa4t\xf7G\xc08^\x92\x16\x14B\xa6\x86\x94G@?\xab\x9f3\xc6\x95\xf0G\xc07\xa8\xb21b\xc1;\x86\x94G@?1\x000\\\xc3\xa0G\xc06@b\xf0\x88F\xfc\x86\x94G@@\x1f[\xc6B\xdf\x04G\xc05\x1d\xd21b\xb6\x0c\x86\x94G@@A\x1c\xda\xbd\xc0\xe0G\xc04e1\xdfw-\xc1\x86\x94G@@Trw\xc1+\x04G\xc04M\xe5\xf3\xf2\x0e\xa2\x86\x94G@@b\xe8\x18.eLG\xc03\xb71\x0b\xd6\xbd\x8a\x86\x94G@@NU\xd3\xea \xacG\xc03k\\\xacC\xf6B\x86\x94G@@S\xd3Ka\x98<G\xc02\xac\x0e\x16\x14)\x96\x86\x94G@@l\xc8=\xbaW\xd0G\xc01\xfa\xa3\x7f\xe4]&\x86\x94G@@l\x7fm\x83\xba\xccG\xc00\xb6\xa1Bs\xb3\xbe\x86\x94G@@*\x03\xbc\x05n\x90G\xc00d^\xf7[\xcd\xa4\x86\x94G@?\xda\x1fU\xe8\xb8\xe1G\xc00Q\xc5P\x1a\xf2\xde\x86\x94G@?\xa2\xf1\x8c\x85\xbbXG\xc00\x12m\xf3\xf1\xfc\x04\x86\x94G@?,M\xec\x18\x7fIG\xc0/\xb8\xcd\x9c\xcc\x11{\x86\x94G@>V\xc5\xbcO\x18\x9fG\xc0/\xc2\xfdX\x87\xcdd\x86\x94G@>F5\xa1\x00\x96\x9fG\xc0/\x03\xfc\xa7\t\x7f\x05\x86\x94G@>-\xf2{t\xa3\xe0G\xc0-\x97\x9aNJS4\x86\x94G@@\x9be)?wTG\xc0+\xf1\x97\xa3\x9f\xa1L\x86\x94G@@\xe5\x14\xe4\xfb4TG\xc0,\xe7V[\xf1\x91\x08\x86\x94G@A\x08L3|\xe9\xd8G\xc0,\xb8K_[`p\x86\x94G@A:\xd5D\x8d\xfb\xc8G\xc0-9\xdcb\xc52\xe0\x86\x94G@ABB\xe1\x91e\xb8G\xc0.\x07\x04\xcf\xff=<\x86\x94G@A'UR5<LG\xc0.\xf5\x10~\x13\xbc4\x86\x94G@A0\xce,\xa9J\x1cG\xc00.\xfd\xc4(\x99\x86\x86\x94G@A\x84S\xe4\xfb7\x10G\xc00\xcd!\xfa\xc5\x9f\xa6\x86\x94G@A\xabfj\x19\xf0\x10G\xc00\x1b\x814\xccp:\x86\x94G@A\xe2\xcd\xc6B\xe6\xc4G\xc0/\xcb1\x14C\x89D\x86\x94G@A\xd7\xea\x8c<\x12\xf4G\xc0-8\xda\xff\xc8\x9c\xac\x86\x94G@A\xa2LcFOxG\xc0+\xc6\x922\xfb\xc9\x84\x86\x94G@At\x1d\x88\xd2A\x08G\xc0+!\x7f_[Yt\x86\x94G@AG\xad\xb8\x9b\xa3@G\xc0+(\xf5s\xd6:\xdc\x86\x94G@A#\xd7=\xbaZ\xf4G\xc0(\x8f_z\xa9\xcf\xd8\x86\x94G@AG\xad\xb8\x9b\xa3@G\xc0'\n@\x10\xd9\x92\x8c\x86\x94t\x94\x85\x94uauK\"}\x94(h\x02\x8c\x0273\x94h\x04h\x05h\x06}\x94(h\x08\x8c\x06Africa\x94h\nG@\xc5\x9a\x00\x00\x00\x00\x00h\x0b\x8c\x03SWZ\x94h\r\x8c\x08eSwatini\x94h\x0fJ\x10c\x16\x00uh\x10(G@>\xad670a\xc8G\xc0;I/d\x96\x04bG@@\t,U\x9f\x07\x98G\xc09\xa9\x02?\n\n\xe2t\x94h\x12]\x94}\x94(h\x04h\x15h\x16(G@@\t,U\x9f\x07\x98G\xc0:\xbb\xdb\xa2\x06\xa5\xca\x86\x94G@?\xde93\xc6\x96\xd1G\xc0;-\x8c\xa5pw\xf7\x86\x94G@?Hc\xd0\xc9\xfe\x10G\xc0;I/d\x96\x04b\x86\x94G@>\xaf\x9b3\xc6\x91\x9fG\xc0:\xbel\xa5pv\x0c\x86\x94G@>\xad670a\xc8G\xc0:e\xe8u\xa7\x11\x8a\x86\x94G@>\xf3\x1d\\\xbcUXG\xc0:\x05\xccS\x84\xed\xc2\x86\x94G@?\x0bH\xcd`,\xc8G\xc09\xbb@u\xa7\x0e\x9c\x86\x94G@?UI\xd0\xc9\xfeHG\xc09\xa9\x02?\n\n\xe2\x86\x94G@?\xd6x\x9d\x96\xcdQG\xc09\xd7\xe4\x97\xc91<\x86\x94G@?\xfc\\\x07g\x04\x90G\xc0:J\xb2\x16\x14K\x1a\x86\x94G@@\t,U\x9f\x07\x98G\xc0:\xbb\xdb\xa2\x06\xa5\xca\x86\x94t\x94\x85\x94uauK#}\x94(h\x02\x8c\x0274\x94h\x04h\x05h\x06}\x94(h\x08\x8c\x06Africa\x94h\nGA\x07\x12@\x00\x00\x00\x00h\x0b\x8c\x03AGO\x94h\r\x8c\x06Angola\x94h\x0fJA=\xbf\x01uh\x10(G@'G\xba\xab\xd0\xbf\xf0G\xc01\xee>1b\xa8\x08G@8\x14t\xab=\xec\x08G\xc0\x11\xc0\x892\xc3\xb9\x90t\x94h\x12]\x94}\x94(h\x04\x8c\x0cMultiPolygon\x94h\x16]\x94((G@)\xfd\xb4n`(\x10G\xc0\x13\x1f\xd9\x84\xafD\xb8\x86\x94G@)Cb\x9e)\x87\xd1G\xc0\x13\xf7\x0f\xcf\xc7- \x86\x94G@(\xef\x9e>\x96\xc0qG\xc0\x14\xfeRzq\xdcO\x86\x94G@(\xdf\x95\x9a\xbf\xb5\xf1G\xc0\x16\xbc\xba(\x86^\xd8\x86\x94G@(][B\x00\x8e G\xc0\x17(\xe3\x8b\x82\xf6\xe8\x86\x94G@'\xd4v\x07\xf9\xb8 G\xc0\x14&\xe5\xff\x90\x90\xf0\x86\x94G@(\xa3 \x89\xae\xa3\xd0G\xc0\x12l\xc7\x99*\"\xf0\x86\x94G@)=\xd47\xc3!PG\xc0\x11\xc0\x892\xc3\xb9\x90\x86\x94G@)\xfd\xb4n`(\x10G\xc0\x13\x1f\xd9\x84\xafD\xb8\x86\x94t\x94\x85\x94(G@(\xa5\x15\xc3\xb5wqG\xc0\x18f~\xa3g\xad\xd8\x86\x94G@)xh`\xb8\xe4\xe0G\xc0\x17\xdc\xdb\xc2\x1f\xfdh\x86\x94G@*\x0c\xbb\xaf:\x9c`G\xc0\x17\xf0\x03\xa6\xd1|\x08\x86\x94G@*\xc0NL>\t@G\xc0\x17t\xfb\xa6\xd1y\xe8\x86\x94G@0S\x97\\\xbc\x15\x08G\xc0\x17\x82\x87\x99*9H\x86\x94G@0\x92\xbb\xec\x18?\x10G\xc0\x1a}\x96\x88\x195Y\x86\x94G@0\xdc5x\n\x99\x10G\xc0\x1c\xe3\xa2\r7\xf88\x86\x94G@1\x17\t\xf9\xbf\x820G\xc0\x1e.\xc9\x17un\x88\x86\x94G@1y\x14\x8f\xefM@G\xc0 #\x19!\xea\x85@\x86\x94G@2\"\\YRL\xc0G\xc0\x1f\xf3a\xbbL\x80\x90\x86\x94G@2v\xd470,\x18G\xc0\x1fcW\xb4x\xdd\xa9\x86\x94G@3\x04I\xd7\x9dh\x88G\xc0\x1f\xf3\xf6\xbe\xb6P\xc8\x86\x94G@3*\xa7,\xf2\xbe\x88G\xc0\x1e\xf3\xe6l\xca\xc7@\x86\x94G@3j\xe1q7\x03\xe8G\xc0\x1c\x9f(\xaa;`\xd7\x86\x94G@4\t\xa8703\x00G\xc0\x1cw'b\x8dK\xb0\x86\x94G@4\x17t\x82H\x17\xf0G\xc0\x1b\xc5\xb9i`\xe9\x10\x86\x94G@4\x9a\x11\x11\xa4C G\xc0\x1b\xc1\xdc\x81E\x9ax\x86\x94G@4\x83\xc6\x89\x1b\xba8G\xc0\x1d2\xcb\xddn\x96\x98\x86\x94G@5\xbaex\n\xae\x80G\xc0\x1d)\xdazr\x00@\x86\x94G@5\xbf\x17\xbcN\xf2\xd8G\xc0\x1f\xae*\xb1\x0f\x0e\xb8\x86\x94G@5\xf2\xfa>\x03\xdb\xd9G\xc0 \x9c\x9f\ro\xa6\x10\x86\x94G@5\xcdB\xdb\x07E\x00G\xc0!\xd1A\xfc^\x9aL\x86\x94G@5\xe0\x0b\xec\x18V`G\xc0#\x0c#m\x02v\xc0\x86\x94G@65p\xdb\x07F\xc8G\xc0#\xca\"\xbb\x84.\xec\x86\x94G@6'\xbf\xa7\xd4\x13YG\xc0&+k\ro\xbe\x84\x86\x94G@6g\x1d\xc9\xf66\x90G\xc0%\xfctb\xc5\x13\x08\x86\x94G@6\xd6\\D\xd7\x80(G\xc0&\t\x05\xb8\x1ah\x94\x86\x94G@7t\xf0>\x03\xe2xG\xc0%\xbcX\x99b\x15X\x86\x94G@7\xe9\x86\xef\x82/\x90G\xc0%\xda\x88\xf8\xf4\xdb\xd8\x86\x94G@8\x04\x94\xab=\xeb\xc0G\xc0&y\x7f(\xbeA\x90\x86\x94G@7\xe7v\x9d\x96\xaahG\xc0'q\xce\xe4z\x01\x90\x86\x94G@8\x14t\xab=\xec\x08G\xc0(a\xf1\xaas2\x10\x86\x94G@7\xeeP\xe8\xae\x8f8G\xc0)!\xb6\xc9+\x87C\x86\x94G@8\x04!\x85\xb1\xf9hG\xc0)\xd2t\xa7\th,\x86\x94G@5\xef\x13,\xf2\xca\xb0G\xc0)\xcb\xff\xf5\x8b\x1c\xf9\x86\x94G@5\xe3I\xa7\xd4\x12/G\xc00\x14\x8f4\xccp\x1c\x86\x94G@6\x8f\xfe\x96\xc3\x04\x10G\xc00\xe6\x00\xe9\xb4\x8f\x02\x86\x94G@77\rjct0G\xc01\x85\xea\xf0\x8820\x86\x94G@5`\x8e\x9d\x96\x9fHG\xc01\xee>1b\xa8\x08\x86\x94G@2\xf4\xc8\xab=\xd5\x80G\xc01\xca\x02\x1c\xe7\xc6\"\x86\x94G@2Ch>\x03\xcb\xa0G\xc01OX\xf0\x881@\x86\x94G@,k^\xab\xd0\xd6\x90G\xc01Zd\xcef\x0fN\x86\x94G@,\x1d\xf3\xe5\xd7\xa8\xe0G\xc01lb\xacC\xedz\x86\x94G@*\xec\xba\xb9x\x10\xe0G\xc00\xf8\xa1V\xee\x96(\x86\x94G@)\xa0\xcfB\x00\x93\xb0G\xc00\xf0\xfb\xd8\xa3~\"\x86\x94G@(nP\xf6\xe8\xa9\xc0G\xc01\x1c\x96L\xb1&\"\x86\x94G@'w\xe8\xe9Ad\xa0G\xc01MH\x9e\x9c\xac\x18\x86\x94G@'G\xba\xab\xd0\xbf\xf0G\xc00\xacS\x0b\xd6\xb0(\x86\x94G@'\x8e\x9cn`\x1dOG\xc0/\x96o\ro\xe7\xf1\x86\x94G@(?E\xfaRy G\xc0-\xc1\xb2\xad\xdd\x19\xe9\x86\x94G@(Y\xea\xb9x\x05\x80G\xc0,\xe5\xf6%T\x8d\x98\x86\x94G@)\x00\x0c|\x07d\x91G\xc0+\x18l\x1e\x80\xe5;\x86\x94G@)z\x19\xdf\x03\xfc\xd0G\xc0*F\x9b\x95\xf8Y\x1c\x86\x94G@*\xa06>\x96\xc7\xe0G\xc0(\xf7\x9ei\x98\xc0\x94\x86\x94G@+Dw\x19\n\xd8PG\xc0(\x13\xc9=9)\xe4\x86\x94G@+z:\x82\xdb\x0f\xe0G\xc0&\x98\x81\x81}g\xa4\x86\x94G@+_m\x1f\xdey?G\xc0%vO\x95\xf8C\xeb\x86\x94G@*\xc6O\xd80b G\xc0$\xbfE\xaas\"\x10\x86\x94G@*=\xf2\x15\xa1\x03\x9fG\xc0#\x88\xa6\xbb\x84-\xcc\x86\x94G@)\xc00n`'\x00G\xc0\"UxU\x1d\xc2\x1c\x86\x94G@)\xdb\xad\xec\xab?`G\xc0!\xeb\r\xfc^\x9a\xbc\x86\x94G@*y\r\xb6\x0e>\xa0G\xc0! \x10\xf8\xf4\xc7\x0c\x86\x94G@)\xdd\xb7x\x9d\x98 G\xc0\x1eb\xdb\x02\xfa\x8e)\x86\x94G@)t\xe3\x86D\xd70G\xc0\x1b\xb5_}\xdb\xca\x10\x86\x94G@(tf\xd4\xc6\x87\xb0G\xc0\x19-\x83\xa6\xd1\x81x\x86\x94G@(\xa5\x15\xc3\xb5wqG\xc0\x18f~\xa3g\xad\xd8\x86\x94t\x94\x85\x94euauK$}\x94(h\x02\x8c\x0275\x94h\x04h\x05h\x06}\x94(h\x08\x8c\x06Africa\x94h\nG@\xbe\xd4\x00\x00\x00\x00\x00h\x0b\x8c\x03BDI\x94h\r\x8c\x07Burundi\x94h\x0fJ\x04\xf8\xae\x00uh\x10(G@=\x06a\x93YPHG\xc0\x11\xff\xfb\xa6\xd1a\xe8G@>\xc0\x92\xcc\xf6\xbePG\xc0\x02\xc9\xb3v\x98:\xb0t\x94h\x12]\x94}\x94(h\x04h\x15h\x16(G@>x<\x884ViG\xc0\x03O\x93\x158Z\xaf\x86\x94G@>\x87\x14\xb9\xcbhPG\xc0\x06v\x01y|\xc3\x80\x86\x94G@>\xbe5\xe7B\x99\xe0G\xc0\x08FDR?g\x80\x86\x94G@>\xc0\x92\xcc\xf6\xbePG\xc0\n\xdf\xdd\xeb\xd9\x01\x80\x86\x94G@>\x81k\x11\xc6\xd1\xe0G\xc0\x0c\x8cs\xab\xc9F\x80\x86\x94G@>\x1d\xc7%\xc3\xde\xf0G\xc0\x10\\Hj\xd2\xdc\xc0\x86\x94G@=\xc0\xe60\\\xbd@G\xc0\x11\xcf?+\xf0\x19\x80\x86\x94G@=W\n\x15\x0e9\xc8G\xc0\x11\xff\xfb\xa6\xd1a\xe8\x86\x94G@=F\xc1\x18x\t\xb8G\xc0\nY\xeb\xff \xe4\x80\x86\x94G@=\x06a\x93YPHG\xc0\x06\xb6\xcc\xd9\x94\xe2 \x86\x94G@=\xa1\xd6K\xab>xG\xc0\x07W\xc5\xcfWt@\x86\x94G@=\xf08K\xab?\xd1G\xc0\x02\xc9\xb3v\x98:\xb0\x86\x94G@>x<\x884ViG\xc0\x03O\x93\x158Z\xaf\x86\x94t\x94\x85\x94uauK%}\x94(h\x02\x8c\x0278\x94h\x04h\x05h\x06}\x94(h\x08\x8c\x06Africa\x94h\nG@\xe1\xff\x80\x00\x00\x00\x00h\x0b\x8c\x03MDG\x94h\r\x8c\nMadagascar\x94h\x0fJ\xd1K~\x01uh\x10(G@E\xa0\x893|\xfe\x0cG\xc09\x99\xf7\x9b3\x00bG@I<\xff)?\x9dLG\xc0(\x14\xc3\xda<\x93\xb9t\x94h\x12]\x94}\x94(h\x04h\x15h\x16(G@H\xc5\x92\x07\x1dy\x1cG\xc0(\xf0\x8d\xee\xb7x\xc3\x86\x94G@H\xe7\x8c\xae^T$G\xc0)\xcab\xc9+\x8a,\x86\x94G@I\x07;\xbfoe\xc0G\xc0+\x1c\x8c\xc2W\xef\x8c\x86\x94G@I\x1b\xd4\xc9\xac\xd6\xc0G\xc0-\x84\x7f\xf5\x8b-T\x86\x94G@I<\xff)?\x9dLG\xc0.s\xf9f/\x08|\x86\x94G@I0E0\x13=\x80G\xc0/i\x81\xee\xb7\x95<\x86\x94G@I\x19\xa2\x99\xe3s\xbcG\xc00\x00\x11Bs\xb0\x9c\x86\x94G@H\xee(R5^\x7fG\xc0.\xd4\x18\xebM\xc2t\x86\x94G@H\xd6\x17\xf9v8\x8cG\xc0/k\x9f\xcc\x95s$\x86\x94G@H\xee\x82\x11Z\xeatG\xc00sw'%0\xe2\x86\x94G@H\xe3$\xeb\xce\xf7\xe8G\xc00\xe0\x02\xc0\xbe\xccZ\x86\x94G@H\xbf\xd2\x85h\x90\xe8G\xc01\x1b%'%3\xc4\x86\x94G@H\xb7\xc2Y\x08\xfd\xfcG\xc01\xf3\xfc\x01\x99E$\x86\x94G@H\x85YtW~\xd4G\xc03\x1ehn\xd3Q\x18\x86\x94G@HF6\x96y\x9f\xe0G\xc04\x7f4\x0f@\x91/\x86\x94G@G\xf7\"\xc9\xac\xd1\xb8G\xc06d9ki\x8fH\x86\x94G@G\xc6\x1b\xcd\x16\xa1\x18G\xc07\xc8.u\xa7\x06\x08\x86\x94G@G\x8cA\xe8e!\xd0G\xc08\xf1\x0e\xa5pn \x86\x94G@G$(:P\xa5$G\xc09-\xaf\xbdU \xa8\x86\x94G@F\xb4j\xbfo[\x88G\xc09\x99\xf7\x9b3\x00b\x86\x94G@Fj\xb2\x8c<'\x15G\xc09X\x9a\x16\x14F\xf0\x86\x94G@F\x05\x15\x8f\xa5\xf5\x88G\xc08\xfd\x041b\xc7\x14\x86\x94G@E\xe1\xc3)?\x8e\x88G\xc08u\xee\xf0\x88P\xb5\x86\x94G@E\xd9P\xc6B\xf84G\xc07\x92\xf9IGrX\x86\x94G@E\xac>f\xb01tG\xc06\xc6\xe3-\xf8\xed\"\x86\x94G@E\xa0\x893|\xfe\x0cG\xc06\x0e\xb2\x9e\x9c\xc1\x03\x86\x94G@E\xb7vKa\xaf\xf0G\xc05V#;\xa0'\xa5\x86\x94G@E\xf2d3|\xffsG\xc05)\xce\x83N;_\x86\x94G@E\xf2\xbcA$@OG\xc04\xd4\x98\xfe/\x81\x97\x86\x94G@F/\xe9\xe4\xfbK\x9cG\xc04\x12\x86\x97\xc9\x17\xda\x86\x94G@F;q_\xdc\x93|G\xc03oy\xed\x1ejb\x86\x94G@F\x1d\xc0\x00I\xcd\x00G\xc02\xf6EIG^\x10\x86\x94G@F\x05\x80=\xbaplG\xc02T\xd5\xca\xfcCd\x86\x94G@E\xfbFY\x08\xf1\xf3G\xc01h\xf2#\xbbd\xe4\x86\x94G@F'\xfe\xf9v,\xc0G\xc00\xd9\xba\x16\x14!\x92\x86\x94G@F9'{+\x15'G\xc007Z#\xbb_\xa4\x86\x94G@Fx\xf3\xae^ItG\xc00-\xebr=\x14b\x86\x94G@F\xc0Y\x85h\x88\x1cG\xc0/\xf2\xe1\x14C\x89\xf4\x86\x94G@F\xef\xbeA$D\xa8G\xc0/\x96?\xa3\x9f\xb1P\x86\x94G@G'\xf7\x96y\x9a\xf8G\xc0/\x8f^\x92\x8e\xa0 \x86\x94G@Gp\xeb\\r\xc8\x99G\xc0.k\x9d\x06\x9cB`\x86\x94G@G\xdaA\xb1\xc8\x1f\xbcG\xc0-0Hpls\x90\x86\x94G@H\x00\xaa\xe1\x91\x83`G\xc0,.\xb6\n\x06\x08\xbc\x86\x94G@G\xef<\xf2\xa2\x94$G\xc0+S\xe6\x92\x8e\x8d\x84\x86\x94G@H%\x9c%\xd5\xc8HG\xc0+\x91qX\x87\xba\xf4\x86\x94G@Hl*\xef8\xc6\x14G\xc0*-\xa8U\x1d\xe4\xa0\x86\x94G@Hn\x87tW~pG\xc0(\xf9\xc9\xd3h\xf78\x86\x94G@H\x98\xeaU\x9f-@G\xc0(\x14\xc3\xda<\x93\xb9\x86\x94G@H\xc5\x92\x07\x1dy\x1cG\xc0(\xf0\x8d\xee\xb7x\xc3\x86\x94t\x94\x85\x94uauK&}\x94(h\x02\x8c\x0280\x94h\x04h\x05h\x06}\x94(h\x08\x8c\x06Africa\x94h\nG@\xaav\x00\x00\x00\x00\x00h\x0b\x8c\x03GMB\x94h\r\x8c\x06Gambia\x94h\x0fJ#M\x1f\x00uh\x10(G\xc00\xd7n(c50G@*B\xb4\x99\xd1\xf0\xf8G\xc0+\xb0\x9f\t\x18;\x87G@+\xc0\xc3\x88\xc0\xe6xt\x94h\x12]\x94}\x94(h\x04h\x15h\x16(G\xc00\xb6\xb6\xee\\a\x00G@+0\x9ej\x08\x92\x10\x86\x94G\xc0/?\xcb\x16\xbf\x8c\x08G@+?F\xd7B\x99(\x86\x94G\xc0.\xcc+\x9fH\x12\x97G@+\xb8\x82A\x12\xd1\xd8\x86\x94G\xc0.)\xd98\xe1\xa9hG@+\xc0\xc3\x88\xc0\xe6x\x86\x94G\xc0-_\xc2\x80\x8f\xbahG@+B\xbe%\xc4N\x19\x86\x94G\xc0,\xc0\xe0\xa2\xb1\xd9\xd0G@+@Y).\x1eH\x86\x94G\xc0,\x18\x0fa\xd7b\xd7G@+\x96\x90\x11InH\x86\x94G\xc0+\xb0\x9f\t\x18;\x87G@+\x02\x94\xd0n\xf7\xb0\x86\x94G\xc0,\x8e.\xed\xc9\xbd\xa0G@*\x8f\xa8\xd7B\x96 \x86\x94G\xc0-l\xa5\x1d\x93$pG@*\x98\xae\x8c*\xb1\x99\x86\x94G\xc0.HF\x8e6\xff@G@+\x04\xde\xb5 u\xff\x86\x94G\xc0/\x06\x0cI\xf2\xbe@G@*\x8e\xa0\xaeL\xd3\x90\x86\x94G\xc0/a\xca\xd2{HWG@*\x8ak\xb8\x8aD \x86\x94G\xc0/\xdc\xd2\xd2{JwG@*B\xb4\x99\xd1\xf0\xf8\x86\x94G\xc00\xd7n(c50G@*M\x83\x81\xed?\xa8\x86\x94G\xc00\xb6\xb6\xee\\a\x00G@+0\x9ej\x08\x92\x10\x86\x94t\x94\x85\x94uauK'}\x94(h\x02\x8c\x0281\x94h\x04h\x05h\x06}\x94(h\x08\x8c\x06Africa\x94h\nG@\xff\xef\x00\x00\x00\x00\x00h\x0b\x8c\x03TUN\x94h\r\x8c\x07Tunisia\x94h\x0fJ\x18\x02\xae\x00uh\x10(G@\x1e\x19\x11\xb73\xfd\x80G@>N\xbb\xfeg\x8f G@&\xfaBY\xe59\x80G@B\xac\xcc\x9d\xec(\xf8t\x94h\x12]\x94}\x94(h\x04h\x15h\x16(G@\"\xf6\xdb\x0bcr\xf0G@>N\xbb\xfeg\x8f \x86\x94G@\"\x1cw\xf3~\xbd\xb0G@@\r%\x02\x9d\x9b\xba\x86\x94G@ \xe0\xd2\x15\xa0\xdapG@@@\xcd\xf1\x8c\x8b\x8c\x86\x94G@ \xdcf\xf0\x14\xe8\x00G@@_\xc9\x84R\x85@\x86\x94G@\x1esXT7h\xe0G@@\xac\x0b\xf4\xf6]\x9a\x86\x94G@\x1e\x19\x11\xb73\xfd\x80G@A\x0cv\xd4\x89%<\x86\x94G@ H.\xb9w\xe2\x10G@AS\xdb\xd2\xd4>Z\x86\x94G@ \xc0\xb3B\x00l\xa0G@A\xbdl\x93\xae\xb48\x86\x94G@ o\x86\xab\xd0\xa1\xe0G@B7rW\xf2\xfa\x96\x86\x94G@ \xd7\x88\xa4\xfd\x03@G@By$\x87\xbc^\xb2\x86\x94G@#\x05\x1d\xdf\x03\xe0pG@B\xac\xcc\x9d\xec(\xf8\x86\x94G@$k\x85q\xc9\xdf\xc1G@B\x9dp\xb2g\t\xfc\x86\x94G@$\\~0\xefkpG@B\\\xadE-\x02\x0c\x86\x94G@&\x0e\xc7\xaf:\x8a\xd0G@B\x8b\xca\tqG \x86\x94G@&36\x90\x829\x80G@Bs3\x11\xf9\xcf<\x86\x94G@%33\xca\x89\x08\xc0G@B4z\xe20k,\x86\x94G@%/\xc3B\x00\x80!G@A\xf9E\xdb\\\xc9\xba\x86\x94G@%\xe1\x08\x97U\xd8\x81G@A\xd9xOjpp\x86\x94G@%\x9d\x9e#H0 G@Aj\xb0]\x11\xafb\x86\x94G@$L\x97j\xf6>\xd0G@A*V\xc5,\xfc\xca\x86\x94G@$\xad\xe7\xbc\xe1\xc5\x90G@@\xe4\x93.\xfd24\x86\x94G@%\xb6\xb3B\x00\x82pG@@\xe2f\x13\xae\xb0t\x86\x94G@&7\x8dd\"\xa6\xcfG@@\xa5\x8cA\xc3*J\x86\x94G@&\xfaBY\xe59\x80G@@\x91\x89\x13\xae\xaf \x86\x94G@&\xddPS\x11\x98\x90G@@/87\x85\xb7\x9e\x86\x94G@%\xe3\xbbx\x9d\x86\xa0G@@\nx\xe7O\x19\xf8\x86\x94G@%F\x17\xf3~\xcb\xa0G@?\xc2\xecyH\xdd4\x86\x94G@#\xe6\x83\xe5\xd7\x84\xc0G@?`F\x19\xb6\x15\x89\x86\x94G@$\x1c\xf7j\xf6=\xe0G@>\xf6:\x94\x97[P\x86\x94G@#\xf0\xa6\x15\xa0\xe7\xe0G@>\x8a\x111\x9a\xc3X\x86\x94G@\"\xf6\xdb\x0bcr\xf0G@>N\xbb\xfeg\x8f \x86\x94t\x94\x85\x94uauK(}\x94(h\x02\x8c\x0282\x94h\x04h\x05h\x06}\x94(h\x08\x8c\x06Africa\x94h\nGA\"\x98\xf0\x00\x00\x00\x00h\x0b\x8c\x03DZA\x94h\r\x8c\x07Algeria\x94h\x0fJ\xe3$q\x02uh\x10(G\xc0!^i\xa6\x1bw\xf0G@3\x0e\xafk\xa1dxG@'\xff\xbf4YK\xafG@B\x8f'\x18\xcdp$t\x94h\x12]\x94}\x94(h\x04h\x15h\x16(G\xc0!^i\xa6\x1bw\xf0G@;eO|\xb2\x9a8\x86\x94G\xc0!T\x8b2\r\xd0\x88G@;\x96\xe8\x19\xb6\x04\xdf\x86\x94G\xc0!T\xc8'\xd0_\xe0G@;\xa8\x0b\x86\xf0\x0c\x00\x86\x94G\xc0!Y%\xc1i\xf9\x90G@<\xd7^\xb6\xb9t4\x86\x94G\xc0\x1c<\xa6-~\x81\x80G@=\x94HPS\x11 \x86\x94G\xc0\x18>\x16d\x1bs?G@=\xbbP\xac|\x07x\x86\x94G\xc0\x14\xf7\xf0\xbc\xda\x8a`G@>\x00\x1d\x08\xa4\xfel\x86\x94G\xc0\x13pG\x15\x99\xa9 G@>\x80M\xd5q\xcdm\x86\x94G\xc0\r\x86\x05\xf4\x96&\x00G@>\xe5\x9e\x9e\xd4\xcb\xc0\x86\x94G\xc0\r.\x13e9\xfb@G@?\xa3%\xb3O\xb0H\x86\x94G\xc0\x08\x8dE\x87\\\t?G@?\xb9x\xb3O\xb0\xac\x86\x94G\xc0\x04\xee\xce}\x1e\x88\x80G@@\x0c\x13\x89q<(\x86\x94G\xbf\xf4\xed'\xa4\xe7_\x81G@@!\xa6W\xf2\xf1j\x86\x94G\xbf\xf1\xfe)Y\xcfn\x00G@@Se\x0e\x8f\xf5\xae\x86\x94G\xbf\xf65s*\x06\x1e\x00G@@n\x98\x0b&%\xf8\x86\x94G\xbf\xfb\xbc:\xd8\x1a\xb1\x00G@@\xf5\xb9&t\xaa\x08\x86\x94G\xbf\xfc\xb0\x11\xe2X&\x00G@AC\x92\xd6>\x0eX\x86\x94G\xc0\x01[\xfb\xb7%L\x80G@A\x95\x8e\x02\x9d\xa2x\x86\x94G\xbf\xf3Vo\xf6\xd2\xde\x00G@A\xdb\x80)\xde~ \x86\x94G\xbf\xc0Pl\xf0\x9b`\x00G@A\xf1\xbf\xb0\xb2\x1e\xee\x86\x94G?\xe0\x1f\xc1\xc7CQ\x00G@B&\x90\x1c7>\x90\x86\x94G?\xf7x\x7f\x9b\xf3\xd7\x80G@BM\x85\xd7\xf2\xfa\xf6\x86\x94G@\tK(\xc3\xbc\xb6\xc0G@BdV\xff3\xd5\xce\x86\x94G@\x13CV\x16\xc6\x93\xe0G@Bn\xb9\x87\xbc^\x84\x86\x94G@\x15G\xcd\x8e>\x14 G@B[\xb6\xe3\xe5S\xf4\x86\x94G@\x19\x0c\x1avYs?G@B\x8e)\xf1\x8c\x95\xac\x86\x94G@\x1dRPo\x85\xe5\x9fG@B\x8f'\x18\xcdp$\x86\x94G@\x1e\xf2\xc4\xb3\xca1\x00G@Bq^\xdd\x11\xb3\xe6\x86\x94G@ \xd7\x88\xa4\xfd\x03@G@By$\x87\xbc^\xb2\x86\x94G@ o\x86\xab\xd0\xa1\xe0G@B7rW\xf2\xfa\x96\x86\x94G@ \xc0\xb3B\x00l\xa0G@A\xbdl\x93\xae\xb48\x86\x94G@ H.\xb9w\xe2\x10G@AS\xdb\xd2\xd4>Z\x86\x94G@\x1e\x19\x11\xb73\xfd\x80G@A\x0cv\xd4\x89%<\x86\x94G@\x1esXT7h\xe0G@@\xac\x0b\xf4\xf6]\x9a\x86\x94G@ \xdcf\xf0\x14\xe8\x00G@@_\xc9\x84R\x85@\x86\x94G@ \xe0\xd2\x15\xa0\xdapG@@@\xcd\xf1\x8c\x8b\x8c\x86\x94G@\"\x1cw\xf3~\xbd\xb0G@@\r%\x02\x9d\x9b\xba\x86\x94G@\"\xf6\xdb\x0bcr\xf0G@>N\xbb\xfeg\x8f \x86\x94G@#\x9c|\x1ct\x86\xc0G@=l\xb5\x19\xb6\x0c\xf0\x86\x94G@#\xb8Q\xa8f\xe0\x10G@<\xf5\xc1\xe3\x19\x07\x80\x86\x94G@#^&#H&@G@<$\xe8\x94\x97N\xff\x86\x94G@#\x83#B\x00x\xc0G@;\xb01\xb6\xb9o0\x86\x94G@#B\x13\xa1\x93=\xa0G@;$\x15\x86\xf0\t\xbc\x86\x94G@#n\xbd\x04\x8f\xd4`G@:\x83\x1f\xf4*\r\xc0\x86\x94G@\"\xa3\x89\xd1\\\x9d\xe0G@:\x18%\xac{\xf7x\x86\x94G@#\xd2FL=\xea\xc0G@9]\x8eo\x0bP\x80\x86\x94G@#\xe5\x82\x82\xda\xee\x80G@8\xef\xdc1\x9a\xaa\xb4\x86\x94G@$\x9b\x91\xd1\\\xa6\x8fG@8a\x1a\xac{\xef\xf0\x86\x94G@%\x8a\xf0*\x1b\xd00G@8\x90\x02\x19\xb5\xf7\x91\x86\x94G@'\x1f\x10\x0e\xcdUoG@8\x19\x10\x94\x97=4\x86\x94G@'\xff\xbf4YK\xafG@7x\xbfB\xab\xb5T\x86\x94G@!%R>\x96\x9e0G@5\x90\xcf#\xf3[\x08\x86\x94G@\x16\xb5\xd3\xd9U\xff G@3\x99\xe8\xb3O{X\x86\x94G@\x11\x11\xd6h\xb2\x0f@G@3'\xbfu\xde\xd5\x8c\x86\x94G@\tC\xdbS\x18\xdf\xc0G@3\x0e\xafk\xa1dx\x86\x94G@\t,\\\x9a\xc6\xf3\xc0G@3\xb1\x8e]\xfa&l\x86\x94G@\x05w\xfd>\x9d\xed\xc0G@3\xdb1\xe6\x82\xaf\xac\x86\x94G@\x00|\xe8\xc3\xbc\x90\x00G@4$ih7\x99\x08\x86\x94G?\xfd+\xf0\xad\x05\x02\x00G@4\x9c^\x01\xd14\xb0\x86\x94G\xbf\xf8\xcd\x06]9\\\x80G@6\xca\xec']0\xa4\x86\x94G\xc0\x13\xb1\x7fY\xdd\xee\xa0G@8\xf9}\xaf\xe5\xc2\xc3\x86\x94G\xc0!^i\xa6\x1bw\xf0G@;eO|\xb2\x9a8\x86\x94t\x94\x85\x94uauK)}\x94(h\x02\x8c\x03154\x94h\x04h\x05h\x06}\x94(h\x08\x8c\x06Africa\x94h\nG@\xc1\xe8\x80\x00\x00\x00\x00h\x0b\x8c\x03ERI\x94h\r\x8c\x07Eritrea\x94h\x0fJ\xc7PZ\x00uh\x10(G@B)_E\xe0\xb4\xe0G@(\xe9,A\x12\xc5\xa0G@E\x8ae\x9dM4PG@1\xff\x91\x12\xe2:0t\x94h\x12]\x94}\x94(h\x04h\x15h\x16(G@B6\xfa/\x05\xa7\x10G@,\xd8\x1e\xcdJ\xa1 \x86\x94G@B)_E\xe0\xb4\xe0G@-\xa5\x1dh\xc6\x93\x00\x86\x94G@B`\x7fw\xafd\x10G@0J\xb7VC\x02\xc0\x86\x94G@Bm\x1f\xb3\xfan\x00G@0\xf4\xe0u\xf6\xfd \x86\x94G@B\x95o\xa8.\x87\xe0G@1C]$\x9eE \x86\x94G@B\xf3\xb6E\xa1\xca\xd0G@1msB\xed\xbb`\x86\x94G@C4}\xd3\xea-\x80G@1\xff\x91\x12\xe2:0\x86\x94G@C~\xcc\xbc\x05}8G@0\xd73F\x15h`\x86\x94G@C\xa2\x0f\xe4\xfb@hG@/\xd8o6\xd5s@\x86\x94G@C\xe8:\xc6B\xef\xb0G@.\xdf\r)..`\x86\x94G@D\x96\xf2{+\r\xf0G@,\xfbn\xc9\x9b_\xe0\x86\x94G@D\xde\x12\xe4\xfbE\xe0G@+\xd7\x92%\xc4P\xa0\x86\x94G@E#o0\x13+\xb8G@*\xb0\x1f\xb8\x8aD\xe0\x86\x94G@EKw=\xbam@G@*\x0076\xd5Y\xc0\x86\x94G@E\x8ae\x9dM4PG@)f7\x07\x0b\xf4@\x86\x94G@Ec\xcbR5N\xf0G@(\xe9,A\x12\xc5\xa0\x86\x94G@E,\xff\xeb\x07J\x88G@)\x15\x9f+\xa9\xd2@\x86\x94G@E\x01?|\xed\x91hG@)\xbbL\xc2Pq\xe0\x86\x94G@D\xcc\x9d\x9d4X\xd0G@*\xe7xW)\xb2\xc0\x86\x94G@D\x93\xdd\x97\xf6+hG@+\x8b\xf1\xe8\xe6\x08\x00\x86\x94G@Dr\xc3\xc9\xee\xcc\x00G@,<\xbea\xcf\xff\x00\x86\x94G@D\x03\\(\xf5\xc2\xa0G@-\n\x07\xb3R\xa8@\x86\x94G@C\xab\x99\x1b\xc5XpG@-\x10'RT`\xa0\x86\x94G@C\x8c\xb9#\xa2\x9cxG@-{5*\x848 \x86\x94G@CA\xa8Xy=\xe0G@-\x02\xcc\xf6\xbe8\x00\x86\x94G@B\xf3\xfa\x1a\x0c\xf1\x90G@-\xeb:h\xb1\x9a\xa0\x86\x94G@B\xcc\x00\xa7\xc5\xacPG@,m\x1bqu\x8e \x86\x94G@B6\xfa/\x05\xa7\x10G@,\xd8\x1e\xcdJ\xa1 \x86\x94t\x94\x85\x94uauK*}\x94(h\x02\x8c\x03162\x94h\x04h\x05h\x06}\x94(h\x08\x8c\x06Africa\x94h\nGA\x11B\xc0\x00\x00\x00\x00h\x0b\x8c\x03MAR\x94h\r\x8c\x07Morocco\x94h\x0fJ_\x98\x06\x02uh\x10(G\xc01\x05:\xcc:@@G@5k\xb5;\xd8\x0b\xe1G\xbf\xf1\xfe)Y\xcfn\x00G@A\xe1GJK\xb88t\x94h\x12]\x94}\x94(h\x04h\x15h\x16(G\xc0\x01[\xfb\xb7%L\x80G@A\x95\x8e\x02\x9d\xa2x\x86\x94G\xbf\xfc\xb0\x11\xe2X&\x00G@AC\x92\xd6>\x0eX\x86\x94G\xbf\xfb\xbc:\xd8\x1a\xb1\x00G@@\xf5\xb9&t\xaa\x08\x86\x94G\xbf\xf65s*\x06\x1e\x00G@@n\x98\x0b&%\xf8\x86\x94G\xbf\xf1\xfe)Y\xcfn\x00G@@Se\x0e\x8f\xf5\xae\x86\x94G\xbf\xf4\xed'\xa4\xe7_\x81G@@!\xa6W\xf2\xf1j\x86\x94G\xc0\x04\xee\xce}\x1e\x88\x80G@@\x0c\x13\x89q<(\x86\x94G\xc0\x08\x8dE\x87\\\t?G@?\xb9x\xb3O\xb0\xac\x86\x94G\xc0\r.\x13e9\xfb@G@?\xa3%\xb3O\xb0H\x86\x94G\xc0\r\x86\x05\xf4\x96&\x00G@>\xe5\x9e\x9e\xd4\xcb\xc0\x86\x94G\xc0\x13pG\x15\x99\xa9 G@>\x80M\xd5q\xcdm\x86\x94G\xc0\x14\xf7\xf0\xbc\xda\x8a`G@>\x00\x1d\x08\xa4\xfel\x86\x94G\xc0\x18>\x16d\x1bs?G@=\xbbP\xac|\x07x\x86\x94G\xc0\x1c<\xa6-~\x81\x80G@=\x94HPS\x11 \x86\x94G\xc0!Y%\xc1i\xf9\x90G@<\xd7^\xb6\xb9t4\x86\x94G\xc0!T\xc8'\xd0_\xe0G@;\xa8\x0b\x86\xf0\x0c\x00\x86\x94G\xc0!\xa2\xbaeA\x05\x10G@;\xa8\x0b\x86\xf0\x0c\x00\x86\x94G\xc0!\x96\xfb\t\x18\x0f\x18G@;\x1e\xe5\xf4*\x10y\x86\x94G\xc0\"\xd3y\xa6\x1b~`G@;\x16\xa6]\xfaG\x00\x86\x94G\xc0#x~\xed\xc9\x95\xa8G@:\xdcf\xdf\xaf.\r\x86\x94G\xc0$`\xfc5w\xae(G@:\xdcf\xdf\xaf.\r\x86\x94G\xc0%\x1a?\x16\xbf_hG@:\xfd\xa5\x91-y\xb8\x86\x94G\xc0&\xc8\xfc\xf4\x9dD\xc0G@:\xe2(\x12\xe2a`\x86\x94G\xc0'o\xba\x80\x8f\xa0HG@:\x1a\xa5\xc0\xf6\xd8\xcc\x86\x94G\xc0(\x0f\xbf\x9fG\xf5\x00G@:\x07\xe6\xd8\xdb\x8a\x00\x86\x94G\xc0)\x00~.\xa4\"\x10G@8\xc5&W&\x9cP\x86\x94G\xc0+\xc8?\x9fH\x05PG@7\xb0\xe5\xf7\x93\xd1\x99\x86\x94G\xc0,q<\xe6\xf6\x1c\xb8G@6Of\xd8\xdby\x9c\x86\x94G\xc0-B\xfc\x80\x8f\xb9\xe8G@5\xdcf\x8d\xc3\x92\xf0\x86\x94G\xc0-\x80}\x1d\x93$\xc8G@5\x80'S\xbc\xbd\xb8\x86\x94G\xc01\x00\xc2\x1a\xbb\xf5\x0cG@5k\xb5;\xd8\x0b\xe1\x86\x94G\xc01\x05:\xcc:@@G@5l\x1c\x86\xef\xf0\xb0\x86\x94G\xc00\xf9&\xc5f\x9f\x90G@5\xe2\xc0'],\xa0\x86\x94G\xc00\x96\xd1\xad\x81\xecPG@6(\x82\x0c\x0e\xac0\x86\x94G\xc00C\rM\xef%\x00G@6\xad\xe91\x9a\xa0\xc0\x86\x94G\xc00S\x8f\xddKN@G@7\x04\x8cyH\xb6\xe0\x86\x94G\xc0/\xf7\x18\xbe\x00i\xa0G@7\xb9.\x05;\x12\xa0\x86\x94G\xc0.\xda\x1d+:k`G@8[\xf0.0\xd8\x10\x86\x94G\xc0.-\xbc\xe6\xf6$`G@8\x85/\xce\x9e\x12\xb0\x86\x94G\xc0-\xa67\xe3\x8cQ\xa0G@9\x1a\x81\x1d\x1f\xca \x86\x94G\xc0-\x9a\x12\xed\xc9\xc2 G@9\xa2\xe2B\xab\xbe\xe0\x86\x94G\xc0,\xe1?\xcf\x11m\x00G@:A!\x91-vp\x86\x94G\xc0+\x8c0'\xd0\x8c\xc0G@:\x9eo\xba#:\xb0\x86\x94G\xc0*G\xa6r\xe8k\xe0G@;\xa3\xe0\xba#?@\x86\x94G\xc0*>D\x1b\x8d\x14\xa0G@;\xa7v8\xc8\xca\x90\x86\x94G\xc0)<\xd8'\xd0\x82\xa0G@<\t\xc6\x86\xf0\r\xd0\x86\x94G\xc0'`\xba\x05\xaeX G@<&\r\x86\xf0\x0e@\x86\x94G\xc0%\xcdJ<KT\xe0G@<\xd5\x07F\x15\x9d \x86\x94G\xc0$\xcc\x97[\x03\xa2`G@=\x19<\xedVx\xc0\x86\x94G\xc0#!.\xed\xc9\x94 G@=\xee\xfe\xaf\xe5\xd8\x90\x86\x94G\xc0#\xa1\"\xc4\xd3\xd3\xc0G@?-\x80\x12\xe2t@\x86\x94G\xc0\"\xde\x9d8\xe1w\xc0G@@\x04\xe0W\xf2\xf0\xf0\x86\x94G\xc0\"\x99\xf4l\x14\xa9\xa0G@@HGi\x04\x03(\x86\x94G\xc0!P\xa0\xbe\x00)@G@@\x9e\xc0[\\\xc3\xd8\x86\x94G\xc0\x1e\x9d\xe0\xf3w\xb8\x00G@@\xd99lm\xd5\xe0\x86\x94G\xc0\x1b\xa6q\xf6\xe1{@G@A\x0e$\x17\x18\x81p\x86\x94G\xc0\x18\xfa4\xca\x81\xdd\x00G@A\x92\xab\xb7\x85\xbd\xc8\x86\x94G\xc0\x17\xb8Pj\xef\x11@G@A\xe1GJK\xb88\x86\x94G\xc0\x14\xc6\x84&\xaa\xbf\x80G@A\xe0\xa9\xcfjp\x88\x86\x94G\xc0\x12]0\xbc\xda~\xc0G@A\xaaT\xc5,\xff\x10\x86\x94G\xc0\r\x1e\xd5\xf4\x96#\x80G@A\xb3.sAz\x08\x86\x94G\xc0\x04\xd5\x9eF\x81\x84\x80G@A\x96\xec\x87\xbcZ\xd8\x86\x94G\xc0\x01[\xfb\xb7%L\x80G@A\x95\x8e\x02\x9d\xa2x\x86\x94t\x94\x85\x94uauK+}\x94(h\x02\x8c\x03163\x94h\x04h\x05h\x06}\x94(h\x08\x8c\x06Africa\x94h\nGA0\xdch\x00\x00\x00\x00h\x0b\x8c\x03EGY\x94h\r\x8c\x05Egypt\x94h\x0fJ\xb0\xba\xc8\x05uh\x10(G@8\xb37\xc9\x9a\xe9%G@6\x00\x00\x00\x00\x00\x00G@Bn\xe0\x9f\xe8h0G@?\x95\xef\x1f\xdd\xeb\xd9t\x94h\x12]\x94}\x94(h\x04h\x15h\x16(G@Bn\xe0\x9f\xe8h0G@6\x00\x00\x00\x00\x00\x00\x86\x94G@@s33330G@6\x00\x00\x00\x00\x00\x00\x86\x94G@=\x05\x1e\xb8Q\xeb\x80G@6\x00\x00\x00\x00\x00\x00\x86\x94G@9\x00\x00\x00\x00\x00\x00G@6\x00\x00\x00\x00\x00\x00\x86\x94G@9\x00\x00\x00\x00\x00\x00G@9\xae\xb8Q\xdb\xe3\xff\x86\x94G@9\x00\x00\x00\x00\x00\x00G@==\x18v\x97cs\x86\x94G@8\xb37\xc9\x9a\xe9%G@>\x0bP\t,\xcfm\x86\x94G@8\xf5&\x95\x95\xfe\xdbG@>\xa9^\x9e\x1b\x08\x9a\x86\x94G@8\xcd\x88\xe3h\xf0\x80G@?\x16\xdb\xb5\x9d\xdc \x86\x94G@9*1\xa4\xbd\xba\nG@?\x91\xb3\xd0|\x84\xb6\x86\x94G@:~\xcd\xf2f\xbaIG@?\x95\xef\x1f\xdd\xeb\xd9\x86\x94G@;u&\x95\x95\xfe\xdbG@?R>\x18i\x83Q\x86\x94G@<sR\xa8C\x80\x89G@?\x06\x98\xdc\xdb7\xc9\x86\x94G@<\xe9\xdd\x1a!\xea5G@>\xde\xbb\x98\xc7\xe2\x82\x86\x94G@=\xae\xf4\x9c\xf5n\xacG@?/\xd6\x0e\x94\xee:\x86\x94G@>\x18S\xe2\xd6#\x8eG@?y0\xbe\r\xed)\x86\x94G@>\xfa\x18\x15\xa0{6G@?\x8eL\xd7I'\x91\x86\x94G@?\xb0\x1e%\x84\xf4\xc8G@?m\xfaC\xfe\\\x92\x86\x94G@?\xf5\xddn\x04\xc0YG@>\xef\x00h\xdb\x8b\xad\x86\x94G@@\x18\xa2\xdba\xbb\x06G@?B\xa5\xa4i\xd74\x86\x94G@@\x7f8\xc5Ck\x90G@?\x06)s\x96\xd0\x92\x86\x94G@@\xe2\xfe\xc5m\\\xfbG@>\xf7\xabud0,\x86\x94G@A!\xf9\xc4\x05\xbd!G@?8'\xcc\xf6\x87@\x86\x94G@A!\xf9\xf0\x1b\x86oG@?8'\xfa\x1a\x0c\xf1\x86\x94G@Ai`\t<\xfd.G@=\xc2\xd60U1\xcc\x86\x94G@Av\x17\xc1\xbd\xa5\x12G@=\x80W)\xb2\x80\xf1\x86\x94G@AR$\x89LD|G@=\x19s\x96\xd0\x91~\x86\x94G@A6\x990\xbe\r\xedG@<X\x0f\xba\x88&\xaa\x86\x94G@A\x13\xc6\xfb\xd2s\xd6G@;\xd2\xc3\xc9\xee\xcb\xfb\x86\x94G@@\xf5\xef\x1f\xdd\xeb\xd9G@;\xa6\x114\x04\xeaK\x86\x94G@@\xcbG0@9\xacG@;\xf8\xab\x0c\x88\xa4\x7f\x86\x94G@@\x91\x81Z\x07\xb3SG@<j\xeb\x1cC,\xa6\x86\x94G@@6,f\x90W\xd2G@=\xd9\xe0`\xfeG\x9a\x86\x94G@@)\x04\xd5Q\xd6\x8cG@=\xc2\xab\x8a\\\xe5\xb4\x86\x94G@@^\x0e\x94\xee9.G@<\xb4\x89\xf4\n(x\x86\x94G@@\xac\xa4*\xed\x13\x94G@;\xb3+\xfd\xb4\xcc%\x86\x94G@A\ra\xe4\xf7e\xfeG@:$k\xce\x853\xb1\x86\x94G@A<\xa7\xc5\xacG\x1cG@9\x99;:h\xb1\x9b\x86\x94G@Ae\xc4\xda\x90\x03\xefG@9\x08\xa3\xd7\n=q\x86\x94G@A\xd8\xa0\xe4\x10\xb61G@7\xed<\xdd\xd6\xe0L\x86\x94G@A\xbf27\x8a\xb0\xc9G@7\xc0\x9bR\x00}\xd5\x86\x94G@A\xc3SP\t,\xcfG@7\x1a9\x82\x01\xcd`\x86\x94G@BXh\x87\xa8\xd6NG@64q\x0c\xb2\x95\xea\x86\x94G@Bn\xe0\x9f\xe8h0G@6\x00\x00\x00\x00\x00\x00\x86\x94t\x94\x85\x94uauK,}\x94(h\x02\x8c\x03164\x94h\x04h\x05h\x06}\x94(h\x08\x8c\x06Africa\x94h\nG@\xf60\xa0\x00\x00\x00\x00h\x0b\x8c\x03LBY\x94h\r\x8c\x05Libya\x94h\x0fJ\x1a\x85e\x00uh\x10(G@\"\xa3\x89\xd1\\\x9d\xe0G@3\x94\x99\xae\x92O@G@9*1\xa4\xbd\xba\nG@@\x91\x89\x13\xae\xaf t\x94h\x12]\x94}\x94(h\x04h\x15h\x16(G@9\x00\x00\x00\x00\x00\x00G@6\x00\x00\x00\x00\x00\x00\x86\x94G@9\x00\x00\x00\x00\x00 G@4\x00\xc7:\xbc\x94\x80\x86\x94G@7\xd9\x99\x99\x99\x99\xc0G@4\x00\x00\x00\x00\x00\x10\x86\x94G@7\xd6p\xe2\xc1+\x00G@3\x94\x99\xae\x92O@\x86\x94G@3\xd9i\x1au\xcd G@5~\xbe7\xde\x93\xb0\x86\x94G@/\xb8\xc1T\xc9\x86 G@7h\xe3h\xf0\x84`\x86\x94G@-\xb3\xdd\x97\xf6+\x80G@6\xdc\xeaJ\x8c\x15p\x86\x94G@,I\xa9q\xca\x02\x80G@6}\xc5\x1d\x1f\xbe\xa0\x86\x94G@+)\xb0|\x07n G@7\n^\x9bj\xd9\x10\x86\x94G@'\xff\xbf4YK\xafG@7x\xbfB\xab\xb5T\x86\x94G@'\x1f\x10\x0e\xcdUoG@8\x19\x10\x94\x97=4\x86\x94G@%\x8a\xf0*\x1b\xd00G@8\x90\x02\x19\xb5\xf7\x91\x86\x94G@$\x9b\x91\xd1\\\xa6\x8fG@8a\x1a\xac{\xef\xf0\x86\x94G@#\xe5\x82\x82\xda\xee\x80G@8\xef\xdc1\x9a\xaa\xb4\x86\x94G@#\xd2FL=\xea\xc0G@9]\x8eo\x0bP\x80\x86\x94G@\"\xa3\x89\xd1\\\x9d\xe0G@:\x18%\xac{\xf7x\x86\x94G@#n\xbd\x04\x8f\xd4`G@:\x83\x1f\xf4*\r\xc0\x86\x94G@#B\x13\xa1\x93=\xa0G@;$\x15\x86\xf0\t\xbc\x86\x94G@#\x83#B\x00x\xc0G@;\xb01\xb6\xb9o0\x86\x94G@#^&#H&@G@<$\xe8\x94\x97N\xff\x86\x94G@#\xb8Q\xa8f\xe0\x10G@<\xf5\xc1\xe3\x19\x07\x80\x86\x94G@#\x9c|\x1ct\x86\xc0G@=l\xb5\x19\xb6\x0c\xf0\x86\x94G@\"\xf6\xdb\x0bcr\xf0G@>N\xbb\xfeg\x8f \x86\x94G@#\xf0\xa6\x15\xa0\xe7\xe0G@>\x8a\x111\x9a\xc3X\x86\x94G@$\x1c\xf7j\xf6=\xe0G@>\xf6:\x94\x97[P\x86\x94G@#\xe6\x83\xe5\xd7\x84\xc0G@?`F\x19\xb6\x15\x89\x86\x94G@%F\x17\xf3~\xcb\xa0G@?\xc2\xecyH\xdd4\x86\x94G@%\xe3\xbbx\x9d\x86\xa0G@@\nx\xe7O\x19\xf8\x86\x94G@&\xddPS\x11\x98\x90G@@/87\x85\xb7\x9e\x86\x94G@&\xfaBY\xe59\x80G@@\x91\x89\x13\xae\xaf \x86\x94G@)S\x9d^J8@G@@ey\xd0\xa6v(\x86\x94G@**\xa1\x0e\x02!`G@@p},{\x89\x18\x86\x94G@+\xd6]9\x96\xfa\x80G@@[!\x81Z\x07\xc0\x86\x94G@.}\xc37!\xd5@G@@!\xee$5ix\x86\x94G@/m\x89\x8b.\x9c\xa0G@?`R\x93J\xca\xf0\x86\x94G@0\x9c\x93 \xd9\x94`G@?.\xa3Y5\xfc0\x86\x94G@2\x05f'|E\xc0G@>\xc3yR\xd24\xe0\x86\x94G@3\x16\x1e\xf7<\x0c G@>D2\"\x91\xfb@\x86\x94G@3\x92\xf4I\x12\x98\xa0G@>\x86\x9c#\xb7\x950\x86\x94G@4\r\xa8Xy=\xd0G@>\xfcZ\xc4q\xb4\x80\x86\x94G@3\xd2\x01%\x99\xed\x90G@?\xc0uO7u\xe0\x86\x94G@4\"K\xdb\xa0\xa5\x10G@@\x1e}Vl\xf4 \x86\x94G@4\xda\xc1\xd2\x9d\xc7@G@@Zxl\"h\x08\x86\x94G@5\x8b\x00\xbc\xbea\xd0G@@k\xed\xfaC\xfe`\x86\x94G@6\xe5P\x87\x01\x10\xb0G@@Q\xbc\xfdK\xf0\x98\x86\x94G@7<\x9e\xec\xbf\xb1`G@@\x18\x82\xbe\x8b\xc1p\x86\x94G@7\x9b\xef\xf1\x95\x03PG@@\x17\xf8\"\xbb\xec\xa8\x86\x94G@7\xedp\xa3\xd7\n@G@@\x02\">\x18i\x90\x86\x94G@8\xeb\xcf\xd4\xbf\t\x80G@?\xe6<t\xfbT\xa0\x86\x94G@9*1\xa4\xbd\xba\nG@?\x91\xb3\xd0|\x84\xb6\x86\x94G@8\xcd\x88\xe3h\xf0\x80G@?\x16\xdb\xb5\x9d\xdc \x86\x94G@8\xf5&\x95\x95\xfe\xdbG@>\xa9^\x9e\x1b\x08\x9a\x86\x94G@8\xb37\xc9\x9a\xe9%G@>\x0bP\t,\xcfm\x86\x94G@9\x00\x00\x00\x00\x00\x00G@==\x18v\x97cs\x86\x94G@9\x00\x00\x00\x00\x00\x00G@9\xae\xb8Q\xdb\xe3\xff\x86\x94G@9\x00\x00\x00\x00\x00\x00G@6\x00\x00\x00\x00\x00\x00\x86\x94t\x94\x85\x94uauK-}\x94(h\x02\x8c\x03165\x94h\x04h\x05h\x06}\x94(h\x08\x8c\x06Africa\x94h\nGA\x05S`\x00\x00\x00\x00h\x0b\x8c\x03ETH\x94h\r\x8c\x08Ethiopia\x94h\x0fJ\x84\x83G\x06uh\x10(G@@z\"\x91\xfb?\xb0G@\x0b``\xfeG\x99\x1cG@G\xe5\x0b\xb6\xedgwG@-\xeb:h\xb1\x9a\xa0t\x94h\x12]\x94}\x94(h\x04h\x15h\x16(G@G\xe5\x0b\xb6\xedgwG@ \x01\x897K\xc6\xa8\x86\x94G@F{W>\xab6{G@\x14\x01\xa8\xac\\\x13\xfd\x86\x94G@E\xd4\x97c^t(G@\x13\xd4\x87\xfc\xb9$\x00\x86\x94G@Eb\x84\x8b\xeb[-G@\x11\x02\xa6\xf3\xf5/\xc2\x86\x94G@E\x10vJ\xdf\xf8\"G@\x10\xef\xbf\xc6T\x0c\xc8\x86\x94G@D\xeds\\\xdf\"[G@\x0fY\xee~3\x0e\xaa\x86\x94G@D\x95\xfd\x8a\xda\xb9\xf5G@\x0fZK\xdb\xa0\xa5&\x86\x94G@Db]\x8dy\xd0\xa6G@\x11\x070@9\xab\xf4\x86\x94G@C\xedn\xac\x86\x05hG@\x0e\xb5\xd7\x88\x11\xb1\xd9\x86\x94G@C\xc7\x99\xe7D\xa5\xf5G@\x0b``\xfeG\x99\x1c\x86\x94G@Cr=\xc4\x86\xad.G@\x0c\x01\x83\xf9\x1edp\x86\x94G@CU\xe7\xea_\x84\xcbG@\x0c\xed\xb6\x1b\xb0_\xb0\x86\x94G@C7\xee\xa2\t\xaa\xa4G@\x0c\xb5D\xbb\x1a\xf3\xa1\x86\x94G@C\x0fz$\x89LDG@\x0c\xc9\xf1k\x11\xc6\xd2\x86\x94G@Bms\xb1\xfa#HG@\x11\xca\x9c\xe4\xd5\x9f\xdc\x86\x94G@B\x14\\\xb0J\xcd\xbfG@\x11\xca\x9c\xe4\xd5\x9f\xdc\x86\x94G@A\xe8\xa2 \x00\x00\x00G@\x13\x1b\x9c\xe3\n\xbc+\x86\x94G@A\xe8\xa2 \x00\x00\x00G@\x15ZY\x82\xd8-\x81\x86\x94G@A\xa6%\x18\xe5d\x97G@\x16\x06$\xdd/\x1a\xa0\x86\x94G@AZ\x7f\xa1\xa0\xcf\x18G@\x1a`{5*\x84\xc0\x86\x94G@A \n|Z\xc4xG@\x1bM\xe5KH\xd4\x00\x86\x94G@A\t\x9c\xe0u\xf7\x00G@\x1c\xe7_o\xd2 @\x86\x94G@@\xc8\xbd\xba\nRxG@\x1e\xdau\xcd\x0b\xb7\x00\x86\x94G@@z\"\x91\xfb?\xb0G@\x1f#\xcf,\xf9]\xc0\x86\x94G@@\xa5\xbc\x01\xa3n@G@ \xb5\x8b\x82\x7f\xa1\xc0\x86\x94G@@\xe9\xa9\xfb\xe7l\xa0G@ \xc2!Bo\xe7@\x86\x94G@@\xfc\xcc%\x07 \x90G@!^~\xa5\xf8L\x80\x86\x94G@@\xfb\x16]9\x97\x08G@#*\xca\xffm3@\x86\x94G@A \xf4\x1f!-\x80G@%B\x9b(\x0f\x12\xc0\x86\x94G@A]\x96R\xbd<HG@%\xd2\x01\xcd_\x9a\x00\x86\x94G@Ajr\xda\x12/\xc0G@&\xa3N\xb9\xa1w \x86\x94G@A\xa1W\xbc\x7fw\xb8G@(*l\xa0<K \x86\x94G@A\xee\x8bm\x86\xec\x18G@)(\x14P\xef\xdc\xc0\x86\x94G@B\"\x96\x91\xa7\\\xd8G@+ l\xca-\xb6`\x86\x94G@B6\xfa/\x05\xa7\x10G@,\xd8\x1e\xcdJ\xa1 \x86\x94G@B\xcc\x00\xa7\xc5\xacPG@,m\x1bqu\x8e \x86\x94G@B\xf3\xfa\x1a\x0c\xf1\x90G@-\xeb:h\xb1\x9a\xa0\x86\x94G@CA\xa8Xy=\xe0G@-\x02\xcc\xf6\xbe8\x00\x86\x94G@C\x8c\xb9#\xa2\x9cxG@-{5*\x848 \x86\x94G@C\xab\x99\x1b\xc5XpG@-\x10'RT`\xa0\x86\x94G@D\x03\\(\xf5\xc2\xa0G@-\n\x07\xb3R\xa8@\x86\x94G@Dr\xc3\xc9\xee\xcc\x00G@,<\xbea\xcf\xff\x00\x86\x94G@D\x93\xdd\x97\xf6+hG@+\x8b\xf1\xe8\xe6\x08\x00\x86\x94G@D\xcc\x9d\x9d4X\xd0G@*\xe7xW)\xb2\xc0\x86\x94G@E\x01?|\xed\x91hG@)\xbbL\xc2Pq\xe0\x86\x94G@E,\xff\xeb\x07J\x88G@)\x15\x9f+\xa9\xd2@\x86\x94G@E\x00\x00\x00\x00\x00\x10G@(33333\x80\x86\x94G@D\xd4\xb4\x8d:\xe6\x98G@'C,\xa5zx`\x86\x94G@D\xde\xaa\xe2\x979\x88G@&\xb5\xd0\xfaX\xf7`\x86\x94G@D\xe0\xb6\x84\x8b\xebxG@&\x1a\x10\xe0\"\x14`\x86\x94G@E(5\xbdQ.\xd8G@&\x11\x82\xa9\x93\x0c\x00\x86\x94G@EG\x07\xf2<\xc8\xf0G@&5\xd0\xfaX\xf7\x80\x86\x94G@Eco\xe1\x91w\xe0G@%\xda\x8f\xd3\xd8\xb1\x80\x86\x94G@EG\x85r\x9b(\x08G@%%)4\xac\xb0@\x86\x94G@Ev\xcc\xa2\xdba\xc8G@$\x0b;\xb8<\xf3 \x86\x94G@E\xa6\x03\xc4\xb0\x9e\xa8G@#\x14\xb9\xcbhI \x86\x94G@E\xd6\xe1G\xae\x14\x80G@\"]\xfe2\xa0f\x80\x86\x94G@Gyc4\x82\xbe\x98G@\x1f\xfc\xce\x1cX%\xc0\x86\x94G@G\xe5\x0b\xb6\xedgwG@ \x01\x897K\xc6\xa8\x86\x94t\x94\x85\x94uauK.}\x94(h\x02\x8c\x03166\x94h\x04h\x05h\x06}\x94(h\x08\x8c\x06Africa\x94h\nG@\xaa\"\x00\x00\x00\x00\x00h\x0b\x8c\x03DJI\x94h\r\x8c\x08Djibouti\x94h\x0fJ\xf33\r\x00uh\x10(G@D\xd4\xb4\x8d:\xe6\x98G@%\xda\x8f\xd3\xd8\xb1\x80G@E\xa8\xafcFa8G@)f7\x07\x0b\xf4@t\x94h\x12]\x94}\x94(h\x04h\x15h\x16(G@E,\xff\xeb\x07J\x88G@)\x15\x9f+\xa9\xd2@\x86\x94G@Ec\xcbR5N\xf0G@(\xe9,A\x12\xc5\xa0\x86\x94G@E\x8ae\x9dM4PG@)f7\x07\x0b\xf4@\x86\x94G@E\xa8\xafcFa8G@(\xc7\xc1\x88\xc0\xd9\x80\x86\x94G@E\xa4\xa8%\xd5\xbdPG@'\xf3)\xcd\x05\x1a@\x86\x94G@E[\xa1\xbfoU\x98G@'x\xa5\xe1\x7f\xf9`\x86\x94G@E\x92\x99Y\x08\xf08G@&\xec\x90w\xaf\xc0\x00\x86\x94G@Eco\xe1\x91w\xe0G@%\xda\x8f\xd3\xd8\xb1\x80\x86\x94G@EG\x07\xf2<\xc8\xf0G@&5\xd0\xfaX\xf7\x80\x86\x94G@E(5\xbdQ.\xd8G@&\x11\x82\xa9\x93\x0c\x00\x86\x94G@D\xe0\xb6\x84\x8b\xebxG@&\x1a\x10\xe0\"\x14`\x86\x94G@D\xde\xaa\xe2\x979\x88G@&\xb5\xd0\xfaX\xf7`\x86\x94G@D\xd4\xb4\x8d:\xe6\x98G@'C,\xa5zx`\x86\x94G@E\x00\x00\x00\x00\x00\x10G@(33333\x80\x86\x94G@E,\xff\xeb\x07J\x88G@)\x15\x9f+\xa9\xd2@\x86\x94t\x94\x85\x94uauK/}\x94(h\x02\x8c\x03167\x94h\x04h\x05h\x06}\x94(h\x08\x8c\x06Africa\x94h\nG@\xc7\xed\x00\x00\x00\x00\x00h\x0b\x8c\x03-99\x94h\r\x8c\nSomaliland\x94h\x0fJ\xe0g5\x00uh\x10(G@EG\x85r\x9b(\x08G@\x1f\xfc\xce\x1cX%\xc0G@Hy^\xd3\xeaD\xa8G@&\xec\x90w\xaf\xc0\x00t\x94h\x12]\x94}\x94(h\x04h\x15h\x16(G@Hy^\xc6\x05\xdb\x0fG@&\xd2<m\xa8!b\x86\x94G@Hy^\xc6\x05\xda\xffG@&\xd2<m\xa8!^\x86\x94G@Hx\x93\xa0\xb7\x11dG@&\xc9\xdd=\xa8\xeb\xf0\x86\x94G@Hx {+\x1f\x08G@%\xf6\xf3\x9d;\xaeH\x86\x94G@Hx\x18\x03\xb3\xa7\x90G@#\xf2n\x99\xd1\xd50\x86\x94G@Hx\x14\xa0\xb7\x11`G@\"\xe7K\xa4\x0fA@\x86\x94G@H>M\\r\xcc G@!\xac\xddX\xf7W(\x86\x94G@G\xe5\x0b\xb6\xedgwG@ \x01\x897K\xc6\xa8\x86\x94G@Gyc4\x82\xbe\x98G@\x1f\xfc\xce\x1cX%\xc0\x86\x94G@E\xd6\xe1G\xae\x14\x80G@\"]\xfe2\xa0f\x80\x86\x94G@E\xa6\x03\xc4\xb0\x9e\xa8G@#\x14\xb9\xcbhI \x86\x94G@Ev\xcc\xa2\xdba\xc8G@$\x0b;\xb8<\xf3 \x86\x94G@EG\x85r\x9b(\x08G@%%)4\xac\xb0@\x86\x94G@Eco\xe1\x91w\xe0G@%\xda\x8f\xd3\xd8\xb1\x80\x86\x94G@E\x92\x99Y\x08\xf08G@&\xec\x90w\xaf\xc0\x00\x86\x94G@E\xbc>\x93\x0f\xc4\x80G@&\x8e/\xfc\xcew\x00\x86\x94G@E\xd5UcFb\x00G@%\xbatc4\xd9\xc0\x86\x94G@F\x0f\x140\x13/\xc0G@$\xe4\x1d\x9d;\xa9\x80\x86\x94G@FN\xa0\n\x87>\x80G@$\xe2h\xbb\xf3\xfb\xc0\x86\x94G@F\xc7I\xd3\xea= G@%ed\x1e\xf0\x94\x00\x86\x94G@GR\x9c\x81\xfe\xbaxG@%\xa2\x12\xc2\xc7\x9f@\x86\x94G@G\xc3H\xbfo`@G@&A$\nu\xb6\xa0\x86\x94G@H\x02\xc3\xaa\xf4\x80\x00G@&b\xd9D|\x8a\xa0\x86\x94G@H0{\xfc\xe0\x05\xf0G@&\xc0?\"Zj@\x86\x94G@Hy^\xd3\xeaD\xa8G@&\xd2=\x008H@\x86\x94G@Hy^\xc6\x05\xdb\x0fG@&\xd2<m\xa8!b\x86\x94t\x94\x85\x94uauK0}\x94(h\x02\x8c\x03168\x94h\x04h\x05h\x06}\x94(h\x08\x8c\x06Africa\x94h\nG@\xf4\xbc \x00\x00\x00\x00h\x0b\x8c\x03UGA\x94h\r\x8c\x06Uganda\x94h\x0fJ\xcd\xca[\x02uh\x10(G@=\x94W\xe5D\xd7\xd8G\xbf\xf7\x17\xd9E\xef[@G@A\x84\x9bR\x00}\xd4G@\x10\xff\xe1\xd6\xf2\xbb\x10t\x94h\x12]\x94}\x94(h\x04h\x15h\x16(G@@\xf3\xac\xce\xfaK\xe6G\xbf\xeefffffg\x86\x94G@?\xdd\xbdQ.\xc6\xd0G\xbf\xf0p\x11\n\x13~\x00\x86\x94G@>\xc5\x15\x8b\x82\x7f\xc0G\xbf\xf0;\x98\xc7\xe2\x82\x00\x86\x94G@>kJt\xa1\x04\x80G\xbf\xf2'\x90P,\xb6\x00\x86\x94G@=\xd2O\n\xd0\xcb@G\xbf\xf7\x17\xd9E\xef[@\x86\x94G@=\x94W\xe5D\xd7\xd8G\xbf\xf5v\x04\xcb\x0e\x0c`\x86\x94G@=\x96|\x89\x1b\xe2 G\xbf\xe2\xcc\x07\x06\xbf\x82\x00\x86\x94G@=\xd1\xca\xf6U\xe9\xf8G\xbf\xcaG\x9af\x15$\x00\x86\x94G@=\xe03\n\xd0\xcbpG?\xe3\x1d\xbcc\x11Z\x00\x86\x94G@>\x16\x0e)\x89\x1eXG?\xf0\xff;\xa2,\xa3\xc0\x86\x94G@>w\xf0\x1b\xe1\xdf0G?\xf9WDaRT\x00\x86\x94G@>\xdaH\x96\xc3(\x89G?\xfd\x97 \xc0\xe5-\x00\x86\x94G@?,\x95\n\xd0\xd18G@\x01\xa2\xbe\xab\x8a\x87\xa0\x86\x94G@>\xc5\xfa\x0e:\x9f\xa9G@\x02\xb8\x14\xc0\x05m\xc0\x86\x94G@>\xd5w\xd7\x9d\x9c\x80G@\x0c\x12\xc5\x9ay\xa4\x80\x86\x94G@>\xd5wZ0\xf5JG@\x0c\x12\xc8\x8f\xe1\x7f\xa8\x86\x94G@?>\xdd\x05)4\xc0G@\x0eAT\xc9\x85\xf0\x80\x86\x94G@?\xe1\xa6\xb5\x0b\x0f G@\x0cwVC\x02\xb4\x00\x86\x94G@@W\xdc\x9cM\xa9\x08G@\x0eV\xab\xde?\xbc\x80\x86\x94G@@\xb1\xeb\x85\x1e\xb8`G@\x0eQ\xeb\x85\x1e\xb8\x00\x86\x94G@A\x00\xa3\xd7\n=qG@\x10\xff\xe1\xd6\xf2\xbb\x10\x86\x94G@A=T!\xc0D(G@\x0cq\xdei\xadC\x80\x86\x94G@ALL\x05\x92\x108G@\x08n\x0f<\xb3\xe6\x80\x86\x94G@A\x84\x9bR\x00}\xd4G?\xfe~R\x15v\x89\xca\x86\x94G@AV\x07_o\xd2 G?\xf2\xd4\xbf\t\x95\xaa\xf7\x86\x94G@A\x17\n=p\xa3\xd7G?\xe0z\xe1G\xae\x14{\x86\x94G@@\xf2`w\xce\xb8PG?\xbc\x1c\xbdq\xc6`\x00\x86\x94G@@\xf3\xac\xce\xfaK\xe6G\xbf\xeefffffg\x86\x94t\x94\x85\x94uauK1}\x94(h\x02\x8c\x03169\x94h\x04h\x05h\x06}\x94(h\x08\x8c\x06Africa\x94h\nG@\xd5t\x80\x00\x00\x00\x00h\x0b\x8c\x03RWA\x94h\r\x8c\x06Rwanda\x94h\x0fJ,\x9a\xb5\x00uh\x10(G@=\x06a\x93YPHG\xc0\x07W\xc5\xcfWt@G@>\xd0\xee70biG\xbf\xf2'\x90P,\xb6\x00t\x94h\x12]\x94}\x94(h\x04h\x15h\x16(G@>kJt\xa1\x04\x80G\xbf\xf2'\x90P,\xb6\x00\x86\x94G@>\xd0\xee70biG\xbf\xfb.\xc0\x86\xc9\xe1?\x86\x94G@>\xc2 \x89\x1b\xe7HG\xc0\x02LI\xdc\xfe\x9e\xf0\x86\x94G@>x<K\t\xe9\x90G\xc0\x03O\x86\x1a`\xd4\x00\x86\x94G@>x<\x884ViG\xc0\x03O\x93\x158Z\xaf\x86\x94G@=\xf08K\xab?\xd1G\xc0\x02\xc9\xb3v\x98:\xb0\x86\x94G@=\xa1\xd6K\xab>xG\xc0\x07W\xc5\xcfWt@\x86\x94G@=\x06a\x93YPHG\xc0\x06\xb6\xcc\xd9\x94\xe2 \x86\x94G@=\x1e\x13\x18x\t\x08G\xc0\x02Vr\xd2\xc1.p\x86\x94G@=A<\xdb\x07e\xc9G\xc0\x01\xb8\x8b\x91\xe6\xb7\xb0\x86\x94G@=J\xb9\x18x\t\xc9G\xbf\xf9\xeb\xbf\xacU\xcd\xff\x86\x94G@=\x94W\xe5D\xd7\xd8G\xbf\xf5v\x04\xcb\x0e\x0c`\x86\x94G@=\xd2O\n\xd0\xcb@G\xbf\xf7\x17\xd9E\xef[@\x86\x94G@>kJt\xa1\x04\x80G\xbf\xf2'\x90P,\xb6\x00\x86\x94t\x94\x85\x94uauK2}\x94(h\x02\x8c\x03176\x94h\x04h\x05h\x06}\x94(h\x08\x8c\x06Africa\x94h\nG@\xd4d\x00\x00\x00\x00\x00h\x0b\x8c\x03SSD\x94h\r\x8c\x08S. Sudan\x94h\x0fJQ\xc3\xc6\x00uh\x10(G@7\xe3\x11\x18\x04\x040G@\x0c\x12\xc8\x8f\xe1\x7f\xa8G@A\xa6%\x18\xe5d\x97G@(~\xfa\xdfl: t\x94h\x12]\x94}\x94(h\x04h\x15h\x16(G@>\xd5wZ0\xf5JG@\x0c\x12\xc8\x8f\xe1\x7f\xa8\x86\x94G@=\xf4\x18\x96\xc3$\x98G@\x10\xb1\xde)e\xd3\xb0\x86\x94G@=\xb7Kx\n\xd1\xa0G@\x12g9\\\x99\x0f\x00\x86\x94G@=(\xb9\\\xbcMxG@\x11\x8e\x9c\x14\xea\xf60\x86\x94G@<\xb2Yx\n\xcd G@\x11\xd1\xff\xc2\xffr@\x86\x94G@<m\xd2\x89\x1b\xdd\x08G@\x11&\x0b\xdeM\xf1\x00\x86\x94G@;\xfa\xdf\xc9\xf6O\x18G@\x11\xa27\x1f(g0\x86\x94G@;_\xcdHAdXG@\x14\xef\x8f\x1f(u\xc0\x86\x94G@;6\xa1\xf9\xbf\xae\xb8G@\x164-%\xfc\x1b\xc0\x86\x94G@:wE\xd7\x9d\x89PG@\x17\xc9pK\x88\x15\x10\x86\x94G@:6\xa2\x96\xc3\x14\x1fG@\x1a/\xb8\xc6ig@\x86\x94G@9\xcb\xf1\x1fK\x9a\xd8G@\x1b\xea\xd1\xc9\xd3?\x10\x86\x94G@9\x1f\xc7\n\xd0\xb6\x98G@\x1e\x00\x16R[\xd0\xd0\x86\x94G@9\x1dl70IXG@\x1fL\xe8\x14\xeb2\xb0\x86\x94G@8\x91?\x18w\xf5\x00G@ uX\x1e\xf0~(\x86\x94G@7\xe3\x11\x18\x04\x040G@!=M6\x82\xd0`\x86\x94G@81\xael\x13\xb5\xd0G@!u\x17\xb4;^\xe0\x86\x94G@8\x89\x94\nH\x0b\x90G@!\xd5\xc7{\xdbV\xa0\x86\x94G@8\xcb\x80@\xefK\xc0G@#\x9e\xd7\xe5\xb8F\xe0\x86\x94G@9\x11\xd1\x8cL\\\xf0G@$\x8c*D\x12a`\x86\x94G@9\xcaf\xf2 \xaa`G@$\xd2{\x8for\xc0\x86\x94G@9\xf6Y\xc1:\xd5 G@$E\xd8\xf8\xb5P`\x86\x94G@:z2.\x89U\xb0G@#\x1a\xffxi\xc6\xa0\x86\x94G@:\xc0\x83y\xe6f\xe0G@\"\xef\x0c\xa9O\x9b\xc0\x86\x94G@;\x1c\xce,\xd0\x8d\xb0G@#F\xf2G\x83\xf1`\x86\x94G@;\xd5c\x92\xa4\xdb\x10G@#5]\xf4\xac\xad@\x86\x94G@;\xf8\x8c8Sc\x90G@\"\xcb\xe4\x03\xa1\x13 \x86\x94G@<\xf7r\xe9\x84\xc2`G@\"\xcb\xe4\x03\xa1\x13 \x86\x94G@=\x00=\x12\xf0d\x80G@#5]\xf4\xac\xad@\x86\x94G@=\x84\x15\x80>\xe5\x00G@#\x96\r\xbcL\xa4\xe0\x86\x94G@=\x9es\xfc\x81\xcbpG@$+z|ri\xc0\x86\x94G@=\xff#\xc4!\xc3@G@$\x94\xf4m~\x03\x80\x86\x94G@>\xd6|\xba\xee\xc8 G@#j\x1a\xed2z \x86\x94G@?ZU(=H\xb0G@#\x9e\xd7\xe5\xb8F\xe0\x86\x94G@?\xd9\xc8\x80\xd5\xf8\x00G@%\x10\x02\xb1`\xe2 \x86\x94G@@35\x8b\xc8\rPG@&)G\xde\xd5'\x00\x86\x94G@@(8\xd8\x01\x82\x98G@'\\\xeb\x88\x8cR\xe0\x86\x94G@@\tuG\x08\xcb\x08G@'\xf2XH\xb2\x17`\x86\x94G@@V^1v\x95\xe8G@(\x0c\xb6\xc4\xf4\xfe\x00\x86\x94G@@_(Z\xe28\x18G@(~\xfa\xdfl: \x86\x94G@@\x9a|\xf2x\xbe\xb0G@([\xd29\xbd\xb1`\x86\x94G@@\x8b\x1b)\xfcb\xe0G@&\xe1\xddD\xa9t\x80\x86\x94G@@\x9a|\xf2x\xbe\xb0G@%p\xb2y\x00\xd9\xc0\x86\x94G@@\xdci)\x1f\xfe\xe0G@$\xa6\x88\xc0UH\x00\x86\x94G@@\xeb\xca\xf1\x9cZ\xb0G@#\xf6\xbd\x83\xec\x9c\xe0\x86\x94G@@\xe9\x98gAr0G@\"\xf7\xd6\xd2\xbb>@\x86\x94G@@\xfbPt\x81\xcd8G@\"\xed\xb6\xcb(\xc1@\x86\x94G@@\xfc\xcc%\x07 \x90G@!^~\xa5\xf8L\x80\x86\x94G@@\xe9\xa9\xfb\xe7l\xa0G@ \xc2!Bo\xe7@\x86\x94G@@\xa5\xbc\x01\xa3n@G@ \xb5\x8b\x82\x7f\xa1\xc0\x86\x94G@@z\"\x91\xfb?\xb0G@\x1f#\xcf,\xf9]\xc0\x86\x94G@@\xc8\xbd\xba\nRxG@\x1e\xdau\xcd\x0b\xb7\x00\x86\x94G@A\t\x9c\xe0u\xf7\x00G@\x1c\xe7_o\xd2 @\x86\x94G@A \n|Z\xc4xG@\x1bM\xe5KH\xd4\x00\x86\x94G@AZ\x7f\xa1\xa0\xcf\x18G@\x1a`{5*\x84\xc0\x86\x94G@A\xa6%\x18\xe5d\x97G@\x16\x06$\xdd/\x1a\xa0\x86\x94G@AOb\x97_\xc4OG@\x13ct$\xe3\xc2\x8b\x86\x94G@A\x00\xa3\xd7\n=qG@\x10\xff\xe1\xd6\xf2\xbb\x10\x86\x94G@@\xb1\xeb\x85\x1e\xb8`G@\x0eQ\xeb\x85\x1e\xb8\x00\x86\x94G@@W\xdc\x9cM\xa9\x08G@\x0eV\xab\xde?\xbc\x80\x86\x94G@?\xe1\xa6\xb5\x0b\x0f G@\x0cwVC\x02\xb4\x00\x86\x94G@?>\xdd\x05)4\xc0G@\x0eAT\xc9\x85\xf0\x80\x86\x94G@>\xd5wZ0\xf5JG@\x0c\x12\xc8\x8f\xe1\x7f\xa8\x86\x94t\x94\x85\x94uauu."
    return pickle.loads(dict_bytes)
